#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <ctype.h>

#ifdef _MSC_VER
#include <io.h> // _chsize
#define stricmp _stricmp
#else
#define stricmp strcasecmp
#endif

#if defined(__APPLE__) || defined(__linux__)
#define _strnicmp strncasecmp
#endif

#include "fd_image.h"

// Util

// x^16 + x^12 + x^5 + 1 poly (0x1021)
static unsigned int update_crc_notable(unsigned int crc, unsigned int data)
{
	for (int b = 0; b < 8; b++) {
		if ( (crc ^ (data << (8 + b))) & 0x8000) {
			crc = ((crc << 1) ^ 0x1021) & 0xffff;
		} else {
			crc = ((crc << 1) & 0xffff);
		}
	}

	return crc;
}

unsigned long crc_table[256];
void make_crc_table()
{
	if (crc_table[0] != 4129) {
		for (unsigned int i = 0; i < 256; i++) {
			crc_table[i] = update_crc_notable(0, i);
		}
	}
}

unsigned int update_crc(unsigned int crc, unsigned int data)
{
	return ((crc << 8) ^ crc_table[(crc >> 8) ^ data]) & 0xffff;
}

void store_crc(int crc, unsigned char* data, int len, unsigned char* res)
{
	for (int i = 0; i < len; i++) {
		crc = update_crc(crc, data[i]);
	}
	res[0] = (crc >> 8) & 0xff;
	res[1] = crc & 0xff;
}

// Store the data rate in si_flags so most formats default it to "unknown"
// Encoding is 8 bits with 0 = unknown, otherwise bit rate = 25*val KBit/second
void SectorInfo::set_rate(int rate)
{
	rate = (SECTOR_rate_scale/2 + rate) / SECTOR_rate_scale;	// A bit of rounding
	if (rate > (SECTOR_rate_mask >> SECTOR_rate_bit))
	{
		rate = (SECTOR_rate_mask >> SECTOR_rate_bit);
	}

	si_flags &= ~SECTOR_rate_mask;
	si_flags |= (rate << SECTOR_rate_bit);
}

// Method will change based on improved user mark support (see set_dam())
int SectorInfo::get_dam(WD17xx_Family chip)
{
	// This is kind of the formatting problem.  Model I should only get FB,FA here
	// It's double density that wants FB/F8 although that's a bit fake, too.
	if (chip != CHIP_WD1771)
		return (si_flags & SECTOR_delete) ? 0xF8 : 0xFB;

	// Partial WD1771 case
	return (si_flags & SECTOR_delete) ? 0xFA : 0xFB;

	// Full WD1771 case
	//int index = !!(si_flags & SECTOR_delete) + 2 * !!(si_flags & SECTOR_damhigh);
	//int dam[4] = { 0xfb, 0xfa, 0xf9, 0xf8 };
	//return dam[index];
}

int SectorInfo::get_idam(WD1010_Family chip)
{
	// For hard disk controllers, IDAM depends on cylinder
	int res;
	int high = (si_track >> 8);

	res = 0xfc | ((high & 3) ^ 2);
	if (si_track >= 1024 && chip == CHIP_WD2010) {
		res = 0xf0 | ((high & 7) ^ 2);
	}

	return res;
}

int SectorInfo::set_dam(WD17xx_Family chip, int dam)
{
	if (chip != CHIP_WD1771) {
		// 0xf8, 0xfb are the only values the WD1793 can write.
		if (dam == 0xf8) {
			si_flags |= SECTOR_delete;
		}
		else if (dam == 0xfb) {
			si_flags &= ~SECTOR_delete;
		}
	}
	else
	{
		switch (dam)
		{
		case 0xf8:
			si_flags |= SECTOR_damhigh;
			si_flags |= SECTOR_delete;
			break;
		case 0xf9:
			si_flags |= SECTOR_damhigh;
			si_flags &= ~SECTOR_delete;
			break;
		case 0xfa:
			si_flags &= ~SECTOR_damhigh;
			si_flags |= SECTOR_delete;
			break;
		case 0xfb:
			si_flags &= ~SECTOR_damhigh;
			si_flags &= ~SECTOR_delete;
			break;
		}
	}

	return get_dam(chip);
}

DiskDrive::DiskDrive(const DriveConfiguration& config, const DiskImageMetadata* metadata)
	: m_config(config)
	, m_chip(CHIP_WD1793)
	, m_file(NULL)
	, m_file_writeable(0)
	, writable(1)
	, current_track(0)
	, current_side(0)
	, tr00(1)
	, ready(0)
	, index_hole_blocked(0)
	, allow_track_append(0)
	, requested_side(0)
{
	make_crc_table();

	tr00 = 1;
	dirty = 0;
	writable = 1;

	// Configure the parameters and header setup based on the drive configuration.
	info.sides = config.sides;
	info.tracks = config.tracks;
	// Typical tracklen values:
	// 5.25" SD:  3125
	// 5.25" DD:  6520
	// 8"    DD: 10416
	int rpm = config.rpm;
	if (rpm == 0)
		rpm = 100;

	tracklen = config.data_rate / 8 * 60 / rpm;

	// No disk[] space allocated
	len = 0;
	len_max = 0;
	disk = NULL;
}

int DiskDrive::ensure_disk_size(int required_len)
{
	if (required_len <= len_max)
		return 1;

	if (len_max < 1024) {
		len_max = 1024;
	}

	// Don't go beyond 32MB disk image size.
	// Should be OK for even extreme floppy sizes and formats.
	while (len_max < required_len && len_max <= 32*1024*1024) {
		len_max *= 2;
	}

	if (len_max < required_len) {
		return 0;
	}

	unsigned char* new_data = new unsigned char[len_max];
	if (disk) {
		memcpy(new_data, disk, len);
		memset(new_data + len, 0x00, len_max - len);	// Initialize unused portion to 0x00
		delete [] disk;
	}

	disk = new_data;

	return 1;
}

const char* DiskDrive::FileError()
{
	if (m_file == NULL)
	{
		return "internal error: no file";
	}

	int err = ferror(m_file);
	if (err == 0)
	{
		return "internal error: no error on file";
	}

	// Get rid of the error in case we try again
	clearerr(m_file);

	return strerror(err);
}

int DiskDrive::GetFileWriteable()
{
	return m_file_writeable;
}

int DiskDrive::GetWriteProtect()
{
	return !writable;
}

void DiskDrive::SetWriteProtect(int on)
{
	writable = !on;
	dirty = 1;
}

int DiskDrive::GetReady()
{
	return ready;
}

void DiskDrive::SetReady(int val)
{
	ready = !!val;
}

void DiskDrive::CheckAllowTrackAppend()
{
	if (current_track == info.tracks - 1)
	{
		allow_track_append = 1;
	}
	else
	{
		allow_track_append = 0;
	}
}

char* DiskDrive::TrackDump(int side, int track)
{
	DMKTrack trk;

	GetDMKTrack(&trk, side, track);
	char* text = trk.Dump(track);

	return text;
}

void DiskDrive::TrackScan(int side, int track, ScanTrack* callback)
{
	// ... the 'fear' of the gap problem holds you back, but, mostly copy code
	// Also, this is the generic DMK version
}

// Output (or count) run-length encoding of buffer
// if f == NULL, returns number of runs, else returns # characters printed
static int run_out(void* ctx, void (*output)(void *ctx, const char* text), int run_char, int run, char marker);
static int run_length_dump(void* ctx, void (*output)(void *ctx, const char* text), unsigned char* data, int offset, int end, char marker, int delta)
{
	int run = 0;
	int run_char = -1;

	int result = 0;
	for (; offset < end; offset += delta)
	{
		if (run_char == data[offset])
		{
			run++;
		}
		else
		{
			result += run_out(ctx, output, run_char, run, marker);
			if (run > 0) marker = ' ';
			run = 1;
			run_char = data[offset];
		}
	}

	result += run_out(ctx, output, run_char, run, marker);

	return result;
}
static int run_out(void* ctx, void (*output)(void *ctx, const char* text), int run_char, int run, char marker)
{
	char f[256];

	if (run == 0 || run_char < 0) return 0;

	if (ctx != NULL)
	{
		int res;
		if (run > 1)
			res = sprintf(f, "%c%02x*%d", marker, run_char, run);
		else
			res = sprintf(f, "%c%02x", marker, run_char);

		output(ctx, f);

		return res;
	}
	
	return 1;
}


// Output printable version of data padding out from the current column
static void ascii_dump(void* ctx, void (*output)(void *ctx, const char* text), unsigned char *ascii_buf, int ascii_buf_len, int curr_column)
{
	char f[256];

	while (curr_column < 20*3 + 2)
	{
		sprintf(f, " "); output(ctx, f);
		curr_column++;
	}

	for (int j = 0; j < ascii_buf_len; j++)
	{
		int ch = ascii_buf[j];
		sprintf(f, "%c", ch >= 32 && ch < 128 ? ch : '.'); output(ctx, f);
	}
}

// Dump segment of data from offset to end - 1
// mark_len is the # of special marker bytes at the head (if any) to surround with {}
// crc_offset is the position of the 2-byte CRC (if any) to surround with []
// Data after crc_offset + 2 is the between marker gap data.
// If crc_offset is -1 the entire segment is gap data.
// delta is the stride between data bytes (1 or 2)
void dump_segment(void* ctx, void (*output)(void *ctx, const char* text), const char* name, int sector, unsigned char* data, int offset, int end, int delta, int mark_len, int crc_offset)
{
	char f[256];

	if (offset >= end)
	{
		return;
	}

	if (name == NULL)
	{
		sprintf(f, "%-7s ", " "); output(ctx, f);
	}
	else
	{
		sprintf(f, "%02x %-4s ", sector, name); output(ctx, f);
	}
	char marker = mark_len > 0 ? '{' : ' ';
	int len = 0;
	unsigned char ascii_buf[16];
	int ascii_buf_len = 0;
	int out_len = 0;

	for ( ; offset < end; offset += delta, len++)
	{
		if (crc_offset == -1)
		{
			break;
		}

		if (mark_len > 0 && mark_len == len)
		{
			marker = '}';
		}
		if (len == crc_offset)
		{
			marker = '[';
		}
		sprintf(f, "%c%02x", marker, data[offset]); output(ctx, f);
		out_len += 3;

		// Only non 'marked' characters are ASCII dumped. 
		if (len >= mark_len && len != crc_offset && len != crc_offset + 1)
		{
			ascii_buf[ascii_buf_len++] = data[offset];
			if (ascii_buf_len >= 16)
			{
				ascii_dump(ctx, output, ascii_buf, ascii_buf_len, out_len);

				sprintf(f, "\n%-7s ", " "); output(ctx, f);
				ascii_buf_len = 0;
				out_len = 0;
			}
		}

		marker = ' ';
		if (crc_offset >= 0 && len + 1 == crc_offset + 2) {
			marker = ']';
			offset += delta;
			len++;
			break;
		}
	}

	// Output any 'gap' (might be the whole segment)
	if (offset < end)
	{
		int num_run = run_length_dump(NULL, output, data, offset, end, marker, delta);
		// Condition here should be modified based on experience
		// If gaps look like they contain data then hexdump them.
		if (num_run > 8)
		{
			out_len += sprintf(f, "%c<%d bytes>", marker, end - offset); output(ctx, f);
			ascii_dump(ctx, output, ascii_buf, ascii_buf_len, out_len);
			// Hexdump the gap data
			ascii_buf_len = 0;
			out_len = 0;
			sprintf(f, "\n%-7s ", "gap"); output(ctx, f);
			for (; offset < end; offset += delta)
			{
				sprintf(f, " %02x", data[offset]); output(ctx, f);
				out_len += 3;
				ascii_buf[ascii_buf_len++] = data[offset];
				if (ascii_buf_len >= 16)
				{
					ascii_dump(ctx, output, ascii_buf, ascii_buf_len, out_len);

					sprintf(f, "\n%-7s ", " "); output(ctx, f);
					ascii_buf_len = 0;
					out_len = 0;
				}
			}
			if (ascii_buf_len != 0)
			{
				ascii_dump(ctx, output, ascii_buf, ascii_buf_len, out_len);
			}
			ascii_buf_len = 0;
		}
		else
		{
			out_len += run_length_dump(ctx, output, data, offset, end, marker, delta);
		}
		marker = ' ';
	}

	sprintf(f, "%c", marker); output(ctx, f);
	out_len++;

	if (ascii_buf_len != 0)
	{
		ascii_dump(ctx, output, ascii_buf, ascii_buf_len, out_len);
	}

	sprintf(f, "\n"); output(ctx, f);
}

void file_out(void *fp, const char* data)
{
	FILE *f = (FILE*)fp;

	fwrite(data, 1, strlen(data), f);
}

void char_size(void *ctx, const char* data)
{
	int extra = 0;
	for (int i = 0; data[i]; i++) {
		if (data[i] == '\n') {
			extra++;
		}
	}

	size_t *len = (size_t*)ctx;

	*len += strlen(data) + extra;
}

void char_out(void *ctx, const char* data)
{
	char **text = (char**)ctx;

	for (int i = 0; data[i]; i++) {
		if (data[i] == '\n') {
			*(*text)++ = '\r';
			*(*text)++ = data[i];
		}
		else {
			*(*text)++ = data[i];
		}
	}
}

// Required for HardDrive
void dump_segment(FILE *f, const char* name, int sector, unsigned char* data, int offset, int end, int delta, int mark_len, int crc_offset)
{
	dump_segment(f, file_out, name, sector, data, offset, end, delta, mark_len, crc_offset);
}

int DiskDrive::GetTrack()
{
	return current_track;
}

void DiskDrive::SetTrack(int track)
{
	if (track < 0) {
		track = 0;
	}

	// Appending new blank tracks only allowed if a write track
	// ocurred on the last track--e.g., we're formatting a drive.
	//BOUNDS: we still have to stop at some point
	if (track > info.tracks - 1 && allow_track_append) {
		//TODO: Geometry lock
		AppendBlankTrack();//RETURN VALUE
	}

	if (track > info.tracks - 1) {
		track = info.tracks - 1;
	}
	current_track = track;

	// TR00 pin: 0 means at track 0
	tr00 = (current_track != 0);
}

int DiskDrive::GetSide()
{
	return current_side;
}

void DiskDrive::SetSide(int side)
{
	if (side < 0) {
		side = 0;
	}

	requested_side = side;

	if (side >= info.sides) {
		side = info.sides - 1;
	}
	current_side = side;
}

int DiskDrive::GetNumSides()
{
	return info.sides;
}

void DiskDrive::StepTrack(int dir)
{
	int track = current_track + ((dir == 1) ? 1 : - 1);

	SetTrack(track);
}

int DiskDrive::AppendBlankTrack()
{
	//Default: appending is not possible.

	return 0;
}

// pin_TR00 state
int DiskDrive::GetTR00()
{
	return tr00;
}

int DiskDrive::GetOffset()
{
	return radial_position;
}

void DiskDrive::SetOffset(int offset)
{
	radial_position = offset % tracklen;
}

void DiskDrive::SetGeometryFlags(int flags)
{
	m_config.geometry_options = flags;
}

int DiskDrive::GetIndexHoleBlocked()
{
	return index_hole_blocked;
}

void DiskDrive::SetIndexHoleBlocked(int blocked)
{
	index_hole_blocked = blocked;
}

int DiskDrive::AtIndexHoleByte(int density)
{
	if (index_hole_blocked)
		return 0;

	// Must be logically equivalent to !IndexPulse() && IndexPulse(pos + stride[density])
	if (radial_position >= tracklen - stride[density]) {
		return 1;
	}

	return 0;
}

int DiskDrive::IndexPulse()
{
	if (len == 0) {//HACKHACK
		//No diskette
		return 0;
	}

	//HACKHACK
	// Does not account for single-density disks (or mixed density)

	// FD200 Disk Drive Operating and Service (1978)(Pertec)(pdf).pdf
	// says index pulse has duration 4.0 msec every 200 msec.

	// The diagram in WD1773.pdf implies the index pulse covers
	// 60 bytes of double-density '4E' -- implying 1.92 milliseconds.

	//ERROR definitely should use drive speed parameters here.
	// Physically pulse is about 2.8 milliseconds, or about 87.5 32-usec bytes.
	return radial_position < 2800 / FDC_MICROSECONDS;
}

// "track" programs
enum {
	TCMD_repeat,
	TCMD_repeat_clock,
	TCMD_id,
	TCMD_begin_crc,
	TCMD_crc,
	TCMD_DAM,
	TCMD_data,
	TCMD_density,
} ;

int track_start_dden[2*3 + 1] = {
	TCMD_density, 0, 0x1,
	TCMD_repeat, 60, 0x4e,
	-1
} ;

int track_sector_dden[15*3 + 1] = {
	TCMD_density, 0, 0x1,
	TCMD_repeat, 12, 0x00,
	TCMD_begin_crc, 0, 0x00,
	TCMD_repeat_clock, 3, 0x0aa1,
	TCMD_repeat, 1, 0xfe,
	TCMD_id, 4, 0x00,
	TCMD_crc, 2, 0x0000,
	TCMD_repeat, 22, 0x4e,
	TCMD_repeat, 12, 0x00,
	TCMD_begin_crc, 0, 0x00,
	TCMD_repeat_clock, 3, 0x0aa1,
	TCMD_DAM, 1, 0x00,
	TCMD_data, 0, 0x00,
	TCMD_crc, 2, 0x0000,
	TCMD_repeat, 24, 0x4e,
	-1,
} ;

int track_end_dden[3*2 + 1] = {
	TCMD_density, 0, 0x1,
	TCMD_repeat, 668, 0x4e,
	-1
} ;

// Single density
// idea is that cmd_offset is divided by two.
// also: need to set recording density command.
int track_start_sden[2*3 + 1] = {
	TCMD_density, 0, 0x0,
	TCMD_repeat, 40, 0xff,
	-1
} ;

int track_sector_sden[13*3 + 1] = {
	TCMD_density, 0, 0x0,
	TCMD_repeat, 6, 0x00,
	TCMD_begin_crc, 0, 0x00,
	TCMD_repeat_clock, 1, 0xc7fe,
	TCMD_id, 4, 0x00,
	TCMD_crc, 2, 0x0000,
	TCMD_repeat, 11, 0xff,
	TCMD_repeat, 6, 0x00,
	TCMD_begin_crc, 0, 0x00,
	TCMD_DAM, 1, 0x00,
	TCMD_data, 0, 0x00,
	TCMD_crc, 2, 0x0000,
	TCMD_repeat, 10, 0xff,
	-1,
} ;

int track_end_sden[2*3 + 1] = {
	TCMD_density, 0, 0x0,
	TCMD_repeat, 334, 0x4e,
	-1
} ;

int tcmd_copy(int *tcmd, int pc, int *prog, int limit)
{
	for (int i = 0; prog[i] >= 0 && pc + 3 <= limit; i += 3, pc += 3)
	{
		tcmd[pc] = prog[i];
		tcmd[pc + 1] = prog[i + 1];
		tcmd[pc + 2] = prog[i + 2];
	}

	return pc;
}

void BlockDrive::DeleteAllSectors()
{
	if (side_info != NULL)
	{
		for (int s = 0; s < info.sides; s++)
		{
			for (int i = 0; i < info.tracks; i++)
			{
				for (int j = 0; j < side_info[s].track_info[i].sectors; j++)
				{
					delete [] side_info[s].sector_data[i][j];
				}
				delete [] side_info[s].sector_info[i];
				delete [] side_info[s].sector_data[i];
			}
			delete [] side_info[s].track_info;
			delete [] side_info[s].sector_info;
			delete [] side_info[s].sector_data;
		}
		delete [] side_info;

		side_info = NULL;
	}
}

int BlockDrive::AppendBlankTrack()
{
	if (info.tracks >= 255)
	{
		return 0;
	}

	int trk = info.tracks;

	for (int s = 0; s < info.sides; s++)
	{
		TrackInfo* new_track_info = new TrackInfo[info.tracks + 1];
		SectorInfo** new_sector_info = new SectorInfo*[info.tracks + 1];
		unsigned char*** new_sector_data = new unsigned char**[info.tracks + 1];

		for (int k = 0; k < info.tracks; k++)
		{
			new_track_info[k] = side_info[s].track_info[k];
			new_sector_info[k] = side_info[s].sector_info[k];
			new_sector_data[k] = side_info[s].sector_data[k];
		}

		// Delete old arrays and replace
		delete [] side_info[s].track_info;
		delete [] side_info[s].sector_info;
		delete [] side_info[s].sector_data;

		side_info[s].track_info = new_track_info;
		side_info[s].sector_info = new_sector_info;
		side_info[s].sector_data = new_sector_data;

		// Initialize the new track
		side_info[s].track_info[trk].sectors = 0;
		side_info[s].sector_info[trk] = NULL;
		side_info[s].sector_data[trk] = NULL;
	}

	info.tracks++;

	return 1;
}

int BlockDrive::AppendBlankSide()
{
	if (info.sides > 1)
	{
		return 1;
	}

	info.sides = 2;
	SideInfo* new_sides = new SideInfo[info.sides];			// probably should just make it static...
	new_sides[0] = side_info[0];
	delete [] side_info;
	side_info = new_sides;

	// Fill in trackinfo for side 2
	int side = 1;
	side_info[side].track_info = new TrackInfo[info.tracks];
	side_info[side].sector_info = new SectorInfo*[info.tracks];
	side_info[side].sector_data = new unsigned char**[info.tracks];

	for (int trk = 0; trk < info.tracks; trk++)
	{
		side_info[side].track_info[trk].sectors = 0;
		side_info[side].sector_info[trk] = NULL;
		side_info[side].sector_data[trk] = NULL;
	}

	return 1;
}

BlockDrive::BlockDrive(const DriveConfiguration &config, const DiskImageMetadata* metadata)
	: DiskDrive(config, metadata)
{
	trkstate = NULL;

	// Block drives always use the same stride
	stride[0] = 2;
	stride[1] = 1;

	side_info = new SideInfo[info.sides];
	for (int s = 0; s < info.sides; s++)
	{
		side_info[s].track_info = new TrackInfo[info.tracks];
		side_info[s].sector_info = new SectorInfo*[info.tracks];
		side_info[s].sector_data = new unsigned char**[info.tracks];
		for (int i = 0; i < info.tracks; i++)
		{
			side_info[s].track_info[i].sectors = 0;
			side_info[s].sector_info[i] = NULL;
			side_info[s].sector_data[i] = NULL;
		}
	}

	current_track = 0;
	radial_position = 0;
	tc_track = 0;

	CreateTrackProgram();
}

BlockDrive::~BlockDrive()
{
	DeleteAllSectors();
}

// ----------------------------------------------------------------------

void TrackProgram::RunCompleteTrackToGetLength()
{
	int *tcmd = track_command;

	tracklen = 0;
    tc_density = 0;

	for (pc = 0; tcmd[pc] >= 0; pc += 3)
	{
        if (tcmd[pc] == TCMD_density) {
            tc_density = tcmd[pc + 2];
        }
		tracklen += tcmd[pc + 1] * stride[tc_density];
	}
}

void TrackProgram::BuildTrackProgram(int num_sector, SectorInfo* _sec_info, unsigned char** _sec_data, const DriveConfiguration& config, WD17xx_Family chip)
{
	int *tcmd = track_command;
	int tcmd_len = (int)(sizeof(track_command)/sizeof(track_command[0]));

	pc = 0;
	cmd_offset = 0;

	m_chip = chip;

	stride[0] = 2;
	stride[1] = 1;

	dirty = 0;

	sector_info = _sec_info;
	sector_data = _sec_data;

	// Build the track program (header, sectors, trailer)
	// and fill in the sector ids and length
	int density = 0;
	for (int sec = 0; sec < num_sector; sec++)
	{
		density = (sector_info[sec].si_flags & SECTOR_dden);
		if (sec == 0)
			pc = tcmd_copy(tcmd, pc, density ? track_start_dden : track_start_sden, tcmd_len);

		int old_pc = pc;
		pc = tcmd_copy(tcmd, pc, density ? track_sector_dden : track_sector_sden, tcmd_len);
		for (; old_pc < pc; old_pc += 3)
		{
			if (tcmd[old_pc] == TCMD_data || tcmd[old_pc] == TCMD_id || tcmd[old_pc] == TCMD_DAM)
			{
				if (tcmd[old_pc] == TCMD_data)
				{
					// Data sections need their length
					tcmd[old_pc + 1] += sector_info[sec].si_length;
				}
				tcmd[old_pc + 2] = sec;
			}
		}
	}
	pc = tcmd_copy(tcmd, pc, density ? track_end_dden : track_end_sden, tcmd_len);
	tcmd[pc] = -1;

	// Compute the track length
	RunCompleteTrackToGetLength();

	// Extend length of track program to match the physical track length
	int ideal_tracklen = config.data_rate / 8 * 60 / config.rpm;

	if (config.density == DISK_FM)
	{
		// Double the target track length since BlockDrive FM programs are double-byte
		ideal_tracklen *= 2;
	}

	if (tracklen < ideal_tracklen)
	{
		tcmd[pc - 2] += (ideal_tracklen - tracklen) / stride[tc_density];

		RunCompleteTrackToGetLength();
	}
	else if (tracklen > ideal_tracklen)
	{
		// Part of the solution to M2 LS-DOS format and backup problems is to
		// truncate excessive track lengths--as long as the repeat count isn't too low.
		int remain = tcmd[pc - 2] + (ideal_tracklen - tracklen) / stride[tc_density];
		if (remain < 10)
		{
			remain = 10;
		}

		tcmd[pc - 2] = remain;

		RunCompleteTrackToGetLength();
	}
	// TODO: if the track isn't the ideal length then:
	// - maybe the drive config is wrong
	// - maybe the source data has too many sectors
	//   - or we're padding too much between sectors

	// Compute CRC values
	radial_position = 0;
	pc = 0;
	tc_in_crc = 0;
	tc_crc = 0;
	for (pc = 0; tcmd[pc] >= 0;)
	{
		// Compute CRC fields
		if (tcmd[pc] == TCMD_begin_crc)
		{
			tc_in_crc = 1;
			tc_crc = 0xffff;
			pc += 3;
		}
		else if (tcmd[pc] == TCMD_density)
		{
			tc_density = tcmd[pc + 2];
			pc += 3;
		}
		else if (tc_in_crc)
		{
			if (tcmd[pc] == TCMD_crc)
			{
				// write crc
				tcmd[pc + 2] = tc_crc;
				tc_in_crc = 0;
			}
			else
			{
				int clock, data;
				data = GetByte(tc_density, &clock);
				tc_crc = update_crc(tc_crc, data);
				AdvanceByte(tc_density);
			}
		}
		else
		{
			radial_position += tcmd[pc + 1] * stride[tc_density];
			pc += 3;
			cmd_offset = 0;
		}
	}

	SetRadialPosition(0);
}

void TrackProgram::SetRadialPosition(int pos)
{
	radial_position = pos;
	//ERROR Ideally we should do a angle conversion, instead, just truncate
	assert(tracklen != 0);
	if (radial_position >= tracklen)
	{
		radial_position %= tracklen;
	}

	// Compute PC for correct radial position
	// Note: tc_in_crc is not computed since it is only
	// used by writing and writing will pass-through the begin_crc, always
	tc_density = 0;
	int tpos = 0;
	int *tcmd = track_command;
	for (pc = 0; tcmd[pc] >= 0; pc += 3)
	{
		if (tcmd[pc] == TCMD_density)
		{
			tc_density = tcmd[pc + 2];
		}

		int len = tcmd[pc + 1] * stride[tc_density];
		if (radial_position < tpos + len)
		{
			cmd_offset = radial_position - tpos;
			break;
		}
		tpos += len;
	}
	assert(tcmd[pc] >= 0);
}

int TrackProgram::GetByte(int , int* clock)
{
	ProcessNonDataCommands();

	int res = 0;
	*clock = 0xff;
	int data = track_command[pc + 2];
    int offset = cmd_offset / stride[tc_density];
	switch (track_command[pc])
	{
	case TCMD_repeat:
		res = data;
		break;
	case TCMD_repeat_clock:
		res = (data & 0xff);
		*clock = ((data >> 8) & 0xff);
		break;
	case TCMD_id:
		switch (offset) {
		case 0:
			res = sector_info[data].si_track;//track
			break;
		case 1:
			res = sector_info[data].si_side;//side
			break;
		case 2:
			res = sector_info[data].si_sector;//sector
			break;
		case 3:
			{
				int code = 4;	// unknown
				switch (sector_info[data].si_length)
				{
				case 128: code = 0; break;
				case 256: code = 1; break;
				case 512: code = 2; break;
				case 1024: code = 3; break;
				default:
					code = 4; // unknown
				}
				res = code;//length
			}
			break;
		}
		break;
	case TCMD_density:
	case TCMD_begin_crc:
		assert(0);
		break;
	case TCMD_crc:
		tc_in_crc = 0;
		res = data;
		if (offset == 0)
		{
			res >>= 8;
		}
		res &= 0xff;
		break;

	case TCMD_DAM:
		res = sector_info[data].get_dam(m_chip);
		if (!tc_density) {
			*clock = 0xc7;
		}
		break;
	case TCMD_data:
		res = sector_data[data][offset]; // data
		break;
	}

	return res;
}

void TrackProgram::WriteByte(int density, int val, int clock)
{
	ProcessNonDataCommands();

	int data = track_command[pc + 2];
    int offset = cmd_offset / stride[tc_density];

	// Change the DAM value written to match the program so CRC calculations are correct.
	if (track_command[pc] == TCMD_DAM)
	{
		int data = track_command[pc + 2];
		int dam = sector_info[data].get_dam(m_chip);
		if (dam != val)
		{
			// Modify the DAM to a supported value.
			val = sector_info[data].set_dam(m_chip, val);
		}
	}

	if (track_command[pc] != TCMD_crc && tc_in_crc)
	{
		tc_crc = update_crc(tc_crc, val);
	}

	switch (track_command[pc])
	{
	case TCMD_density:
	case TCMD_begin_crc:
		assert(0);
		//... actually begin_crc can help with testing.
		//	add an 'in crc' flag.
		break;
	case TCMD_crc:
		// Possibly we can get a bad CRC.
		tc_in_crc = 0;
		{
			int crc = track_command[pc + 2];
			if (offset == 0)
			{
				track_command[pc + 2] = ((crc & 0xff) | (val << 8));
			}
			else
			{
				track_command[pc + 2] = ((crc & 0xff00) | val);
				// TODO: if the controller CRC doesn't match the computed one,
				// we should set the CRC error flag.  However, if it doesn't match,
				// there is more likely a problem with track sync'ing for WriteByte.
				assert( track_command[pc + 2] == tc_crc);
			}
		}
		break;

	case TCMD_data:
		sector_data[data][offset] = val;
		dirty = 1;
		break;
	}
}

// Process commands which don't generate data bytes
void TrackProgram::ProcessNonDataCommands()
{
	while (track_command[pc] >= 0 && track_command[pc + 1] == 0)
	{
		if (track_command[pc] == TCMD_density)
		{
			tc_density = track_command[pc + 2];
		}
		else if (track_command[pc] == TCMD_begin_crc)
		{
			tc_crc = 0xffff;
			tc_in_crc = 1;
		}
		pc += 3;
	}
}

void TrackProgram::AdvanceByte(int density)
{
	ProcessNonDataCommands();

    int mul = stride[density];
	radial_position += mul;
	cmd_offset += mul;
	int cmd_len = track_command[pc + 1] * stride[tc_density];
	if (cmd_offset >= cmd_len)
	{
		if (track_command[pc] == TCMD_begin_crc || track_command[pc] == TCMD_density || track_command[pc + 1] == 0)
		{
			printf("uh-oh");
		}
		pc += 3;
		cmd_offset -= cmd_len;

		if (track_command[pc] < 0)
		{
			pc = 0;
			radial_position = 0;
		}
	}

	assert(radial_position < tracklen);
}

void BlockDrive::CreateTrackProgram()
{
	trk_program.BuildTrackProgram(side_info[current_side].track_info[tc_track].sectors, side_info[current_side].sector_info[tc_track], side_info[current_side].sector_data[tc_track], m_config, m_chip);
	trk_program.SetRadialPosition(radial_position);
	radial_position = trk_program.radial_position;
	tracklen = trk_program.tracklen;

	// Track program uses the same stride
	stride[0] = 2;
	stride[1] = 1;
}

int BlockDrive::GetByte(int , int* clock)
{
	int res = trk_program.GetByte(0, clock);
	radial_position = trk_program.radial_position;

	return res;
}

struct WriteTrackState {
	DMKTrack wt_track;
} ;

void BlockDrive::WriteByte(int density, int val, int clock)
{
	if (trkstate)
	{
		trkstate->wt_track.WriteByte(density, val, clock);
		return;
	}

	trk_program.WriteByte(density, val, clock);
	if (trk_program.dirty) {
		dirty = 1;
	}
	radial_position = trk_program.radial_position;
}

void BlockDrive::NotifyWriteTrack(int writeTrackOn)
{
	// The block drive could guess that write track is occurring
	// by looking at the position of the writes. Adding an explicit
	// flag makes it a bit safer to change the code.
	if (!writeTrackOn)
	{
		// Replace the current track with the formatted one.
		if (trkstate != NULL)
		{
			TrackInfo tinfo;

			trkstate->wt_track.GetTrackInfo(tinfo);
			SectorInfo* sinfo = new SectorInfo[tinfo.sectors];
			unsigned char ** tdata = new unsigned char*[tinfo.sectors];

			for (int i = 0; i < tinfo.sectors; i++)
			{
				trkstate->wt_track.GetSectorInfo(sinfo[i], i);

				//TODO: a JV3-specific hack here is to force sinfo[i].si_side to be 0 or 1 based on the current side.
				//But can IMD (or any other 'block' format) deal with the SIDE IDAM value being different from the logical side???
#ifdef JV3_SIDE_FIX
				if (ImageFormat() == DISK_JV3) {
					sinfo[i].si_side = (current_side == 0) ? 0 : 1;
				}
#endif

				tdata[i] = new unsigned char[sinfo[i].si_length];
				trkstate->wt_track.GetSectorData(tdata[i], i);
			}

			dirty = 1;
			// Update info
			//TODO: may need to change tracks/sides DiskInfo info;

			for (int i = 0; i < side_info[current_side].track_info[current_track].sectors; i++)
			{
				delete [] side_info[current_side].sector_data[current_track][i];
			}
			delete [] side_info[current_side].sector_data[current_track];
			side_info[current_side].sector_data[current_track] = tdata;

			delete [] side_info[current_side].sector_info[current_track];
			side_info[current_side].sector_info[current_track] = sinfo;

			side_info[current_side].track_info[current_track] = tinfo;

			//ERROR JV3 or others might need additional data structure updates
			delete trkstate;
			trkstate = NULL;

			CreateTrackProgram();
		}
	}
	else
	{
		if (trkstate == NULL)
		{
			trkstate = new WriteTrackState;

			GetDMKTrack(&trkstate->wt_track, current_side, current_track);

			// Change tracklen (and stride) to match DMK track length
			tracklen = trkstate->wt_track.GetTracklenAndStride(stride);

			CheckAllowTrackAppend();

			if (requested_side == 1 && info.sides == 1)
			{
				if (!(m_config.geometry_options & DRV_side_lock))
				{
					AppendBlankSide();
				}
			}
		}
	}
}

void BlockDrive::AdvanceByte(int density)
{
	if (trkstate)
	{
		trkstate->wt_track.AdvanceByte(density);
		//ERROR tracklen has to be identical otherwise we'll miss the index pulse
		radial_position = trkstate->wt_track.GetPosition();
		return;
	}

	trk_program.AdvanceByte(density);
	radial_position = trk_program.radial_position;
}

int BlockDrive::IsModified()
{
	return dirty;
}

void BlockDrive::GetDiskInfo(DiskInfo& res)
{
    res = info;
}

void BlockDrive::GetTrackInfo(TrackInfo& ret_info, int side, int track)
{
	if (side >= 0 && side < info.sides && track >= 0 && track < info.tracks) //BOUNDS
	{
		ret_info = side_info[side].track_info[track];
	}
	else
	{
		ret_info.sectors = 0;
	}
}

void BlockDrive::GetSectorInfo(SectorInfo& ret_info, int side, int track, int sector)
{
	if (side >= 0 && side < info.sides && track >= 0 && track < info.tracks) //BOUNDS
	{
		ret_info = side_info[side].sector_info[track][sector];
	}
	else
	{
		memset(&ret_info, 0, sizeof(ret_info));
	}
}

void BlockDrive::GetSectorData(unsigned char* res, int side, int track, int sector)
{
 	if (side >= 0 && side < info.sides && track >= 0 && track < info.tracks) //BOUNDS
	{
		unsigned char *data = side_info[side].sector_data[track][sector];

		memcpy(res, data, side_info[side].sector_info[track][sector].si_length);
	}
}

void BlockDrive::GetDMKTrack(class DMKTrack *trk, int side, int track)
{
	TrackProgram tprog;
	tprog.BuildTrackProgram(side_info[side].track_info[track].sectors, side_info[side].sector_info[track], side_info[side].sector_data[track], m_config, m_chip);

	int a1_run = 0;
	int fe_run = 0;
	int curr_idam = 0;
	int track_len = 0;
	unsigned char raw[16384];
	int idam[DMK_MAX_IDAM];

	for (int j = 0; j < DMK_MAX_IDAM; j++)
	{
		idam[j] = 0;
	}

	// Use GetByte() to raw data, finding IDAM marks/density changes along the way
	// The code here (like TrackProgram itself) presumes double-density
	int density = 1;
	for (;;)
	{
		int data, clock;
		data = tprog.GetByte(density, &clock); // density is ignored.
		raw[track_len] = data;
		if (data == 0xfe)
		{
			if (a1_run == 3)
			{
				// double density IDAM
				idam[curr_idam++] = ((track_len & 0x3fff) | 0x8000);
			}
			else if (clock == 0xc7)
			{
				// single density IDAM
				if (fe_run == 1)	// trigger on second marker
				{
					idam[curr_idam++] = ((track_len - 1) & 0x3fff);	// mark at the previous location
				}
			}
		}
		track_len++;

		// Update run counter for DDEN A1 marker bytes
		if (data == 0xa1 && clock == 0x0a)
		{
			a1_run++;
		}
		else
		{
			a1_run = 0;
		}
		// Update run counter for SDEN marker
		if (data == 0xfe && clock == 0xc7)
		{
			fe_run++;
		}
		else
		{
			fe_run = 0;
		}

		tprog.AdvanceByte(density);
		if (tprog.radial_position == 0)
		{
			break;
		}
	}
	trk->SetData(raw, track_len, 0);
	trk->SetIdam(idam);

	// Pad track to actual size otherwise format may be truncated
	int geometry_tracklen = m_config.data_rate / 8 * 60 / m_config.rpm;

	if (m_config.density == DISK_FM)
	{
		// Double the target track length since BlockDrive FM programs are double-byte
		geometry_tracklen *= 2;
	}

	trk->SetMinimumTracklen(geometry_tracklen);
	trk->SetDataRate(m_config.data_rate);
}

void BlockDrive::PutDMKTrack(class DMKTrack *dmktrk, int side, int track)
{
	current_side = side;
	current_track = track;

	// TODO: Copied from NotifyWrite, should merge
	TrackInfo tinfo;

	dmktrk->GetTrackInfo(tinfo);
	SectorInfo* sinfo = new SectorInfo[tinfo.sectors];
	unsigned char ** tdata = new unsigned char*[tinfo.sectors];

	for (int i = 0; i < tinfo.sectors; i++)
	{
		dmktrk->GetSectorInfo(sinfo[i], i);

		//TODO: a JV3-specific hack here is to force sinfo[i].si_side to be 0 or 1 based on the current side.
		//But can IMD (or any other 'block' format) deal with the SIDE IDAM value being different from the logical side???
#ifdef JV3_SIDE_FIX
		if (ImageFormat() == DISK_JV3) {
			sinfo[i].si_side = (current_side == 0) ? 0 : 1;
		}
#endif

		tdata[i] = new unsigned char[sinfo[i].si_length];
		dmktrk->GetSectorData(tdata[i], i);
	}

	dirty = 1;
	// Update info
	//TODO: may need to change tracks/sides DiskInfo info;

	for (int i = 0; i < side_info[current_side].track_info[current_track].sectors; i++)
	{
		delete [] side_info[current_side].sector_data[current_track][i];
	}
	delete [] side_info[current_side].sector_data[current_track];
	side_info[current_side].sector_data[current_track] = tdata;

	delete [] side_info[current_side].sector_info[current_track];
	side_info[current_side].sector_info[current_track] = sinfo;

	side_info[current_side].track_info[current_track] = tinfo;

	//ERROR JV3 or others might need additional data structure updates
	delete trkstate;
	trkstate = NULL;

	CreateTrackProgram();
}

void BlockDrive::SetTrack(int track)
{
	// TODO: range checking.
	DiskDrive::SetTrack(track);

	if (current_track != tc_track)
	{
		tc_track = current_track;
		CreateTrackProgram();
	}
}

void BlockDrive::SetSide(int side)
{
	int old_side = current_side;

	DiskDrive::SetSide(side);

	if (old_side != current_side)
	{
		tc_track = current_track;
		CreateTrackProgram();
	}
}

int BlockDrive::GetRawByte(int offset)
{
	return 0;
}

const char* DiskDrive::LoadFromFile(FILE* f, unsigned char* data, int count, int writeable)
{
	if (!ensure_disk_size(count))
	{
		return "Disk image too large";
	}

	m_file = f;
	m_file_writeable = writeable;

	Load(data, count);//ERROR: possible assert()s not captured

	return NULL;
}

const char* DiskDrive::SaveToFile()
{
	if (m_file == NULL)
	{
		return "err_no_path";
	}

	if (!m_file_writeable)
	{
		return "err_readonly";
	}

	if (fseek(m_file, 0, SEEK_SET) == EOF) return FileError();

	const char* res = Save(m_file);

	if (fflush(m_file) == EOF) return FileError();
#ifdef _MSC_VER
    //ERROR OSX fix
	if (_chsize( _fileno(m_file), ftell(m_file)) == EOF) return FileError();
#endif

	return res;
}

const char *DiskDrive::SaveAs(const char *filename)
{
	FILE *fp = fopen(filename, "wb");
	if (!fp) return FileError();

	const char *res = Save(fp);

	if (fclose(fp) == EOF) return FileError();

	return res;
}

const char *DiskDrive::Close()
{
	// No file to close, then all is fine.
	if (!m_file)
		return NULL;

	const char *res = NULL;
	if (fclose(m_file) == EOF)
		res = FileError();

	m_file = NULL;

	return res;
}

DiskDrive::~DiskDrive()
{
	if (m_file)
	{
		//TODO: caller should be forced to close before delete to retrieve the error
		assert( fclose(m_file) != EOF );//ERROR
	}
	m_file = NULL;

	delete[] disk;
}

/*
Sketch:
sector_info structure which provides a common "get_physical_sector_info(int side, int track)" interface.
(corresponding: get_physical_sector_data(int side, int track, byte* buffer, int &len))
I think one needs:
sector #, length, "error" flag, DAM marker type, density, encoding.
	density/encoding kind of go together.
GetByte() would be implemented on top of it.
DMK would overrite GetByte() since it does it 'naturally'.
(but: would ideally _implement_ the block interface for utilities).

- really need:
  driveinfo - RPM, tracks.  Physical information about the drive.
    - the disk data files don't really determine this.

  diskinfo - just # tracks, 2-sided or not.
  trackinfo - # sectors in the track, at least.
  sectorinfo
		FM/FMF track,side,sector,length (header) + DAM & CRC error indication
		DAM has 4 possibilities, I think, deleted then two different markers.
  Note that access has to be by physical track, sector # (the actual order)
  The values in the headers will commonly vary, at least for sector numbers
  there is 0 vs. 1-indexed and will typically be interleaved.
  The track and side can also be 'wrong' compared to the desired addressing.
*/

// recognizing formats.
// DMK bytes 03 02 are track length--can sanity check. (must be 0x3fff or less)
// and...bytes 0c-0f should be zero which is handy.
// JV1 - I'd go with the file size being multiple of 256*10
// JV3 - has these funny headers.
// triples of track / sector / flags.
// Detect 0x00 0xfe as JV1 - looking for model I 'boot' trick
// For JV3, check last byte of header is 0x00 or 0xff

// IMD has a variable length header.  Check for 'sensible'
// ascii followed by 1A in first 512 or so.

// Either loads disk image from file called "fname" or if fname is NULL
// then expects disk data passed in data + len arguments.

// Checks order/types checked based on extension
// .imd - IMD only
// .dsk .dmk .jv1 .jv3 - either of those, but choose jv3/dmk first based on extension (no IMD)


static int Detect_DMK(unsigned char *data, int len)
{
	if (len < 16)
		return 0;

	int tracklen = data[2] + (data[3]<<8);
	if (tracklen >= 256 && tracklen <= 0x3fff) {
		// Insist the file size >= 16+tracks*sides*tracklen (or close to it)

		int tracks = data[1];
		int sides = !!(data[4] & 0x10) ? 1 : 2;
		int missing = len - (16 + tracks*sides*tracklen);
		// Allow up to 8 tracks to be missing at the tail of the image
		// or: allow up to 16 missing if you claim more than 30 tracks.
		// The second rule is for "Lifeboat CP-M v1.4 Disk System Utilities (1981)(Author Unknown)[M1][Bad Sectors].dmk"
		// which has 15 missing tracks (but claims to have 43)
		if ((tracks >= 2 && missing >= -8*tracklen)
			|| (tracks >= 30 && missing >= -16*tracklen)
			)
		{
			if (data[0x0c] == 0 && data[0x0d] == 0 && data[0x0e] == 0 && data[0x0f] == 0) {
				return 1;
			}
		}
	}

	return 0;
}

static int Detect_JV1(unsigned char *data, int len)
{
	if (len < 2)
		return 0;

	if (data[0x00] == 0x00 && data[0x01] == 0xfe) {
		return 1;
	}

	return 0;
}

static int Detect_JV3(unsigned char *data, int len)
{
	if (len < 0x2200)
		return 0;

	// JV3--could also detect that sector count is consistent.
	// 0x21ff = JV3_RECORD_COUNT*3
	if (data[0x21ff] == 0x00 || data[0x21ff] == 0xff) {
		// Exclude file it it starts with a typical Model I boot block (so likely JV1)
		if (len % 256 == 0 && data[0] == 0x00 && data[1] == 0xfe && data[2] == 0x11)
		{
			return 0;
		}
		return 1;
	}

	return 0;
}

static int Detect_IMD(unsigned char *data, int len)
{
	// Probably don't need this
	//if (len < 512)
	//	return 0;

	// Try IMD format
    //TODO: there really is a header line:
    //IMD v.vv: dd/mm/yyyy hh:mm:ss

	// At least check for 'IMD' and some kind of version
	if (data[0] == 'I' && data[1] == 'M' && data[2] == 'D' &&
		data[3] == ' ' && isdigit(data[4]) && data[5] == '.' && isdigit(data[6]) && isdigit(data[7])
		)
	{
		return 2;	// "strong" detection
	}

	// Allow as IMD if it is ASCII at the start with a 0x1a terminator
	int OK = 0;
	for (int i = 0; i < 512 && i < len; i++)
	{
		if (data[i] < 9 || data[i] >= 0x80)
		{
			// unlikely to be valid ascii
			OK = 0;
			break;
		}
		if (data[i] == 0x1a)
		{
			// found end.
			OK = 1;	// "weak" detection
			break;
		}
	}

	return OK;
}

static int Detect_HFE(unsigned char *data, int len)
{
	if (len < HFE_BLOCK_SIZE)
		return 0;

	if (memcmp(&data[0], HFE_MAGIC, strlen(HFE_MAGIC)) == 0) {
		return 1;
	}

	if (memcmp(&data[0], HFEv3_MAGIC, strlen(HFEv3_MAGIC)) == 0) {
		return -1;	// Unsupported
	}

	return 0;
}

int HasExtension(const char* name, const char* ext)
{
	const char *p = name ? strrchr(name, '.') : 0;

	// Skip detection from ROMs
	if (p && stricmp(p, ext) == 0)
	{
		return 1;
	}

	return 0;
}


// Format name has optional double-sided and track number qualifiers
// FMT[-ds][#tracks]
int DiskFormatName(const char* name, int *sidesp, int* tracksp)
{
	int res = -1;
	int sides = 1;
	int tracks = 0; // meaning: use drive config

	if (_strnicmp(name, "dmk", 3) == 0) res = DISK_DMK;
	if (_strnicmp(name, "jv1", 3) == 0) res = DISK_JV1;
	if (_strnicmp(name, "jv3", 3) == 0) res = DISK_JV3;
	if (_strnicmp(name, "imd", 3) == 0) res = DISK_IMD;
	if (_strnicmp(name, "hfe", 3) == 0) res = DISK_HFE;

	if (res >= 0)
	{
		name += 3;
		if (*name == '-' && _strnicmp(name, "-ds", 3) == 0)
		{
			sides = 2;
			name += 3;
		}
		if (*name == '#' && isdigit(name[1]))
		{
			name++;
			tracks = atoi(name);
			name += strspn(name, "0123456789");
		}
	}

	// Don't allow extra characters (e.g. dmkfoo.dsk is a file name, not a blank)
	if (*name != 0) {
		res = -1;
	}

	if (sidesp) *sidesp = sides;
	if (tracksp) *tracksp = tracks;

	return res;
}

DiskDrive* LoadBlankDisk(const char* format, const char*& error, const DriveConfiguration& _config, const DiskImageMetadata *metadata)
{
	DiskDrive *res = NULL;

	DriveConfiguration config = _config;

	int sides, tracks;
	int fmt = DiskFormatName(format, &sides, &tracks);

	if (sides > config.sides)
		config.sides = sides;

	if (tracks > config.tracks)
		config.tracks = tracks;

	switch (fmt)
	{
		default: break;
#ifndef DMK2_TESTING
		case DISK_DMK: res = fd_create_dmk(config, metadata); break;
#else
		case DISK_DMK: res = fd_create_dmk2(config, metadata); break;
#endif
		case DISK_JV1: res = fd_create_jv1(config, metadata); break;
		case DISK_JV3: res = fd_create_jv3(config, metadata); break;
		case DISK_IMD: res = fd_create_imd(config, metadata); break;
		case DISK_HFE: res = fd_create_hfe(config, metadata); break;
	}

	if (res == NULL)
	{
		static char error_msg[256];
		sprintf(error_msg, "Unable to create blank %d-track %s-sided <%s> image",
			config.tracks,
			config.sides == 1 ? "one" : "two",
			format);

		error = error_msg;
	}

	return res;
}

DiskDrive* LoadDisk(const char* fname, const unsigned char *data_in, int len, const char*& error, const DriveConfiguration& config)
{
	unsigned char *data = (unsigned char *)data_in;

	error = NULL;

	FILE *f = NULL;
	int can_write = 0;

	if (fname) {
		can_write = 1;
		f = fopen(fname, "rb+");	
		if (f == NULL) {
			can_write = 0;

			f = fopen(fname, "rb");
			if (f == NULL)
			{
				error = strerror(errno);

				return NULL;
			}
		}

		if (fseek(f, 0, SEEK_END) == EOF)
		{
			error = strerror(errno);

			return NULL;
		}

		len = ftell(f);
		data = new unsigned char[len];
		if (data == NULL)
		{
			error = "out of memory";

			return NULL;
		}

		if (fseek(f, 0, SEEK_SET) == EOF)
		{
			error = strerror(errno);

			fclose(f);
			delete [] data;

			return NULL;
		}
		int readlen = (int)fread(data, 1, len, f);
		if (readlen == 0 || readlen != len)
		{
			error = readlen != len ? "fread incomplete" : strerror(ferror(f));

			fclose(f);
			delete [] data;

			return NULL;
		}
	}

	DiskImageFormat format_guess = DISK_DMK;

	if (fname != NULL)
	{
		if (HasExtension(fname, ".dmk")) { format_guess = DISK_DMK; }
		if (HasExtension(fname, ".jv1")) { format_guess = DISK_JV1; }
		if (HasExtension(fname, ".jv3")) { format_guess = DISK_JV3; }
		if (HasExtension(fname, ".imd")) { format_guess = DISK_IMD; }
		if (Detect_IMD(data, len) > 1) { format_guess = DISK_IMD; }	// Only if IMD and version found
		if (HasExtension(fname, ".hfe")) { format_guess = DISK_HFE; }
		if (Detect_HFE(data, len) < 0) { error = "HFE Version 3 not supported"; return NULL; }
		if (Detect_HFE(data, len)) { format_guess = DISK_HFE; }
	}

	int order[4] = { format_guess, -1, -1, -1 };

	switch (format_guess)
	{
		case DISK_DMK: order[1] = DISK_IMD; order[2] = DISK_JV3; order[3] = DISK_JV1; break;
		case DISK_JV1: order[1] = DISK_JV1; order[2] = DISK_JV3; order[3] = DISK_DMK; break;
		case DISK_JV3: order[1] = DISK_JV3; order[2] = DISK_JV1; order[3] = DISK_DMK; break;
		case DISK_IMD: break;
		case DISK_HFE: break;
		case DISK_empty: break;
	}
	DiskDrive *res = NULL;
	for (int i = 0; res == NULL && i < 4 && order[i] >= 0; i++)
	{
		switch ((DiskImageFormat)order[i])
		{
#ifndef DMK2_TESTING
		case DISK_DMK: if (Detect_DMK(data, len)) { res = fd_create_dmk(config, NULL); } break;
#else
		case DISK_DMK: if (Detect_DMK(data, len)) { res = fd_create_dmk2(config, NULL); } break;
#endif
		case DISK_JV1: if (Detect_JV1(data, len)) { res = fd_create_jv1(config, NULL); } break;
		case DISK_JV3: if (Detect_JV3(data, len)) { res = fd_create_jv3(config, NULL); } break;
		case DISK_IMD: if (Detect_IMD(data, len)) { res = fd_create_imd(config, NULL); } break;
		case DISK_HFE: if (Detect_HFE(data, len)) { res = fd_create_hfe(config, NULL); } break;

		case DISK_empty: break;
		}
	}
	if (res == NULL)
	{
		error = "disk image format not recognized";
	}
	else
	{
		error = res->LoadFromFile(f, data, len, can_write);
		if (error != NULL)
		{
			delete res;
			res = NULL;
		}
	}

	if (data != data_in)
		delete[] data;

	if (res == NULL && f != NULL)
	{
		fclose(f);
	}

	return res;
}

void SaveDiskASCII(FILE *f, DiskDrive* dsk, const char* options)
{
	bool trackdump = false;
	DiskInfo info;
	dsk->GetDiskInfo(info);


	fprintf(f, "DISK sides=%d tracks=%02x%s\n", info.sides, info.tracks, dsk->GetWriteProtect() ? " WriteProtect" : "");

	if (options != NULL && strchr(options, 'T') != NULL)
	{
		trackdump = true;
	}

	for (int side = 0; side < info.sides; side++)
	{
		if (info.sides > 1)
		{
			fprintf(f, "\nSIDE %d\n", side);
		}

		for (int track = 0; track < info.tracks; track++)
		{
			TrackInfo tinfo;
			dsk->GetTrackInfo(tinfo, side, track);

			if (trackdump)
			{
				if (dsk->ImageFormat() == DISK_HFE)
				{
					dsk->TrackDump(f, side, track);
				}
				else
				{
					DMKTrack trk;

					fprintf(f, "\n");
					dsk->GetDMKTrack(&trk, side, track);
					trk.Dump(f, track);
				}

				continue;
			}

			for (int sec_index = 0; sec_index < tinfo.sectors; sec_index++)
			{
				SectorInfo sinfo;
				dsk->GetSectorInfo(sinfo, side, track, sec_index);
				fprintf(f, "\n\n");
				if (info.sides > 1)
				{
					fprintf(f, "%d:", side);
				}
				
				fprintf(f, "%02x:%02x track %02x sector %02x side %d %s %d KB/s%s\n", track, sec_index, sinfo.si_track, sinfo.si_sector, sinfo.si_side, sinfo.si_flags & SECTOR_dden ? "MFM" : "FM", sinfo.get_rate(), sinfo.si_flags & SECTOR_unavail ? " [no data]" : "");
				if (!(sinfo.si_flags & SECTOR_unavail))
				{
					unsigned char sdata[16384];
					dsk->GetSectorData(sdata, side, track, sec_index);
					for (int i = 0; i < sinfo.si_length; i += 16)
					{
						for (int j = 0; j < 16; j++)
						{
							fprintf(f, "%02x ", sdata[i + j]);
						}
						fprintf(f, "   ");
						for (int j = 0; j < 16; j++)
						{
							int ch = sdata[i + j];
							fprintf(f, "%c", ch >= 32 && ch < 128 ? ch : '.');
						}
						fprintf(f, "\n");
					}
					if (sinfo.si_flags & SECTOR_err)
					{
						fprintf(f, "[crc error]\n");
					}
				}
			}
		}
	}
}

Notifier::Notifier()
{
	for (int i = 0; i < max_notify_level; i++)
	{
		m_note_count[i] = 0;
		m_notifications[i] = NULL;
	}
}

Notifier::~Notifier()
{
	Reset();
	for (int i = 0; i < max_notify_level; i++)
	{
		delete [] m_notifications[i];
	}
}

#include <stdarg.h>

void Notifier::Notify(Level lev, const char* fmt, ...)
{
	char buf[4096];
	va_list args;

	va_start(args, fmt);
#ifdef _WIN32
	vsprintf_s(buf, sizeof(buf), fmt, args);
#else
	vsprintf(buf, fmt, args);
#endif

	int len = strlen(buf) + 1;

	if (m_notifications[lev] == NULL)
	{
		m_notifications[lev] = new char[len];
		strcpy(m_notifications[lev], buf);
	}
	else
	{
		char* replace = new char[strlen(m_notifications[lev]) + len + 1];
		strcpy(replace, m_notifications[lev]);
		strcat(replace, "\n");
		strcat(replace, buf);
		delete [] m_notifications[lev];
		m_notifications[lev] = replace;
	}

	m_note_count[lev]++;
}

void Notifier::Reset()
{
	for (int i = 0; i < max_notify_level; i++)
	{
		m_note_count[i] = 0;
		delete [] m_notifications[i];
		m_notifications[i] = NULL;
	}
}

// switch on TRACK_PROGRAM, remove old code?
// Need to test this on 'sledge' etc.
// why does len == 0 cause NR and TR00 to flap?
// - RDY might be about the motor
// - TR00 might be Release specific
/* no, it does flap
16516.694         +8 @5693 z rd _37ec 04 STATUS:04 |0|0|0|0|0|TR0|0|0| (I)
18943.802         +8 @5693 z rd _37ec 00 STATUS:00  (I)
18943.828         +8 @5693 z rd _37ec 80 STATUS:80 |NR |0|0|0|0|0|0|0| (I)
19025.672         +8 @5693 z rd _37ec 82 STATUS:82 |NR |0|0|0|0|0|IDX|0| (I)
19031.178         +8 @569c z rd _37ec 80 STATUS:80 |NR |0|0|0|0|0|0|0| (I)
19223.670         +8 @56a8 z rd _37ec 82 STATUS:82 |NR |0|0|0|0|0|IDX|0| (I)
19229.142         +8 @56b8 z rd _37ec 80 STATUS:80 |NR |0|0|0|0|0|0|0| (I)
19339.712         +8 @5693 z rd _37ec 04 STATUS:04 |0|0|0|0|0|TR0|0|0| (I)
21805.734         +8 @5693 z rd _37ec 00 STATUS:00  (I)
21805.758         +8 @5693 z rd _37ec 80 STATUS:80 |NR |0|0|0|0|0|0|0| (I)
21887.752         +8 @5693 z rd _37ec 82 STATUS:82 |NR |0|0|0|0|0|IDX|0| (I)
21893.234         +8 @569c z rd _37ec 80 STATUS:80 |NR |0|0|0|0|0|0|0| (I)
TR00 changes because the target dsk changes as the motor goes off.
The selected disk goes back to the first one.  Seems odd.

m1_expansion.c does it (also true for m3_fdc.cpp)

		if (exp.fdc_motor <= 0) {
			fdc_set_drive(0);
		}

TODO:
- not ready change to blank other bits
  - some documentation to confirm that would be nice
- add JV1 to format tests
  - maybe add pre-formatted disks to format tests?

- maybe fix the index pulse
  - idea is to accumulate time instead of just radial position.
  - would help for 'odd' DMK drives.
*/
