#include <stdio.h>
#include <string.h>
#include <math.h>

#include <assert.h>

#include "fd_image.h"
#include "hd_image.h"

#ifdef _MSC_VER
#define stricmp _stricmp
#define strnicmp _strnicmp
#else
#define stricmp strcasecmp
#define strnicmp strncasecmp
#endif

void HardDriveImage::Fail(const char* msg)
{
	report->Notify(Notifier::NOTIFY_Fail, "%s", msg);
}
void HardDriveImage::Warn(const char* msg)
{
	report->Notify(Notifier::NOTIFY_Warn, "%s", msg);
}

BlockHardDrive::BlockHardDrive(Notifier* err)
	: HardDriveImage(err)
	, cylinders(256)//guess default (153 is 'typical' for 8-meg
	, heads(4)//guess default
	, sectors_per_head(17)
	, sector_length(512)
	, min_sector_id(1)
	, interleave(1)
	, m_file_len(0)
	, m_file(NULL)
{
}

void BlockHardDrive::SetGeometry(int heads_, int cylinders_, int sectors_per_track_, int sector_size_, int interleave_, int min_id_)
{
	heads = heads_;
	cylinders = cylinders_;
	sectors_per_head = sectors_per_track_;
	sector_length = sector_size_;
	interleave = interleave_;
	min_sector_id = min_id_;

	// Modify m_config, mostly for the DREM Hard Drive support which uses this directly.
	m_config.sectors_per_track = sectors_per_track_;
	m_config.sides = heads;
	m_config.tracks = cylinders;
}

void BlockHardDrive::NotifyFormat(int cylinder, int head, int sector_count, int sector_size, unsigned char* buffer, int gap3)
{
	// There needs to be checks here, various cases (and we need a track and cylinder)
	// if track 0, we can 'reset' the drive
	// If not, we don't allow format to change. (# sectors, size)
	// The buffer needs to be scanned to get sector IDs
	// - we don't allow gaps in the IDs.
	// We do allow the format to expand the # of cylinders.
	// We allow cylinder 0 to expand the # of heads.
	// (This is a bit fancy, but it helps in debugging)

	// Also, look for bad block markers or generally just 'junk' in the format.
	
	// This code is in line to change to test LSDOS formatting.
#if 0
	// Adaptable size would be really nice.
	if (data == NULL)
	{
		int min = -1;
		int max = -1;
		for (int i = 0; i < sector_count; i++)
		{
			assert(buffer[i*2] == 0);		// if 0x80, a bad block, we don't support it.
			int snum = buffer[i*2+1];
			if (min == -1 || snum < min) min = snum;
			if (max == -1 || snum > max) max = snum;
		}
		if (min == -1 || max == -1 || (max - min + 1) != sector_count)
		{
			Warn("we will have some problems\n");
		}
		min_sector_id = min;

		sectors_per_head = sector_count;
		sector_length = sector_size;
		data_len = cylinders * heads * sector_count * sector_size;
		
		data = new unsigned char[data_len];
		memset(data, 0, data_len);
	}
#endif
}

void BlockHardDrive::GetTrackInfo(TrackInfo& info, int head, int cylinder)
{
	info.sectors = sectors_per_head;
}


int BlockHardDrive::GetMinSectorID(int head, int cylinder)
{
	return min_sector_id;
}

void BlockHardDrive::GetSectorInfo(SectorInfo& info, int head, int cylinder, int sector_id, int expected_length)
{
	info.si_flags = SECTOR_dden;		// MFM encoding
	info.si_length = sector_length;
	info.si_sector = sector_id;
	info.si_track = cylinder;
	info.si_side = head;
}

static void dunsel() { }

void BlockHardDrive::GetSectorData(unsigned char* dst, const SectorInfo& info)
{
	assert(info.si_length == sector_length);

#if 0
	if (SeekToSector(info.si_side, info.si_track, info.si_sector))
	{
		int len = fread(dst, 1, sector_length, m_file);
		assert(len == sector_length);
	}
	else
	{
		memset(dst, 0, sector_length);
	}
#else
	// Initially result to zero so empty drives look empty
	memset(dst, 0, info.si_length);

	// No-check version for formatting
	(void)SeekToSector(info.si_side, info.si_track, info.si_sector);
	// gcc says you must pay attention to fread result.
	if (fread(dst, 1, sector_length, m_file))
		dunsel();
#endif
}

void BlockHardDrive::PutSectorData(const unsigned char* src, const SectorInfo& info)
{
	assert(info.si_length == sector_length);
#if 0
	if (SeekToSector(info.si_side, info.si_track, info.si_sector))
	{
		int len = fwrite(src, 1, sector_length, m_file);
		assert(len == sector_length);
	}
#else
	// No-check version for formatting
	(void)SeekToSector(info.si_side, info.si_track, info.si_sector);
	(void)fwrite(src, 1, sector_length, m_file);
#endif
}

int BlockHardDrive::SeekToSector(int head, int cylinder, int sector)
{
	if (m_file == NULL)
	{
		return 0;
	}

#if 0
	assert(head >= 0 && head < heads);
	assert(cylinder >= 0 && cylinder < cylinders);
	assert(sector >= min_sector_id && sector < sectors_per_head + min_sector_id);

	int offset = cylinder * sectors_per_head * heads + head * sectors_per_head + ((sector - min_sector_id) % sectors_per_head);
	offset *= sector_length;

	assert(offset < m_file_len && offset + sector_length <= m_file_len);

	fseek(m_file, offset, SEEK_SET);
	int pos = ftell(m_file);
	assert( pos == offset );
#else
	// No-check version for formatting
	int offset = cylinder * sectors_per_head * heads + head * sectors_per_head + ((sector - min_sector_id) % sectors_per_head);
	offset *= sector_length;

	(void)fseek(m_file, offset, SEEK_SET);
#endif
	return 1;
}

void BlockHardDrive::Load(FILE* fp, bool read_only, const DiskImageMetadata *stamp)
{
	m_file = fp;

	if (fseek(fp, 0, SEEK_END) != 0)
	{
		Fail("Seek error");
		return;
	}

	m_file_len = ftell(fp);

	// sanity check
	int expected_size = heads * cylinders * sectors_per_head * sector_length;

	if (expected_size != m_file_len)
	{
		Warn("Not the expected size");
	}
}

void BlockHardDrive::SaveToFile()
{
	fclose(m_file);
	m_file = NULL;
}

// Code to check magic numbers in formats given a small buffer.

static int Detect_HDV1(unsigned char *buf, int length)
{
	if (length >= 256)
	{
		if (buf[0] == 0x56 && buf[1] == 0xcb)
		{
			return 1;
		}
	}

	return 0;
}

static unsigned char* skip_bom(unsigned char* buf, int length)
{
	const char* UTF8_BOM = "\xef\xbb\xbf";
	if (memcmp(buf, UTF8_BOM, 3) == 0) {
		buf += 3;
		length -= 3;
	}

	return buf;
}

static int Detect_DREM(unsigned char *buf, int length)
{
	if (length < 32)
	{
		return 0;
	}

	buf = skip_bom(buf, length);

	// Detect DSK inside any header line e.g. [*DSK*]
	int state = 0;
	for (int i = 0; i < length; i++)
	{
		unsigned int ch = buf[i];

		// Back to start of line regardless of state
		if (ch == '\n' || ch == '\r') {
			state = 0;
			continue;
		}

		switch (state)
		{
		case -1:// failed on this line, waiting for a newline.
			break;

		case 0://start of line
			if (ch == '[')
				state = 1;
			else if (!(ch == '\t' || ch == ' ')) {
				state = -1;
			}
			break;

		// This is a bit sloppy, we are allowing almost anything to occur after '['
		case 1:// inside '['
			if (ch == ']')
				state = -1;
			else if (i + 3 <= length && strncmp((const char*)&buf[i], "DSK", 3) == 0)
				state = 2;
			break;

		case 2:// detected [..DSK..]
			if (ch == ']') {
				return 1;
			}
		}
	}

	return 0;
}

static unsigned char emu_magic[8] = { EMU_DRIVE_MAGIC };

static int Detect_EMU(unsigned char *buf, int length)
{
	if (length >= 16)
	{
		// byte 11 is the file type
		if (memcmp(buf, emu_magic, sizeof(emu_magic)) == 0 && buf[11] == 2)
		{
			return 1;
		}
	}
	return 0;
}

#ifdef ENABLE_HDV_UPGRADE
char* hdv1_model2_upgrade(const char *filename, FILE* fp, const char* op)
{
	static char message[1024];

	sprintf(message, "Apply changes to update HDV file '%s' to Model II compatible:\n\n", filename);

	char* need_upgrade = NULL;

	int do_fix = (strcmp(op, "fix") == 0);

	(void)fseek(fp, 0, SEEK_SET);
	unsigned char header[256];

	int header_len = fread(header, 1, sizeof(header), fp);

	(void)fseek(fp, 0, SEEK_SET);

/*
Byte 8, bit 1
    0 = original format
    1 = extended format

If bit 1 is set then:

Byte 29 is sectors/track (not sectors/cylinder)

bits[3:2]  of byte 8 give the sector size:
    00 = 128  01 = 256  10 = 512  11 = 1024
*/

	if (Detect_HDV1(header, header_len))
	{
		// HDV1 file, detect possible need to upgrade

		// Check Model II bit is set
		if ((header[8] & 2) == 0)
		{
			need_upgrade = message;
			sprintf(message + strlen(message), "* Add Model II flag 0x02 to byte 8\n");
			if (do_fix) {
				header[8] |= 2;
			}
		}
		// Check sector size is 512 bytes
		if ((header[8] & 0xc) != 8)
		{
			need_upgrade = message;
			sprintf(message + strlen(message), "* Flag sector size as 512\n");
			if (do_fix) {
				header[8] &= ~0xc;			// clear bits
				header[8] |= 8;				// 0b10 = 512 byte sectors.
			}
		}

		// Check if sectors_per_cylinder looks wrong
		int heads = header[26];
		int cylinders = ((header[27] & 7) << 8) | header[28];
		if (cylinders == 0)
		{
			cylinders = 256;
		}
		int sectors_per_cylinder = header[29];

		// This check is for model 2 images created before Model 2 hinting
		if (heads != 0)
		{
			if ((sectors_per_cylinder * heads >= 256 || sectors_per_cylinder > 50) && sectors_per_cylinder % heads == 0)
			{
				need_upgrade = message;
				sprintf(message + strlen(message), "* Change %d sectors/cylinder to %d sectors/track\n", sectors_per_cylinder, sectors_per_cylinder / heads);
				// Must be actual sectors_per_cylinder change to sectors_per_head
				// change it to the expected sectors_per_head
				// change the header and save it
				if (do_fix)
				{
					header[29] = sectors_per_cylinder / heads;
				}
			}
		}

		if (do_fix)
		{
			// Write fixed header
			(void)fseek(fp, 0, SEEK_SET);
			fwrite(header, 1, header_len, fp);
			(void)fseek(fp, 0, SEEK_SET);
		}
	}

	return need_upgrade;
}
#endif

// Add error delegate, produce errors if not HDV file (e.g. bad magic)
// replaces *&error at least.
HardDriveImage* OpenHardDriveImage(FILE *fp, bool read_only, const char* path, const char* format_name, const unsigned char *data_in, int len, Notifier* report, const DriveConfiguration& config, const DiskImageMetadata *stamp)
{
	//TODO: data means a pre-arranged image
	assert(data_in == NULL);

	HardDriveImage* drive_image = NULL;

	unsigned char buf[4096];
	if (fseek(fp, 0, SEEK_END) != 0)
	{
		report->Notify(Notifier::NOTIFY_Fail, "Seek Error");

		return NULL;
	}

	if (fseek(fp, 0, SEEK_SET) != 0)
	{
		report->Notify(Notifier::NOTIFY_Fail, "Seek Error");

		return NULL;
	}

	int buf_len = fread(buf, 1, sizeof(buf), fp);

	// Seek back to start (EMU loading at least required it)
	if (fseek(fp, 0, SEEK_SET) != 0)
	{
		report->Notify(Notifier::NOTIFY_Fail, "Seek Error");

		return NULL;
	}

	if (strcmp(format_name, "emu") == 0)
	{
		extern void make_crc_table();

		make_crc_table();

		if (buf_len > 0 && !Detect_EMU(buf, buf_len))
		{
			report->Notify(Notifier::NOTIFY_Fail, "Not a DG EMU hard drive image");

			return NULL;
		}

		drive_image = hd_create_emu(report);
		drive_image->Load(fp, read_only, stamp);

		return drive_image;
	}
	else if (strcmp(format_name, "drem") == 0)
	{
		if (buf_len == 0)
		{
			hd_image_create_blank_drem(fp, report);
		}

		if (buf_len > 0 && !Detect_DREM(buf, buf_len))
		{
			report->Notify(Notifier::NOTIFY_Fail, "Not a DREM hard drive config file");
			return NULL;
		}

		drive_image = hd_create_drem(report, path, read_only, fp);
		if (drive_image) {
			drive_image->Load(fp, read_only, stamp);
		}
	}
	else
	{
		if (buf_len > 0 && !Detect_HDV1(buf, buf_len))
		{
			report->Notify(Notifier::NOTIFY_Fail, "Not an HDV1 hard drive image");
			return NULL;
		}

		drive_image = hd_create_hdv1(report);
		drive_image->Load(fp, read_only, stamp);
	}

	if (drive_image)
	{
		//TODO: fetch geometry
	}

	return drive_image;
}

#include <errno.h>

FILE* OpenOrCreateFile(const char* filename, bool& read_only, Notifier* report)
{
	read_only = false;
	FILE *fp;

	bool readonly_ok = false;
	fp = fopen(filename, "rb");
	if (fp != NULL)
	{
		readonly_ok = true;
		fclose(fp);
	}

	fp = fopen(filename, "r+b");	// existing file, read/write
	if (fp == NULL)
	{
		fp = fopen(filename, "w+b");	// read/write but create or truncate
	}
	if (fp == NULL)
	{
		if (!readonly_ok)
		{
			report->Notify(Notifier::NOTIFY_Fail, "Could not create file %s: %s", filename, strerror(errno));

			return NULL;
		}

		fp = fopen(filename, "rb");		// read only
		if (fp == NULL)
		{
			report->Notify(Notifier::NOTIFY_Fail, "Could not open file %s: %s", filename, strerror(errno));

			return NULL;
		}
		read_only = true;
	}
	
	return fp;
}

FILE* OpenFileOnly(const char* filename, bool& read_only)
{
	read_only = false;
	FILE *fp;

	fp = fopen(filename, "r+b");	// existing file, read/write
	if (fp == NULL)
	{
		fp = fopen(filename, "rb");		// read only existing file.
		if (fp != NULL)
		{
			read_only = true;
		}
	}

	return fp;
}

char *ReplaceExtension(const char* path, const char *ext)
{
	int len = strlen(path);
	const char *p = path ? strrchr(path, '.') : 0;
	if (p)
	{
		len = p - path;
	}

	char* res = new char[len + 1 + strlen(ext) + 1];
	memcpy(res, path, len);
	res[len] = '.';
	strcpy(&res[len + 1], ext);

	return res;
}
