
enum WD17xx_Family {
	CHIP_WD1771,
	// 179[1-7] series
	// 1791/1793 (Model II/III) the same except for inverted data lines
	// 179[24] are single density only versions
	// 179[57] has a side-select output for double-sided drives.
	CHIP_WD1793
} ;

// Hard drive controllers
enum WD1010_Family {
	CHIP_8X300,		// Implemented using Synetics 8X300 processor
	CHIP_WD1010,	// WD1010 chip (1024 tracks)
	CHIP_WD2010,	// Supports 2048 tracks
} ;

#define FDC_MICROSECONDS (32)

// This was 43 from Model II tech manual.
// I suspect it should be larger as I may
// be measuring from the wrong point.
// Increased to 48 -- enough to boot robotd disk.
//ERROR yes this is wrong and needs to change (and change for density)
#define DAM_MARKER_TIMEOUT (48)

// Correct values, but must be counted after the end of the CRC.
#define DAM_MARKER_TIMEOUT_FM (30)
#define DAM_MARKER_TIMEOUT_MFM (43)

// Maximum sector size supported by all formats
#define MAX_SECTOR_LENGTH (1024)

//#define DMK2_TESTING

// JV3_SIDE_FIX is the right idea but should generate warnings
#define JV3_SIDE_FIX


// Known formats
enum DiskImageFormat
{
	DISK_DMK,
	DISK_JV1,
	DISK_JV3,
	DISK_IMD,
	DISK_HFE,

	// Not image formats
	DISK_empty,
} ;

// Known encoding formats
enum DiskEncoding
{
	DISK_FM,		// "low" density
	DISK_MFM,		// "high" density
} ;

// Error/Notification interface
class Notifier
{
public:
	enum Level
	{
		NOTIFY_Fail,
		NOTIFY_Warn,
		NOTIFY_Note,
	} ;

	static const int max_notify_level = NOTIFY_Note + 1;

	Notifier();
	virtual ~Notifier();

	virtual void Notify(Level lev, const char* fmt, ...);
	virtual void Reset();

	int m_note_count[max_notify_level];
	char* m_notifications[max_notify_level];
} ;

// Basic drive information (representing the drive hardware)
// The values are generally advisory

// ERROR: data_rate is a mess. It really means the flux or bit-clock rate
// The actual rate of reading data bytes depends on the encoding.
// For MFM, the byte rate will be the data rate / 8,
// but with FM the byte rate will be half that (data rate / 4)
// HOWEVER, the machine configuration is passing
// down 125 Kbit/s for the Model I.

// Option bits for DriveConfiguration
enum {
	DRV_side_lock = 1<<0,		// No automatic promotion to double-sided
	// Potentially a "track lock" to restrict appending tracks
} ;

struct DriveConfiguration
{
	int rpm;
	int data_rate;			// bits/s
	DiskEncoding density;
	int sides;
	int tracks;
	int sectors_per_track;	// typical, say given 256-byte sectors
	int physical_size;		// 8 or 5(.25) inches.
	int geometry_options;
} ;

// Metadata in an image
struct DiskImageMetadata
{
	// Creation time
	int yyyy, mm, dd, hour, min, sec;
	// TRS-80 Model (zero if unknown or not a TRS-80)
	// Otherwise: 1 = model I/III/4/4p
	// 2 = model 2/12/16/6000
	int trs80_model_family;
	// Some images include controller or OS information
} ;

// Common block definition structs
struct DiskInfo {
    int tracks;
    int sides;
} ;

struct TrackInfo {
    int sectors;
} ;

enum SectorFlag {
    SECTOR_dden = 1<<0,		// Encoding: 1 = MFM (double-density) 0 = FM (single density)
    SECTOR_err = 1<<1,
	SECTOR_delete = 1<<2,
	SECTOR_unavail = 1<<3,
	SECTOR_damhigh = 1<<4,
	// Potentially: SECTOR_nonibm = 1 << 6; // sector length is multiple of 16
	SECTOR_savemarker = 1<<6,	// temporary for JV3 saving
	SECTOR_rate_bit = 8,
	SECTOR_rate_mask = 255 << SECTOR_rate_bit,
	SECTOR_rate_scale = 25,
};
struct SectorInfo {
    int si_flags;  // SectorFlag values
    int si_length; // Actual length in bytes
    int si_sector; // marked sector number
    int si_track; // marked track number
    int si_side; // marked side number
	void set_rate(int rate);
	int get_rate() { return ((si_flags & SECTOR_rate_mask) >> SECTOR_rate_bit) * SECTOR_rate_scale; }
	int get_dam(WD17xx_Family chip);
	int get_dam(WD1010_Family chip) { return 0xf8; }
	int set_dam(WD17xx_Family chip, int dam);
	int get_full_dam(int /*density*/);
	int get_idam(WD17xx_Family /*chip*/) { return 0xfe; }
	int get_idam(WD1010_Family chip);
} ;

struct SideInfo
{
    TrackInfo *track_info;
    SectorInfo **sector_info;
	unsigned char ***sector_data;
} ;

class BlockInterface
{
public:
    virtual void GetDiskInfo(DiskInfo& info) { }
    virtual void GetTrackInfo(TrackInfo& info, int side, int track) { }
    virtual void GetSectorInfo(SectorInfo& info, int side, int track, int sector) { }
    virtual void GetSectorData(unsigned char* res, int side, int track, int sector) { }
    
    virtual void PutDiskInfo(const DiskInfo& info) { }
    virtual void PutTrackInfo(const TrackInfo& info, int side, int track) { }
    virtual void PutSectorInfo(const SectorInfo& info, int side, int track, int sector) { }
    virtual void PutSectorData(const unsigned char* res, int side, int track, int sector) { }

	virtual void GetDMKTrack(class DMKTrack *trk, int side, int track) { }
protected:
	int radial_position, tracklen;
	int stride[2]; // How much to advance given density (0=FM, 1=MFM)
} ;

//WIP: Callout class used for scanning tracks (debug text dumps or converting to block formats)
//WIP: Also for bit-level dumps of HFE disk images
class ScanTrack
{
public:
	virtual void Scan_Begin();
	virtual void Scan_Gap();
	virtual void Scan_Idam();
	virtual void Scan_Dam();
	virtual void Scan_End();

	int m_side, m_track;
} ;

// Base class for disk drive format implementations.
class DiskDrive : public BlockInterface {
public:
	DiskDrive(const DriveConfiguration& config, const DiskImageMetadata* metadata);
	virtual ~DiskDrive();
	virtual int Load(unsigned char *data, int count) = 0;
	virtual const char* Save(FILE *fp) = 0;
	virtual int IsModified() = 0;
	virtual void DebugDump() = 0;
	virtual void SetTrack(int track);
	virtual int GetTrack();
	virtual int GetSide();
	virtual int GetNumSides();
	virtual void SetReady(int rdy);
	virtual void SetOffset(int offset);
	virtual int GetOffset();
	virtual int AtIndexHoleByte(int density);
	virtual int AppendBlankTrack();
	virtual int AppendBlankSide() { return 0; }
	virtual DiskImageFormat ImageFormat() = 0;

	virtual const DriveConfiguration* GetConfig() { return &m_config; }

	// Pin interface
	virtual int GetTR00();
	virtual int GetReady();
	virtual int IndexPulse();
	virtual int GetFileWriteable();
	virtual int GetWriteProtect();
	virtual void SetWriteProtect(int on);
	virtual void StepTrack(int dir);	// combination DIRC/STEP
	virtual void SetSide(int side);
	virtual int GetIndexHoleBlocked();
	virtual void SetIndexHoleBlocked(int blocked);
	virtual void SetGeometryFlags(int flags);		// Can read flags via GetConfig()

	// Byte routines combine RD/WD (and FM/MFM encoding)
	virtual int GetByte(int density, int* clock) = 0;
	virtual void WriteByte(int density, int val, int clock) = 0;
	virtual void AdvanceByte(int density) = 0;
	virtual void AlignToMark(int density) { (void)density; }
	virtual void NotifyWriteTrack(int writeTrackOn) { }

	// Physical configuration of drive
	// Loaded images may exceed some parameters (especially tracks)
	DriveConfiguration m_config;

	// Chip type accessing drive, needed for JV3 support
	WD17xx_Family m_chip;

	// FILE* based loading
	// Load/Save routines return NULL for OK otherwise error string
	FILE* m_file;
	int m_file_writeable;
	const char* LoadFromFile(FILE* f, unsigned char* data, int count, int writeable);
	const char* SaveToFile();
	const char* SaveAs(const char *filename);
	const char* Close();

	// TODO: This was added for debugging
	virtual int GetRawByte(int offset) = 0;

	//WIP: Debug/Viewing
	virtual char *TrackDump(int side, int track);
	virtual void TrackDump(FILE* f, int side, int track){}
	virtual void TrackScan(int side, int track, ScanTrack* callback);

	// The basic geometry of the drive should be fixed.
	// The data files are usually truncated to the last used track,
	// not the last physical track.
protected:
	int ensure_disk_size(int required_len);
	const char* FileError();
	int len;
	int len_max;
	unsigned char* disk;
    DiskInfo info;
	int writable;
	int dirty;

	// Drive state
	int current_track;
	int current_side;
	int tr00;
	int ready;
	int index_hole_blocked;

	void CheckAllowTrackAppend();
	int allow_track_append;
	int requested_side;
} ;

// Code for handling a single track
// Very similar to code inside of DMKDisk
#define DMK_MAX_IDAM (64)
#define DMK_INDEX_LENGTH (DMK_MAX_IDAM*2)

class DMKTrack
{
public:
	DMKTrack()
		: dirty(0)
		, prev_clocked_a1(0)
		, radial_position(0)
		, single_density(0)
		, tracklen(0)
		, data_rate(0)
	{
		for (int i = 0; i < DMK_MAX_IDAM; i++)
		{
			idamoffset[i] = 0;
		}
		for (size_t i = 0; i < sizeof(trackdata)/sizeof(trackdata[0]); i++)
		{
			trackdata[i] = 0x00;
		}
	}
	void SetIdam(int* idam);
	void SetData(unsigned char* data, int len, int single_d);
	void GetIdam(int* idam);
	void GetData(unsigned char* dst, int len);
	void SetMinimumTracklen(int len);
	void SetDataRate(int rate);

	int GetPosition();
	void SetPosition(int pos);

	void RemoveIDAM(int index);
	void InsertIDAM(int offset, int density);
	int GetByte(int density, int* clock);
	int GetRawByte(int offset);
	void WriteByte(int density, int val, int clock);
	void AdvanceByte(int density);
	int IsModified()
	{
		return dirty;
	}
	void SetWritten()
	{
		dirty = 0;
	}

	// Read from DMK data
	void ReadDMKTrack(
		unsigned char* dmkdata,
		int dmkdata_len,
		int opt_single_density,
		int data_rate
	);

	void GetTrackInfo(TrackInfo& info);
	void GetSectorInfo(SectorInfo& info, int sector);
	void GetSectorData(unsigned char* res, int sector);
	int GetTracklenAndStride(int *_stride = NULL);

	int FindDAM(int sector, int& delta, int& idam_offset, int& idam_len, int& dam_offset, int& dam_len);	// returns density

	void Dump(FILE *f, int tracknum);
	void Dump(void (*output)(void *ctx, const char* text), int tracknum, void* ctx);
	char* Dump(int tracknum);

protected:
	void setup_stride();
	int stride[2];
	void Reset();

	int GetSectorLength(int sector);
	int dirty;
	int prev_clocked_a1;
	int radial_position;
	int single_density;
	int tracklen;

	int data_rate;
	int idamoffset[DMK_MAX_IDAM];
	unsigned char trackdata[16384];		//ERROR: not a fixed allocation (but it is the real max)
} ;

// TrackProgram converts block-based sector information into a readable track
class TrackProgram
{
public:
	void BuildTrackProgram(int num_sector, SectorInfo* sec_info, unsigned char** sec_data, const DriveConfiguration& config, WD17xx_Family chip);

	int GetByte(int , int* clock);
	void WriteByte(int density, int val, int clock);
	void AdvanceByte(int density);

	void SetRadialPosition(int pos);

	int radial_position;
	int tracklen;
	SectorInfo* sector_info;
	unsigned char** sector_data;

	int pc, cmd_offset, tc_density, tc_crc, tc_in_crc;
	int dirty;

	int stride[2];

	WD17xx_Family m_chip;
	int track_command[16384];
private:
	void ProcessNonDataCommands();
	void RunCompleteTrackToGetLength();
} ;

// Type-detecting loader.
DiskDrive* LoadDisk(const char* fname, const unsigned char *data, int len, const char*& error, const DriveConfiguration& config);

// Return >= 0 if string is a known disk type format
int DiskFormatName(const char* name, int *sides = NULL, int* tracks = NULL);

// Creating blank disks
DiskDrive* LoadBlankDisk(const char* format, const char*& error, const DriveConfiguration& config, const DiskImageMetadata *metadata = NULL);

// Saving ASCII dump of disk for debugging
void SaveDiskASCII(FILE *f, DiskDrive* dsk, const char* options);

// DMK format drive - reimplemented using DMKTrack
class DMKDiskDrive2 : public DiskDrive
{
	// Disk data
	int options;
	int opt_ignore_density, opt_single_density, opt_single_sided;
	int magic;

public:
	DMKDiskDrive2(const DriveConfiguration &config, const DiskImageMetadata* metadata = NULL);
	virtual int Load(unsigned char *data, int count);
	virtual const char* Save(FILE *fp);

	int IsModified();

	void SetTrack(int track);
	void SetSide(int side);

	int GetByte(int density, int* clock);
	void WriteByte(int density, int val, int clock);
	virtual void AdvanceByte(int density);
	virtual DiskImageFormat ImageFormat() { return DISK_DMK; }

	virtual int GetRawByte(int offset);
	virtual void DebugDump();

    virtual void GetDiskInfo(DiskInfo& info);
    virtual void GetTrackInfo(TrackInfo& info, int side, int track);
    virtual void GetSectorInfo(SectorInfo& info, int side, int track, int sector);
    virtual void GetSectorData(unsigned char* res, int side, int track, int sector);

	virtual void GetDMKTrack(class DMKTrack *trk, int side, int track);

protected:
	void setup_stride();
	//notes:
	//- # tracks in DriveConfiguration or 	int writable, sides, tracks;
	//- load and save do copying to fill in DMKTrack
	//- IsModified() sees if any track was written.
	//- radial position based on current side/track.
	DMKTrack track[2][256];	//blah huge
} ;

// Common class for drives formats which only read blocks
// It provides common code to simulate tracks
class BlockDrive : public DiskDrive
{
public:
    BlockDrive(const DriveConfiguration &config, const DiskImageMetadata* metadata = NULL);
    virtual ~BlockDrive();

	void SetTrack(int track);
	void SetSide(int side);
	int GetRawByte(int offset);

    virtual void GetDiskInfo(DiskInfo& info);
    virtual void GetTrackInfo(TrackInfo& info, int side, int track);
    virtual void GetSectorInfo(SectorInfo& info, int side, int track, int sector);
    virtual void GetSectorData(unsigned char* res, int side, int track, int sector);

 	virtual void GetDMKTrack(class DMKTrack *trk, int side, int track);
	virtual void PutDMKTrack(class DMKTrack *trk, int side, int track);

	void CreateTrackProgram();
	//TODO
	int IsModified();
	void DebugDump() {}
	virtual int GetByte(int density, int* clock);
	void WriteByte(int density, int val, int clock);
	virtual void AdvanceByte(int density);
	virtual void NotifyWriteTrack(int writeTrackOn);
	virtual int AppendBlankTrack();
	virtual int AppendBlankSide();

protected:
	void DeleteAllSectors();
	int tc_track;
	TrackProgram trk_program;

	struct WriteTrackState *trkstate;

	SideInfo *side_info;
	int track_command[16384];		// current track
} ;


// JV1 format drive.
class JV1DiskDrive : public BlockDrive
{
	public:
	JV1DiskDrive(const DriveConfiguration &config, const DiskImageMetadata* metadata = NULL);
	int Load(unsigned char *data, int count);
	virtual const char* Save(FILE *fp);
	virtual DiskImageFormat ImageFormat() { return DISK_JV1; }
} ;

// JV3 format drive.
struct JV3SectorInfo {
	int trk;
	int sec;
	int density;
	int dam;
	int side;
	int badCRC;
	int nonibm;
	int lenCode;
	int len;
	unsigned char* data;	// sector data (temp) or free data
	int sec_index;		// index for sector data;
	int raw_flags;
} ;

#define JV3_RECORD_COUNT (2901)

class JV3DiskDrive : public BlockDrive
{
	public:
	JV3DiskDrive(const DriveConfiguration &config, const DiskImageMetadata* metadata = NULL);
	int Load(unsigned char *data, int count);
	virtual const char* Save(FILE *fp);
	virtual DiskImageFormat ImageFormat() { return DISK_JV3; }
	virtual void NotifyWriteTrack(int writeTrackOn);
private:
	const char* WriteSecInfo(FILE *fp, const SectorInfo* info);
	const char* WriteSecData(FILE* fp, int seclen, const unsigned char* data);
	int numSector;
	int secInfoValid;
	JV3SectorInfo secInfo[JV3_RECORD_COUNT];
} ;

// IMD format drive.
class IMDDiskDrive : public BlockDrive
{
	public:
	IMDDiskDrive(const DriveConfiguration &config, const DiskImageMetadata* metadata = NULL);
	int Load(unsigned char *data, int count);
	virtual const char* Save(FILE *fp);
	virtual DiskImageFormat ImageFormat() { return DISK_IMD; }
    
	// Methods added for xtar.  Definitely need to be generalized into BlockDrive
	void MakeTrack(int track, int sector_count, int sector_len, int first_id, int density);
	void WriteSector(int side, int track, int sector_index, int sector_len, const unsigned char* data);

    int comment_len;
} ;

#define HFE_MAGIC "HXCPICFE"
#define HFEv3_MAGIC "HXCHFEV3"

#define HFE_BLOCK_SIZE (512)

// HFE (version 0) format (HxC floppy format used in HxC / GoTek / FlashFloppy / Lotharek

struct HFETrackSectorInfo
{
	HFETrackSectorInfo()
		: m_data_offset(-1)
	{
	}

	SectorInfo m_info;
	unsigned int m_mark;
	int m_data_offset;

	// Position of data in flux stream
	int m_bitpos;	//TODO: I don't think this is accurately recorded
} ;

enum HFETrackState
{
	HFETRACK_Reset,
	HFETRACK_Read,
	HFETRACK_Write,
} ;

struct HFETrack
{
	HFETrack()
		: m_state(HFETRACK_Reset)
        , m_flux_len(0)
        , m_flux_data(NULL)
		, m_num_sector(0)
		, m_info(NULL)
		, m_data_len(0)
		, m_data(NULL)
		, m_prevWrite(0)
	{
	}
	~HFETrack()
	{
		delete [] m_info;
		delete [] m_data;
		delete [] m_flux_data;
	}

	void reset()
	{
		m_state = HFETRACK_Reset;
		m_num_sector = 0;
		m_data_len = 0;
		delete [] m_info;
		delete [] m_data;
		delete [] m_flux_data;
		m_info = NULL;
		m_data = NULL;
		m_flux_data = NULL;
		m_prevWrite = 0;
	}
    
	HFETrackState m_state;

	int m_flux_len;
	unsigned char* m_flux_data;

	int m_num_sector;
	HFETrackSectorInfo *m_info;
	int m_data_len;
	unsigned char* m_data;

	unsigned char m_prevWrite;
} ;

class HFEDiskDrive : public DiskDrive
{
public:
	HFEDiskDrive(const DriveConfiguration &config, const DiskImageMetadata* metadata = NULL);
	int Load(unsigned char *data, int count);
	virtual const char* Save(FILE *fp);
	virtual DiskImageFormat ImageFormat() { return DISK_HFE; }
    
	// Block interface
    virtual void GetDiskInfo(DiskInfo& info);
    virtual void GetTrackInfo(TrackInfo& info, int side, int track);
    virtual void GetSectorInfo(SectorInfo& info, int side, int track, int sector);
    virtual void GetSectorData(unsigned char* res, int side, int track, int sector);

	// Byte interface
	virtual void SetTrack(int track);
	virtual void SetSide(int side);
	virtual int GetByte(int density, int* clock);
	virtual void WriteByte(int density, int val, int clock);
	virtual void AdvanceByte(int density);
	virtual void AlignToMark(int density);
	virtual void NotifyWriteTrack(int writeTrackOn);
	virtual int AppendBlankTrack();
	virtual int AppendBlankSide();
	virtual int GetRawByte(int offset);
	virtual int IsModified();
	virtual void DebugDump();

	// Debug/Viewing
	virtual char *TrackDump(int side, int track);
	virtual void TrackDump(FILE* f, int side, int track);
	virtual void TrackScan(int side, int track, ScanTrack* callback);

protected:
	void update_header();

	void flush_track(int density);
    int scan_track(int SIDE, int TRACK, HFETrack& track, int pass, void (*output)(void *ctx, const char* text) = NULL, void* ctx = NULL);
    int set_track(int SIDE, int TRACK, int density);
	void go_track(int SIDE, int TRACK, int density);
	void read_track(int SIDE, int TRACK, HFETrack& track);
	void write_track(int SIDE, int TRACK, HFETrack& track);

	// Header data
	unsigned char magic[9];
	int version;
	int encoding;
	int rate;
	int rpm;
	int uimode;
	int track_list;		// 512-byte block number
	//int write_allow;	// generally FF or 00
	int single_step;	// meaning?

	struct track_encode {
		int use_alt;	// 0x00 mean use
		int encode;
	} ;
	track_encode alt[2];

	struct track_position {
		int block;
		int len;
	};
	track_position track_pos[256];	// block, len pairs

	// State
	HFETrack m_currentTrack;
	int radial_bit_position;
} ;


DiskDrive* fd_create_dmk(const DriveConfiguration &config, const DiskImageMetadata* metadata);
DiskDrive* fd_create_imd(const DriveConfiguration &config, const DiskImageMetadata* metadata);
DiskDrive* fd_create_jv3(const DriveConfiguration &config, const DiskImageMetadata* metadata);
DiskDrive* fd_create_jv1(const DriveConfiguration &config, const DiskImageMetadata* metadata);
DiskDrive* fd_create_hfe(const DriveConfiguration &config, const DiskImageMetadata* metadata);

DiskDrive* fd_create_nodiskette(const DriveConfiguration &config);

// Util

unsigned int update_crc(unsigned int crc, unsigned int data);
void make_crc_table();
unsigned int update_crc_table(unsigned int crc, unsigned int data);
void store_crc(int crc, unsigned char* data, int len, unsigned char* res);

// Debug util
void file_out(void *fp, const char* data);
void char_size(void *ctx, const char* data);
void char_out(void *ctx, const char* data);
void dump_segment(void* ctx, void (*output)(void *ctx, const char* text), const char* name, int sector, unsigned char* data, int offset, int end, int delta, int mark_len = 0, int crc_offset = -1000);

