#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <ctype.h>

#ifdef _MSC_VER
#define stricmp _stricmp
#else
#define stricmp strcasecmp
#endif

#if defined(__APPLE__) || defined(__linux__)
#define _strnicmp strncasecmp
#endif

#include "fd_image.h"

// DMK format drive.
class DMKDiskDrive : public DiskDrive
{
	// Disk data
	int options;
	int opt_ignore_density, opt_single_density, opt_single_sided;
	int magic;
	unsigned char *trackptr[2][256];
	int idamoffset[2][256][64];

	int prev_clocked_a1;
	void RemoveIDAM(int index);
	void InsertIDAM(int offset, int density);

	int find_data(unsigned char* track, int off);

	unsigned char* TrackIndex(int side, int track);

public:
	DMKDiskDrive(const DriveConfiguration &config, const DiskImageMetadata* metadata = NULL);
	virtual int Load(unsigned char *data, int count);
	virtual const char* Save(FILE *fp);
	int IsModified();
	void DebugDump();
	int GetByte(int density, int* clock);
	int GetRawByte(int offset);
	void WriteByte(int density, int val, int clock);
	virtual void AdvanceByte(int density);
	virtual void NotifyWriteTrack(int writeTrackOn);
	virtual int AppendBlankTrack();
	virtual int AppendBlankSide();
	virtual DiskImageFormat ImageFormat() { return DISK_DMK; }

    virtual void GetDiskInfo(DiskInfo& info);
    virtual void GetTrackInfo(TrackInfo& info, int side, int track);
    virtual void GetSectorInfo(SectorInfo& info, int side, int track, int sector);
    virtual void GetSectorData(unsigned char* res, int side, int track, int sector);

	virtual void GetDMKTrack(class DMKTrack *trk, int side, int track);
private:
	void WriteBackIDAMOffsets(int side, int track);
	void setup_stride();

	// Length of a track in file (includes IDAM data)
	int dmk_tracklen;
} ;


// An empty disk drive (distinguished from a blank diskette)
class NoDiskette : public DMKDiskDrive
{
public:
	NoDiskette(const DriveConfiguration& config) : DMKDiskDrive(config)
	{
		// len = 0 causes DMK to return 0 on read and never an IndexPulse
		len = 0;

		// Disable the index pulse
		index_hole_blocked = 1;
	}

	virtual DiskImageFormat ImageFormat() { return DISK_empty; }
	virtual void SetIndexHoleBlocked(int) override {} // blank disk never has index pulse
} ;

// DMK Format
DMKDiskDrive::DMKDiskDrive(const DriveConfiguration &config, const DiskImageMetadata* metadata)
	: DiskDrive(config, metadata)
{
	// Preserved comment about FM.  This is true, but blank tracklen still computed the same way
	// which has been moved to common DiskDrive code
	// Note: DMK supports a special FM (single density) mode
	// so the tracklen does not need correcting for FM data rate as it does for the BlockDrives

	dmk_tracklen = tracklen + DMK_INDEX_LENGTH;	// add extra for the IDAM offset table.
	len = 0x10 + info.tracks * info.sides * dmk_tracklen;
	int size_ok = ensure_disk_size(len);
	assert(size_ok);

	opt_single_density = config.density == DISK_FM ? 1 : 0;
	opt_ignore_density = 0;		//TODO: old flag, probably correct?
	opt_single_sided = (info.sides == 1);

	setup_stride();

	// Copy the options to the 16-byte header and set up a blank image (no sectors)
	memset(disk, 0x00, len);

	disk[0] = writable ? 0x00 : 0xff;

	disk[1] = info.tracks;
	disk[2] = (dmk_tracklen & 0xff);
	disk[3] = (dmk_tracklen >> 8) & 0xff;

	disk[4] = 0; // options
	if (info.sides == 1) disk[4] |= 0x10;
	if (opt_single_density) disk[4] |= 0x40;
	if (opt_ignore_density) disk[4] |= 0x80;
	options = disk[4];
	magic = 0;

	for (int i = 0; i < info.tracks; i++) {
		for (int side = 0; side < info.sides; side++) {
			trackptr[side][i] =  TrackIndex(side, i);
			for (int j = 0; j < DMK_MAX_IDAM; j++) {
				idamoffset[side][i][j] = 0;
			}
		}
	}

	// write checking
	prev_clocked_a1 = 0;

	current_track = 0;
	current_side = 0;
	radial_position = 0;
}

int DMKDiskDrive::AppendBlankTrack()
{
	if (info.tracks > 255)
	{
		return 0;
	}

	unsigned char* old_disk = disk;

	int newlen = 0x10 + (info.tracks + 1) * info.sides * dmk_tracklen;
	if (!ensure_disk_size(newlen))
	{
		return 0;
	}

	// Fix track pointers if disk was reallocated
	if (old_disk != disk) {
		for (int i = 0; i < info.tracks; i++) {
			for (int side = 0; side < info.sides; side++) {
				trackptr[side][i] =  TrackIndex(side, i);
			}
		}
	}

	int trk = info.tracks;

	info.tracks++;
	disk[1] = info.tracks;
	len = newlen;

	for (int side = 0; side < info.sides; side++) {
		trackptr[side][trk] =  TrackIndex(side, trk);
		for (int j = 0; j < DMK_MAX_IDAM; j++) {
			idamoffset[side][trk][j] = 0;
		}
	}

	return 1;
}

int DMKDiskDrive::AppendBlankSide()
{
	if (info.sides > 1)
	{
		return 1;
	}

	// Check available space
	int newlen = 0x10 + info.tracks * 2/*sides*/ * dmk_tracklen;
	if (!ensure_disk_size(newlen))
	{
		return 0;
	}
	len = newlen;

	// Move original track data to new positions
	for (int trk = info.tracks - 1; trk >= 0; trk--)
	{
		info.sides = 1;
		unsigned char* src = TrackIndex(0, trk);
		info.sides = 2;
		unsigned char* dst = TrackIndex(0, trk);
		memcpy(dst, src, tracklen);
	}
	info.sides = 2;
	disk[4] &= ~0x10;	// double-sided

	// Fix track pointers for the original side
	int side = 0;
	for (int trk = 0; trk < info.tracks; trk++) {
		trackptr[side][trk] =  TrackIndex(side, trk);
	}

	// Build empty idam information for new side
	side = 1;
	for (int trk = 0; trk < info.tracks; trk++) {
		trackptr[side][trk] =  TrackIndex(side, trk);
		memset(trackptr[side][trk], 0, tracklen);
		for (int j = 0; j < DMK_MAX_IDAM; j++) {
			idamoffset[side][trk][j] = 0;
		}
	}

	return 1;
}


void DMKDiskDrive::setup_stride()
{
	stride[0] = opt_single_density ? 1 : 2;
	stride[1] = 1;
}

int DMKDiskDrive::find_data(unsigned char* track, int off)
{
	// Count 4e and 00 bytes
	// For double density, the data sheet says 22 4E followed by 12 00
	// I suspect the real controller allows some variation.
	int c4, c0, data;
	for (c4 = 0; (data = track[off]) == 0x4e && c4 < 22; off++) {
		c4++;
	}
	for (c0 = 0; (data = track[off]) == 0x00 && c0 < 16; off++) {
		c0++;
	}
	if (data == 0xa1 && c4 == 22 && c0 == 12) {
		return off;
	}

	return -1;
}

unsigned char* DMKDiskDrive::TrackIndex(int side, int track)
{
	return disk + (track * info.sides + side) * dmk_tracklen + 0x10;
}

int DMKDiskDrive::Load(unsigned char *data, int count)
{
	int size_ok = ensure_disk_size(count);
	assert(size_ok);

	len = count;
	memcpy(disk, data, count);
	int i;

	writable = 1;
	if (disk[0] == 0xff) {
		writable = 0;
	}
	info.tracks = disk[1];
	dmk_tracklen = disk[2] + (disk[3]<<8);

	//printf("writable=%d tracks=%d tracklen=%d\n", writable, tracks, dmk_tracklen);

	options = disk[4];
	
	for (i = 5; i <= 0xb; i++) {
		if (disk[i] != 0) {
			printf("reserved byte at $%02x is not zero ($%02x)\n", i, disk[i]);
		}
	}

	magic = (disk[0xc] | (disk[0xd] << 8) | (disk[0xe] << 16) | (disk[0xf] << 24));
	
	//printf("options=%02x magic=%08x\n", options, magic);
	if (magic == 0x12345678) {
		printf("real drive? not possible\n");
	}

	opt_ignore_density = !!(options & 0x80);
	opt_single_density = !!(options & 0x40);
	opt_single_sided = !!(options & 0x10);
	info.sides = opt_single_sided ? 1 : 2;
	//printf("option detail: MIXED:%d SingleD:%d OneSided:%d\n", opt_ignore_density, opt_single_density, opt_single_sided);

	setup_stride();

	// Check on file size
	int expected_size = info.sides * info.tracks * dmk_tracklen + 0x10;	// 16 byte header
	//int file_err = len - expected_size;
	if (len < expected_size) {
		printf("**ERROR File appears truncated\n");
		int size_ok = ensure_disk_size(expected_size);
		assert(size_ok);
		len = expected_size;
	} else if (len > expected_size) {
		printf("**ERROR File is too long\n");
	}
		
	// Set up track pointers and sector (IDAM) offsets
	int sector_count[2];
	sector_count[0] = sector_count[1] = 0;
	for (i = 0; i < info.tracks; i++) {
		for (int side = 0; side < info.sides; side++) {
			unsigned char* trackindex = TrackIndex(side, i);
			trackptr[side][i] = trackindex; // but starts with 128 bytes of index
			//printf("s%d trk %02d:", side, i);
			for (int j = 0; j < 64; j++) {
				idamoffset[side][i][j] = (trackindex[j*2] | (trackindex[j*2+1] << 8));
				// high bit means, double-density, next highest is ignored
				int off = (idamoffset[side][i][j] & 0x3fff);
				if (off >= dmk_tracklen) {
					printf("***ERROR IDAM offset %04x too large (side %d track %d sector %d)\n", off, side, i, j);
				}
				if (off != 0) {
					int density = ((idamoffset[side][i][j] >> 15) & 1);
					//[unused] int sector = trackptr[side][i][off+3*stride[density]];
					//printf(" %04x%s%02x/%d", off, density ? ":" : "*", trackptr[side][i][off], sector);
					sector_count[density]++;

#define A1_REPAIR
#ifdef A1_REPAIR
					if (density) {
						/*
						Disks I've been looking at:

						C:\q\18q3\RandomDMK\SuperUtility1\SuperUtility-1su.dmk
						C:\q\18q3\RandomDMK\SuperUtility1\x.dmk
						-m4 C:\q\18q3\RandomDMK\doga.dmk

						"C:/Projects/george/trs80/rips/CosmicFighter/download/cosftrv3/Cosmic Fighter v3.0 (1981)(Big Five Software)[M13][Dual Boot][Set 1].dmk"
						"C:/Projects/george/trs80/rips/CosmicFighter/download/cosftrv3/Cosmic Fighter v3.0 (1981)(Big Five Software)[M13][Dual Boot][Set 2].dmk"

						I'm thinking the user presentation of the feature should be something like:
						1. Prompt that inconsistencies detected with a brief list.
						2. Options are to cancel the load, repair and continue, do not repair and continue, or view details.
						3. View details gives a complete text log.  I'm the real user of that, either given the information from a user or viewing it myself.

						There are cases where there are only inconsistencies.  Maybe generically "modify and continue".
						*/

						// IDAM marker check
						unsigned char *idam = trackptr[side][i] + off;
						unsigned char a, b, c, mark;
						a = idam[ -3 * stride[density] ];
						b = idam[ -2 * stride[density] ];
						c = idam[ -1 * stride[density] ];
						mark = idam[0];
						if (a != 0xa1 || b != 0xa1 || c != 0xa1) {
							printf("double density IDAM incomplete 0xa1 leader: {%02x %02x %02x %02x) (side %d track %d sector %d)\n", a, b, c, mark, side, i, j);
							// Repair
							idam[ -3 * stride[density] ] = 0xa1;
							idam[ -2 * stride[density] ] = 0xa1;
							idam[ -1 * stride[density] ] = 0xa1;
						}
						// DAM marker check.  The DAM might not exist.
						// Look for the first $fb or $f8 which is proceed by a $a1
						// (within the marker timeout and past the CRC
						int found = 0;
						for (int k = 7 + 1; k < DAM_MARKER_TIMEOUT + 7; k++)
						{
							mark = idam[ k * stride[density]];
							if (mark == 0xf8 || mark == 0xfb)
							{
								found = 1;
								a = idam[ (k-3) * stride[density] ];
								b = idam[ (k-2) * stride[density] ];
								c = idam[ (k-1) * stride[density] ];
								if (a != 0xa1 || b != 0xa1 || c != 0xa1) {
									printf("double density DAM incomplete 0xa1 leader: {%02x %02x %02x %02x) (side %d track %d sector %d)\n", a, b, c, mark, side, i, j);

									// Repair
									idam[ (k-3) * stride[density] ] = 0xa1;
									idam[ (k-2) * stride[density] ] = 0xa1;
									idam[ (k-1) * stride[density] ] = 0xa1;
								}

								break;
							}
						}
						if (!found)
						{
							// This is not an error, just unusual
							printf("DAM not found (side %d track %d sector %d)\n", side, i, j);
						}
					}
#endif
				}
			}
			//printf("\n");
		}
	}
	//printf("\n %d sectors, %d SD %d DD\n\n", sector_count[0] + sector_count[1], sector_count[0], sector_count[1]);
	// In double-density, all IDAM marks should be $fe

	//printf("File size error: %d\n", file_err);

	tracklen = dmk_tracklen - DMK_INDEX_LENGTH;

	return 0;
}

const char* DMKDiskDrive::Save(FILE *fp)
{
	// Copy idamoffset table back
	for (int i = 0; i < info.tracks; i++) {
		for (int side = 0; side < info.sides; side++) {
			unsigned char* trackindex = TrackIndex(side, i);
			trackptr[side][i] = trackindex; // but starts with 128 bytes of index
			for (int j = 0; j < 64; j++) {
				trackindex[j*2] = idamoffset[side][i][j] & 0xff;
				trackindex[j*2 + 1] = (idamoffset[side][i][j] >> 8) & 0xff;
			}
		}
	}

	disk[0] = writable ? 0 : 0xff;

	int res = (int)fwrite(disk, 1, len, fp);

	if (res == 0)
	{
		return FileError();
	}

	if (res == len) {
		dirty = 0;
	}
	else
	{
		return "fwrite lost bytes";
	}

	return NULL;
}

int DMKDiskDrive::IsModified()
{
	return dirty;
}

void DMKDiskDrive::DebugDump()
{
	int t, side;

	// OK: state info.
	// You have current track number & offset (unless in-between seeking)
	// current byte: trackptr[i] + 128 + byteoffset;
	// 0 <= byteoffset < tracklen.
	// If you are single density then both bytes should be the same:
	// current byte: trackptr[i] + 128 + byteoffset (and at +1)
	// In theory, you need to look at the high bit to determine the density.
	// (unless single density is flagged--presumably--all tracks are single)

	// Dump sector headers of each track
	for (side = 0; side < info.sides; side++) {
		for (t = 0; t < info.tracks; t++) {
			for (int i = 0; (idamoffset[side][t][i] & 0x3fff) != 0; i++) {
				int off = (idamoffset[side][t][i] & 0x3fff);
				printf("%06lx:", trackptr[side][t] - disk + off);
				int j;
				for (j = 0; j < 5; j++) {
					printf(" %02x", trackptr[side][t][off + j]);
				}
				printf(" %02x%02x", trackptr[side][t][off + j], trackptr[side][t][off + j + 1]);
				printf("\n");
			}
			printf("\n");
		}
	}

	// Verify CRCs of each sector
	int head_err = 0;
	int sect_err = 0;
	int miss_err = 0;
	for (side = 0; side < info.sides; side++) {
		for (t = 0; t < info.tracks; t++) {
			for (int i = 0; (idamoffset[side][t][i] & 0x3fff) != 0; i++) {
				int off = (idamoffset[side][t][i] & 0x3fff);
				unsigned int crc = 0xffff;
				printf("head  ");
				for (int j = -3; j < 7; j++) {
					unsigned int data = trackptr[side][t][off + j];
					printf(" %02x", data);
					crc = update_crc(crc, data);
				}
				if (crc != 0) {
					head_err++;
				}
				printf(" : %04x%s\n", crc, crc == 0 ? "" : " !!!! CRC ERROR !!!");
				if (crc == 0) {
					off = find_data(trackptr[side][t], off + 7);
					if (off < 0) {
						miss_err++;
						printf("  !!! track data not found\n");
					} else {
						printf("  trk ");
						crc = 0xffff;
						int j;
						for (j = 0; j < 4; j++) {
							unsigned int data = trackptr[side][t][off + j];
							printf(" %02x", data);
							crc = update_crc(crc, data);
						}
						printf(" ...");
						// 256 should really be based on 'length' field
						for (; j < 4 + 256; j++) {
							unsigned int data = trackptr[side][t][off + j];
							crc = update_crc(crc, data);
						}
						for (; j < 4 + 256 + 2; j++) {
							unsigned int data = trackptr[side][t][off + j];
							printf(" %02x", data);
							crc = update_crc(crc, data);
						}

						if (crc != 0) {
							sect_err++;
						}
						printf(" : %04x%s\n", crc, crc == 0 ? "" : " !!!! CRC ERROR !!!");
					}
				}
			}
		}
	}
	printf("header CRC errors: %d\n", head_err);
	printf("sector CRC errors: %d\n", sect_err);
	printf("sector not found errors: %d\n", miss_err);
}

void DMKDiskDrive::GetDiskInfo(DiskInfo& res)
{
	res = info;
}

void DMKDiskDrive::WriteBackIDAMOffsets(int side, int track)
{
	//TODO: code in DMKDiskDrive::Save() should call this.
	unsigned char* trackindex = TrackIndex(side, track);
	trackptr[side][track] = trackindex; // but starts with 128 bytes of index
	for (int j = 0; j < 64; j++) {
		trackindex[j*2] = idamoffset[side][track][j] & 0xff;
		trackindex[j*2 + 1] = (idamoffset[side][track][j] >> 8) & 0xff;
	}
}

// Crude implementation of the reader part of the BlockInterface
void DMKDiskDrive::GetTrackInfo(TrackInfo& info, int side, int track)
{
	DMKTrack copy;

	WriteBackIDAMOffsets(side, track);
	copy.ReadDMKTrack(TrackIndex(side, track), dmk_tracklen, opt_single_density, m_config.data_rate);

	copy.GetTrackInfo(info);
}

void DMKDiskDrive::GetSectorInfo(SectorInfo& info, int side, int track, int sector)
{
	DMKTrack copy;

	WriteBackIDAMOffsets(side, track);
	copy.ReadDMKTrack(TrackIndex(side, track), dmk_tracklen, opt_single_density, m_config.data_rate);

	copy.GetSectorInfo(info, sector);
}

void DMKDiskDrive::GetSectorData(unsigned char* res, int side, int track, int sector)
{
	DMKTrack copy;

	WriteBackIDAMOffsets(side, track);
	copy.ReadDMKTrack(TrackIndex(side, track), dmk_tracklen, opt_single_density, m_config.data_rate);

	copy.GetSectorData(res, sector);
}

void DMKDiskDrive::GetDMKTrack(class DMKTrack *trk, int side, int track)
{
	WriteBackIDAMOffsets(side, track);
	trk->ReadDMKTrack(TrackIndex(side, track), dmk_tracklen, opt_single_density, m_config.data_rate);
}

void DMKDiskDrive::WriteByte(int density, int val, int clock)
{
	if (len == 0) { //HACKHACK
		//No diskette
		return;
	}

	dirty = 1;

	if (density == 0) {
		trackptr[current_side][current_track][radial_position + 128] = val;
		if (stride[density] == 2) {
			trackptr[current_side][current_track][(radial_position + 1) % tracklen + 128] = val;
		}
	}
	// Erase old marker
	for (int i = 0; i < 64; i++) {
		int off = (idamoffset[current_side][current_track][i] & 0x3fff);
		if (off == radial_position + 128) {
			RemoveIDAM(i);
			break;
		}
		// Check for altered position if we're doing FM (double-byte) writes
		if (density == 0 && stride[density] == 2)
		{
			if (off == radial_position + 127) {
				RemoveIDAM(i);
			}
		}
	}
	
	if (density != 0)
	{
		// Add new marker
		if (prev_clocked_a1 == 3 && val == 0xfe) {
			InsertIDAM(radial_position + 128, density);
		}

		trackptr[current_side][current_track][radial_position + 128] = val;

		if (val == 0xa1 && clock != 0xff) {
			prev_clocked_a1++;
		}
		else
		{
			prev_clocked_a1 = 0;
		}
	}
	else
	{
		if (val == 0xfe && clock == 0xc7)
		{
			InsertIDAM(radial_position + 128, density);
		}
	}
}

void DMKDiskDrive::RemoveIDAM(int index)
{
	int i;
	for (i = index; i < 64 - 1; i++) {
		idamoffset[current_side][current_track][i] = idamoffset[current_side][current_track][i + 1];
	}
	idamoffset[current_side][current_track][i] = 0;
}

void DMKDiskDrive::InsertIDAM(int offset, int density)
{
	int i, j;

	for (i = 0; i < 64; i++) {
		int curr_offset = (idamoffset[current_side][current_track][i] & 0x3fff);
		if (curr_offset > offset || curr_offset == 0) {
			break;
		}
	}
	if (i >= 64) {
		//assert(0); // Just ignore the problem.  Occurs when SuperUtilty writes ROM to Track 3
		return;
	}
	// High bit set only for double-density disks.
	if (density != 0)
	{
		offset |= 0x8000;
	}
	// shift up
	for (j = 63; j > i; j--) {
		idamoffset[current_side][current_track][j] = idamoffset[current_side][current_track][j - 1];
	}

	idamoffset[current_side][current_track][i] = offset;
}

int DMKDiskDrive::GetByte(int density, int* clock)
{
	int res;
	int i;

	if (len == 0) {//HACKHACK
		//No diskette
		return 0;
	}
	*clock = 0xff;	// not a valid MFM clock.
	res = trackptr[current_side][current_track][radial_position + 128];	// have to skip IDAM table.  Messy.

	// look for marker bytes
	if (density != 0) {
		// Double density
		// I suppose this could be pre-calculated or at least moved to AdvanceByte()
		if (res == 0xa1) {
			for (i = 0; i < 64; i++) {
				int off = (idamoffset[current_side][current_track][i] & 0x3fff);
				// The idam offset points at the 'marker' byte
				// but in double density the preceeding 3 A1 values are the special bytes.
				if (off != 0) {
					int delta = off - (radial_position + 128);
					if (delta >= 1 && delta <= 3) {
						*clock = 0x0a;
						break;
					}

					// The format does not mark DAM fields.  We'll use heuristics.
					delta = (radial_position + 128) - off;
					// Consider bytes past the IDAM data but within 43 bytes of the mark as markers.
					//ERROR: the timeout should strictly be 43 (from datasheet) but FROM THE END OF THE CRC
					// This calculation is from the marker itself.
					if (delta >= 7 && delta < DAM_MARKER_TIMEOUT) {
						// It might be safer to only allow if we see [xx A1 A1], [A1 xx A1] or [A1 A1 xx]
						*clock = 0x0a;
						break;
					}
				}
			}
		}
	}
	else {
		// Single density markers
		if (res >= 0xf8 && res <= 0xfe) {
			for (i = 0; i < 64; i++) {
				int off = (idamoffset[current_side][current_track][i] & 0x3fff);
				if (off != 0) {
//HACKHACK
//This fails for robtattk.dmk as the stream is offset by a byte
//The marker points at $145, position is $146.  Oddly, the two
// $fe bytes are at $145 and $146
					//ERROR checking pos-1+128 fixes GAMES1.DMK which also has odd offsets.
					// Fundamentally, the GetByte() code should use the IDAM markers to sync the reading offset
					// instead of assuming all single-density double-bytes start on an even boundary.
					// And the various "+ 128" constants should be #defined.  In this case,
					// the +128 converts a radial position to an IDAM position (which is offset
					// from the base of the 128-byte table).
					if (off == (radial_position + 128) || off == (radial_position - 1 + 128)) {
						*clock = 0xc7;
						break;
					}
					// For DAM fields, we will consider in-range byte after the offset.
					// This will return unnecessary clocks at time but the FDC will
					// have to ignore them.

					int delta = (radial_position + 128) - off;
					//ERROR: the timeout should strictly be 30 (from datasheet) but FROM THE END OF THE CRC
					// This calculation is from the marker itself.
					//ERROR: and since we're in single density, bytes are doubled, so delta /= 2.
					//ERROR: I probably should only consider the first Fx byte to appear in this window
					// to be a data mark.  Perhaps only a subset of Fx bytes should be allowed.
					int marker_limit = DAM_MARKER_TIMEOUT * stride[density];
					// Consider bytes past the IDAM data but within 43 bytes of the mark as markers.
					if (delta >= 7 && delta < marker_limit) {
						*clock = 0xc7;
						break;
					}
				}
			}
		}
	}

	return res;
}

int DMKDiskDrive::GetRawByte(int offset)
{
	return trackptr[current_side][current_track][(radial_position + offset) % tracklen + 128];
}

void DMKDiskDrive::AdvanceByte(int density)
{
	int inc = stride[density];

	// Safety--adjust radial position
	if (inc != 1) {
		if (radial_position % 2 != 0) {
			radial_position--;
		}
	}

	SetOffset(radial_position + inc);
}

void DMKDiskDrive::NotifyWriteTrack(int writeTrackOn)
{
	// DMK doesn't neeed notice of to format a track, but does use it
	// to enable the automatic track extension
	if (writeTrackOn)
	{
		CheckAllowTrackAppend();

		if (requested_side == 1 && info.sides == 1)
		{
			if (!(m_config.geometry_options & DRV_side_lock))
			{
				AppendBlankSide();
			}
		}
	}
}

DiskDrive* fd_create_nodiskette(const DriveConfiguration &config)
{
	return new NoDiskette(config);
}

DiskDrive* fd_create_dmk(const DriveConfiguration &config, const DiskImageMetadata* metadata)
{
	return new DMKDiskDrive(config, metadata);
}
