#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <ctype.h>

#ifdef _MSC_VER
#define stricmp _stricmp
#else
#define stricmp strcasecmp
#endif

#if defined(__APPLE__) || defined(__linux__)
#define _strnicmp strncasecmp
#endif

#include "fd_image.h"

// A re-implementation of DMK using DMKTrack which was
// never quite good enough to ship

DMKDiskDrive2::DMKDiskDrive2(const DriveConfiguration &config, const DiskImageMetadata* metadata) : DiskDrive(config, metadata)
{
	tracklen += DMK_INDEX_LENGTH;	// add extra for the IDAM offset table.
	len = 0x10 + info.tracks * info.sides * tracklen;
	int size_ok = ensure_disk_size(len);
	assert(size_ok);

	opt_single_density = config.density == DISK_FM ? 1 : 0;
	opt_ignore_density = 0;		//TODO: old flag, probably correct?
	opt_single_sided = (info.sides == 1);

	setup_stride();

	// Copy the options to the 16-byte header and set up a blank image (no sectors)
	memset(disk, 0x00, len);

	disk[0] = writable ? 0x00 : 0xff;

	disk[1] = info.tracks;
	disk[2] = (tracklen & 0xff);
	disk[3] = (tracklen >> 8) & 0xff;

	disk[4] = 0; // options
	if (info.sides == 1) disk[4] |= 0x10;
	if (opt_single_density) disk[4] |= 0x40;
	if (opt_ignore_density) disk[4] |= 0x80;
	options = disk[4];
	magic = 0;

	tracklen -= DMK_INDEX_LENGTH;	// Change to actual track length

	// Track constructor is blank, but set density and tracklen
	unsigned char data[16384];
	memset(data, 0x00, sizeof(data));
	for (int i = 0; i < info.tracks; i++) {
		for (int side = 0; side < info.sides; side++) {
			track[side][i].SetData(data, tracklen, opt_single_density);
		}
	}

	radial_position = 0;
}

void DMKDiskDrive2::setup_stride()
{
	stride[0] = opt_single_density ? 1 : 2;
	stride[1] = 1;
}

int DMKDiskDrive2::Load(unsigned char *data, int count)
{
	int size_ok = ensure_disk_size(count);
	assert(size_ok);

	//TODO We don't need the disk field for DMKDiskDrive2
	len = count;
	memcpy(disk, data, count);
	int i;

	writable = 1;
	if (disk[0] == 0xff) {
		writable = 0;
	}
	info.tracks = disk[1];
	int dmk_tracklen = disk[2] + (disk[3]<<8);
	
	options = disk[4];
	
	for (i = 5; i <= 0xb; i++) {
		if (disk[i] != 0) {
			printf("reserved byte at $%02x is not zero ($%02x)\n", i, disk[i]);
		}
	}

	magic = (disk[0xc] | (disk[0xd] << 8) | (disk[0xe] << 16) | (disk[0xf] << 24));
	
	printf("options=%02x magic=%08x\n", options, magic);
	if (magic == 0x12345678) {
		printf("real drive? not possible\n");
	}

	opt_ignore_density = !!(options & 0x80);
	opt_single_density = !!(options & 0x40);
	opt_single_sided = !!(options & 0x10);
	info.sides = opt_single_sided ? 1 : 2;
	printf("option detail: MIXED:%d SingleD:%d OneSided:%d\n", opt_ignore_density, opt_single_density, opt_single_sided);

	setup_stride();

	// Check on file size
	int expected_size = info.sides * info.tracks * dmk_tracklen + 0x10;	// 16 byte header
	//int file_err = len - expected_size;
	if (len < expected_size) {
		printf("**ERROR File appears truncated\n");
		int size_ok = ensure_disk_size(expected_size);
		assert(size_ok);
		len = expected_size;
	} else if (len > expected_size) {
		printf("**ERROR File is too long\n");
	}
	
	for (i = 0; i < info.tracks; i++) {
		for (int side = 0; side < info.sides; side++) {
			unsigned char* dmkdata = disk + (i * info.sides + side) * dmk_tracklen + 0x10; //TrackIndex(side, i);
			track[side][i].ReadDMKTrack(dmkdata, dmk_tracklen, opt_single_density, m_config.data_rate);
		}
	}

	//TODO: add blank tracks.

	tracklen = dmk_tracklen - DMK_INDEX_LENGTH;

	return 0;
}

const char* DMKDiskDrive2::Save(FILE *fp)
{
	// Copy idamoffset table and track data back
	for (int i = 0; i < info.tracks; i++) {
		for (int side = 0; side < info.sides; side++)
		{
			//ERROR TODO: check if track[side][i].IsModified()
			{
				int idam[DMK_MAX_IDAM];

				unsigned char* trackindex = disk + (i * info.sides + side) * (tracklen + DMK_INDEX_LENGTH) + 0x10; //TrackIndex(side, i);

				track[side][i].GetIdam(idam);
				track[side][i].GetData(trackindex + DMK_INDEX_LENGTH, tracklen);

				for (int j = 0; j < DMK_MAX_IDAM; j++)
				{
					int offset = idam[j];
					// Add back the 128 byte offset
					if (offset != 0)
					{
						offset = (((offset + DMK_INDEX_LENGTH) & 0x3fff) | (offset & 0xc000));
					}

					trackindex[j*2] = offset & 0xff;
					trackindex[j*2 + 1] = (offset >> 8) & 0xff;
				}
			}
			track[side][i].SetWritten();
		}
	}

	disk[0] = writable ? 0 : 0xff;

	int res = (int)fwrite(disk, 1, len, fp);

	if (res == 0)
	{
		return FileError();
	}

	if (res == len) {
		dirty = 0;
	}
	else
	{
		return "fwrite lost bytes";
	}

	return NULL;
}


int DMKDiskDrive2::IsModified()
{
	for (int s = 0; s < info.sides; s++)
	{
		for (int i = 0; i < info.tracks; i++)
		{
			if (track[s][i].IsModified())
			{
				return 1;
			}
		}
	}

	return 0;
}


int DMKDiskDrive2::GetByte(int density, int* clock)
{
	return track[current_side][current_track].GetByte(density, clock);
}

void DMKDiskDrive2::WriteByte(int density, int val, int clock)
{
	track[current_side][current_track].WriteByte(density, val, clock);
}

void DMKDiskDrive2::AdvanceByte(int density)
{
	track[current_side][current_track].AdvanceByte(density);

	radial_position = track[current_side][current_track].GetPosition();
}

void DMKDiskDrive2::SetTrack(int trk)
{
	int radial_pos = track[current_side][current_track].GetPosition();

	DiskDrive::SetTrack(trk);

	track[current_side][current_track].SetPosition(radial_pos);
}

void DMKDiskDrive2::SetSide(int side)
{
	int radial_pos = track[current_side][current_track].GetPosition();

	DiskDrive::SetSide(side);

	track[current_side][current_track].SetPosition(radial_pos);
}

int DMKDiskDrive2::GetRawByte(int offset)
{
	return track[current_side][current_track].GetRawByte(offset);
}

void DMKDiskDrive2::DebugDump()
{
}

void DMKDiskDrive2::GetDiskInfo(DiskInfo& res)
{
	res = info;
}

void DMKDiskDrive2::GetTrackInfo(TrackInfo& info, int side, int track_num)
{
	track[side][track_num].GetTrackInfo(info);
}

void DMKDiskDrive2::GetSectorInfo(SectorInfo& info, int side, int track_num, int sector)
{
	track[side][track_num].GetSectorInfo(info, sector);
}

void DMKDiskDrive2::GetSectorData(unsigned char* res, int side, int track_num, int sector)
{
	track[side][track_num].GetSectorData(res, sector);
}

void DMKDiskDrive2::GetDMKTrack(class DMKTrack *trk, int side, int track_num)
{
	*trk = track[side][track_num];
}

