#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <ctype.h>

#ifdef _MSC_VER
#define stricmp _stricmp
#else
#define stricmp strcasecmp
#endif

#if defined(__APPLE__) || defined(__linux__)
#define _strnicmp strncasecmp
#endif

#include "fd_image.h"

// DMKTrack implements a single track in DMK format

int DMKTrack::GetPosition()
{
	return radial_position;
}

void DMKTrack::SetPosition(int pos)
{
	radial_position = pos % tracklen;
}

void DMKTrack::RemoveIDAM(int index)
{
	int i;
	for (i = index; i < DMK_MAX_IDAM - 1; i++) {
		idamoffset[i] = idamoffset[i + 1];
	}
	idamoffset[i] = 0;
}

void DMKTrack::InsertIDAM(int offset, int density)
{
	int i, j;

	for (i = 0; i < DMK_MAX_IDAM; i++) {
		int curr_offset = (idamoffset[i] & 0x3fff);
		if (curr_offset > offset || curr_offset == 0) {
			break;
		}
	}
	// High bit set only for double-density disks.
	if (density != 0)
	{
		offset |= 0x8000;
	}
	if (i >= DMK_MAX_IDAM) {
		// failed--too many IDAM offsets to store.
		assert(0);
		return;
	}
	// shift up
	for (j = DMK_MAX_IDAM - 1; j > i; j--) {
		idamoffset[j] = idamoffset[j - 1];
	}

	idamoffset[i] = offset;
}

void DMKTrack::WriteByte(int density, int val, int clock)
{
	dirty = 1;

	if (density == 0) {
		trackdata[radial_position] = val;
		if (stride[density] == 2) {
			trackdata[(radial_position + 1) % tracklen] = val;
		}
	}
	// Erase old marker
	for (int i = 0; i < DMK_MAX_IDAM; i++) {
		int off = (idamoffset[i] & 0x3fff);
		if (off == radial_position) {
			RemoveIDAM(i);
			break;
		}
		// Check for altered position if we're doing FM (double-byte) writes
		if (density == 0 && stride[density] == 2) {
			if (off == radial_position - 1)
			{
				RemoveIDAM(i);
			}
		}
	}
	
	if (density != 0)
	{
		// Add new marker
		if (prev_clocked_a1 == 3 && val == 0xfe) {
			InsertIDAM(radial_position, density);
		}

		trackdata[radial_position] = val;

		if (val == 0xa1 && clock != 0xff) {
			prev_clocked_a1++;
		}
		else
		{
			prev_clocked_a1 = 0;
		}
	}
	else
	{
		if (val == 0xfe && clock == 0xc7)
		{
			InsertIDAM(radial_position, density);
		}
	}
}

int DMKTrack::GetByte(int density, int* clock)
{
	int res;
	int i;

	*clock = 0xff;	// not a valid MFM clock.
	res = trackdata[radial_position];

	// look for marker bytes
	if (density != 0) {
		// Double density
		// I suppose this could be pre-calculated or at least moved to AdvanceByte()
		if (res == 0xa1) {
			for (i = 0; i < DMK_MAX_IDAM; i++) {
				int off = (idamoffset[i] & 0x3fff);
				// The idam offset points at the 'marker' byte
				// but in double density the preceeding 3 A1 values are the special bytes.
				if (off != 0) {
					int delta = off - radial_position;
					if (delta >= 1 && delta <= 3) {
						*clock = 0x0a;
						break;
					}

					// The format does not mark DAM fields.  We'll use heuristics.
					delta = radial_position - off;
					// Consider bytes past the IDAM data but within 43 bytes of the mark as markers.
					//ERROR: the timeout should strictly be 43 (from datasheet) but FROM THE END OF THE CRC
					// This calculation is from the marker itself.
					if (delta >= 7 && delta < DAM_MARKER_TIMEOUT) {
						// It might be safer to only allow if we see [xx A1 A1], [A1 xx A1] or [A1 A1 xx]
						*clock = 0x0a;
						break;
					}
				}
			}
		}
	}
	else {
		// Single density markers
		if (res >= 0xf8 && res <= 0xfe) {
			for (i = 0; i < DMK_MAX_IDAM; i++) {
				int off = (idamoffset[i] & 0x3fff);
				if (off != 0) {
//HACKHACK
//This fails for robtattk.dmk as the stream is offset by a byte
//The marker points at $145, position is $146.  Oddly, the two
// $fe bytes are at $145 and $146
					//ERROR checking pos-1+128 fixes GAMES1.DMK which also has odd offsets.
					// Fundamentally, the GetByte() code should use the IDAM markers to sync the reading offset
					// instead of assuming all single-density double-bytes start on an even boundary.
					// And the various "+ 128" constants should be #defined.  In this case,
					// the +128 converts a radial position to an IDAM position (which is offset
					// from the base of the 128-byte table).
					if (off == radial_position || off == radial_position - 1) {
						*clock = 0xc7;
						break;
					}
					// For DAM fields, we will consider in-range byte after the offset.
					// This will return unnecessary clocks at time but the FDC will
					// have to ignore them.

					int delta = radial_position - off;
					//ERROR: the timeout should strictly be 30 (from datasheet) but FROM THE END OF THE CRC
					// This calculation is from the marker itself.
					//ERROR: and since we're in single density, bytes are doubled, so delta /= 2.
					//ERROR: I probably should only consider the first Fx byte to appear in this window
					// to be a data mark.  Perhaps only a subset of Fx bytes should be allowed.
					int marker_limit = DAM_MARKER_TIMEOUT * stride[density];
					// Consider bytes past the IDAM data but within 43 bytes of the mark as markers.
					if (delta >= 7 && delta < marker_limit) {
						*clock = 0xc7;
						break;
					}
				}
			}
		}
	}

	return res;
}

void DMKTrack::AdvanceByte(int density)
{
	int inc = stride[density];

	// Safety--adjust radial position
	if (inc != 1) {
		if (radial_position % 2 != 0) {
			radial_position--;
		}
	}

	radial_position = (radial_position + inc) % tracklen;
}

int DMKTrack::GetRawByte(int offset)
{
	return trackdata[(radial_position + offset) % tracklen];
}

void DMKTrack::Reset()
{
	dirty = 0;
	prev_clocked_a1 = 0;
	radial_position = 0;
}

void DMKTrack::SetIdam(int* idam)
{
	for (int i = 0; i < DMK_MAX_IDAM; i++)
	{
		idamoffset[i] = idam[i];
	}

	Reset();
}

void DMKTrack::SetData(unsigned char* data, int len, int single_d)
{
	single_density = single_d;
	tracklen = len;
	memcpy(trackdata, data, tracklen);

	setup_stride();

	Reset();
}

void DMKTrack::SetMinimumTracklen(int len)
{
	if (len > tracklen)
	{
		// Pad with zeros
		for (int i = tracklen; i < len; i++)
		{
			trackdata[i] = 0x00;
		}

		tracklen = len;
	}
}

int DMKTrack::GetTracklenAndStride(int *_stride)
{
	if (_stride)
	{
		_stride[0] = stride[0];
		_stride[1] = stride[1];
	}

	return tracklen;
}

void DMKTrack::SetDataRate(int rate)
{
	data_rate = rate;
}

void DMKTrack::GetIdam(int* idam)
{
	for (int i = 0; i < DMK_MAX_IDAM; i++)
	{
		idam[i] = idamoffset[i];
	}
}

void DMKTrack::GetData(unsigned char* dst, int len)
{
	if (len > tracklen)
	{
		len = tracklen;
	}

	memcpy(dst, trackdata, len);
}

void DMKTrack::GetTrackInfo(TrackInfo& info)
{
// idamoffset[] quickly tells you about the existence of tracks
// and the density.  Deleted or other special things are trickier
	info.sectors = 0;
	for (int i = 0; i < DMK_MAX_IDAM; i++)
	{
		//ERROR: possibly should consider delete
		if ((idamoffset[i] & 0x3fff) != 0)
		{
			// M2 LS-DOS format and backup hack.
			// If an IDAM marker refers to a DAM which is past the track, ignore rest of IDAM table.
			// This occurs when garbage accidentally gets into the write track data.
			// TODO: A proper fix might allow it but mark data as SECTOR_unavail
			// or might discard the IDAM for having an invalid CRC.

			int off = (idamoffset[i] & 0x3fff);
			int density = !!(idamoffset[i] & 0x8000);
			int delta = stride[density];
			int timeout = density ? DAM_MARKER_TIMEOUT_MFM : DAM_MARKER_TIMEOUT_FM;
			if (off + delta * 7 + timeout >= tracklen)	// delta*7 to skip past CRC
			{
				break;
			}
			info.sectors++;
		}
	}
}

int DMKTrack::GetSectorLength(int sector)
{
	int density = !!(idamoffset[sector] & 0x8000);
	int off = idamoffset[sector] & 0x3fff;
	int delta = stride[density];

	off += delta * 4;
	int leng = trackdata[off]; off += delta;
	static int sector_lengths[4] = { 128, 256, 512, 1024 };
	leng = sector_lengths[leng & 0x3];

	return leng;
}

void DMKTrack::GetSectorInfo(SectorInfo& info, int sector)
{
	int density = !!(idamoffset[sector] & 0x8000);
	int off = idamoffset[sector] & 0x3fff;
	int delta = stride[density];

	/*[unused] int mark = trackdata[off];*/ off += delta;
	int trck = trackdata[off]; off += delta;
	int side = trackdata[off]; off += delta;
	int sect = trackdata[off]; off += delta;
	int leng = GetSectorLength(sector); off += delta;

	info.si_flags = 0;
	info.si_track = trck;
	info.si_side = side;
	info.si_sector = sect;
	info.si_length = leng;
	if (density) info.si_flags |= SECTOR_dden;

	info.set_rate(data_rate / 1000);

	//ERROR _more_ copied code, unfortunately FindDAM() uses GetSectorInfo...
	// FindDAM() no longer uses GetSectorInfo()
	off += delta * 7;	// skip past CRC
	int dam_offset = 0;

	// TODO: what if CRC is invalid for the IDAM? Probably we should completely ignore this entry.
	// ERROR NOT TRUE if the disk is only single density but then 'density' should be zero.
	int timeout = density ? DAM_MARKER_TIMEOUT_MFM : DAM_MARKER_TIMEOUT_FM;

	for (int i = 0; i < timeout; i++)
	{
		if (trackdata[off] >= 0xf8 && trackdata[off] <= 0xfb)
		{
			dam_offset = off;
			break;
		}
		off += delta;
	}

	if (dam_offset > 0)
	{
		int dam_mark = trackdata[dam_offset];
		if (density == 0 && dam_mark == 0xfa)		//TODO: also compare 0xf8 and/or SECTOR_damhigh
		{
			info.si_flags |= SECTOR_delete;
		}

		// Check data CRC
		unsigned int crc = 0xffff;
		if (density)
		{
			crc = update_crc(crc, 0xa1);
			crc = update_crc(crc, 0xa1);
			crc = update_crc(crc, 0xa1);
		}
		crc = update_crc(crc, dam_mark);

		int off = dam_offset;
		for (int j = 0; j < info.si_length; j++)
		{
			off += delta;
			crc = update_crc(crc, trackdata[off]);
		}
		off += delta;
		crc = update_crc(crc, trackdata[off]);
		off += delta;
		crc = update_crc(crc, trackdata[off]);
		if (crc != 0)
		{
			info.si_flags |= SECTOR_err;
		}
	}
	else
	{
		info.si_flags |= SECTOR_unavail;
	}

	//ERROR still lots of TODO here

	// *delta
	// +1 track
	// +2 side			masked?
	// +3 sector
	// +4 length		sector_lengths[z.reg_DSR & 0x3];
	// +4,5 CRC
	// Note: bad CRC is error.
	// Not sure if we should check DAM CRC, too.
	//
				//for (int j = -3; j < 7; j++) {
				//	unsigned int data = trackptr[side][t][off + j];
				//	printf(" %02x", data);
				//	crc = update_crc(crc, data);
				//}
/*
    SECTOR_dden = 1<<0,
    SECTOR_err = 1<<1,
	SECTOR_delete = 1<<2,
	SECTOR_unavail = 1<<3,
	SECTOR_rate_bit = 8,
	SECTOR_rate_mask = 255 << SECTOR_rate_bit,
	SECTOR_rate_scale = 25,
*/
}

int DMKTrack::FindDAM(int sector, int& delta, int& idam_offset, int& idam_len, int& dam_offset, int& dam_len)
{
	int density = !!(idamoffset[sector] & 0x8000);
	int off = idamoffset[sector] & 0x3fff;

	delta = stride[density];

	idam_offset = off;
	idam_len = 7;
	off += delta * idam_len;	// skip past CRC
	dam_offset = dam_len = 0;

	if (idam_offset == 0)
	{
		// maybe should be "return density" to be consistent?
		return -1;
	}

	int sector_length = GetSectorLength(sector);	// get length

	// ERROR NOT TRUE if the disk is only single density but then 'density' should be zero.
	int timeout = density ? DAM_MARKER_TIMEOUT_MFM : DAM_MARKER_TIMEOUT_FM;

	for (int i = 0; i < timeout; i++)
	{
		if (trackdata[off] >= 0xf8 && trackdata[off] <= 0xfb)
		{
			dam_offset = off;
			dam_len = sector_length + 3;	// include header and CRC
			if (density != 0)
			{
				dam_offset -= 3 * delta; // leading a1 a1 a1
				dam_len += 3 * delta;
			}
			return density;
		}
		off += delta;
	}

	//TODO: this assert happens on 'blank' (or missing/invalid) tracks
	// AND on: CosmicFighter/download/cosftrv3/Cosmic Fighter v3.0 (1981)(Big Five Software)[M13][Dual Boot][Set 1].dmk
	// Nov 2019: Removed the assert to allow track dumps of the disk.
	//assert(0);	// Same assert() as in GetSectorData()

	// dam_offset of 0 will indicate DAM not found.
	return density;
}

void DMKTrack::GetSectorData(unsigned char* res, int sector)
{
	//TODO: USE "FindDAM()" here.

	int density = !!(idamoffset[sector] & 0x8000);
	int off = idamoffset[sector] & 0x3fff;
	int delta = stride[density];

	off += delta * 7;	// skip past CRC

	SectorInfo info;
	GetSectorInfo(info, sector);	// get length

	// ERROR NOT TRUE if the disk is only single density but then 'density' should be zero.
	int timeout = density ? DAM_MARKER_TIMEOUT_MFM : DAM_MARKER_TIMEOUT_FM;

	for (int i = 0; i < timeout; i++)
	{
		if (trackdata[off] >= 0xf8 && trackdata[off] <= 0xfb)
		{
			for (int j = 0; j < info.si_length; j++)
			{
				off += delta;
				res[j] = trackdata[off];
			}

			return;
		}
		off += delta;
	}

	//TODO: this assert happens on 'blank' (or missing/invalid) tracks
	// Usually avoided as GetSectorData() should set SECTOR_unavail
	memset(res, 0xfe, info.si_length);

#if defined(_DEBUG) && defined(_WIN32)
	{
		extern void ods(char *str);

		int off = idamoffset[sector] & 0x3fff;

		char buf[256];
		sprintf(buf, "Did not find sector %d with offset 0x%04x and density %d (len=%d)\n", sector, off, density, info.si_length);
		ods(buf);
		// Output 128 bytes at initial offset
		char *p = buf;
		*p = 0;
		for (int i = 0; i < 128; i++) {
			if (i % 16 == 0 && buf[0]) {
				sprintf(p, "\n");
				ods(buf);
				p = buf;
				*p = 0;
			}
			sprintf(p, "%02x ", trackdata[off + i]);
			p += strlen(p);
		}
		if (buf[0]) {
			sprintf(p, "\n");
			ods(buf);
		}

		ods("\n");
	}
#endif
}

void DMKTrack::setup_stride()
{
	stride[0] = single_density ? 1 : 2;
	stride[1] = 1;
}

void DMKTrack::ReadDMKTrack(
		unsigned char* dmkdata,
		int dmkdata_len,
		int opt_single_density,
		int _data_rate
		)
{
	// Set up track pointers and sector (IDAM) offsets
	int idamoffset[DMK_MAX_IDAM];
	unsigned char* trackindex = dmkdata;

	for (int j = 0; j < DMK_MAX_IDAM; j++) {
		idamoffset[j] = (trackindex[j*2] | (trackindex[j*2+1] << 8));
		// high bit means, double-density, next highest is ignored
		int off = (idamoffset[j] & 0x3fff);
		if (off >= dmkdata_len) {
			printf("***ERROR IDAM offset %04x too large\n", off);
		}
		if (off != 0) {
			// Must take away the 128 offset
			idamoffset[j] = (((off - DMK_INDEX_LENGTH) & 0x3fff) | (idamoffset[j] & 0xc000));
		}
		else
		{
			idamoffset[j] = 0;
		}
	}
	unsigned char* trackdata = dmkdata + DMK_INDEX_LENGTH;
	SetIdam(idamoffset);
	SetData(trackdata, dmkdata_len - DMK_INDEX_LENGTH, opt_single_density);
	data_rate = _data_rate;
}

void DMKTrack::Dump(void (*output)(void *ctx, const char* text), int tracknum, void* ctx)
{
	char f[256];

	// Check if density is mixed or not.
	{
		int fm = 0, mfm = 0;
		for (int sector = 0; sector < DMK_MAX_IDAM; sector++)
		{
			int density = !!(idamoffset[sector] & 0x8000);
			int off = idamoffset[sector] & 0x3fff;

			if (off == 0) {
				break;
			}
			if (density) mfm++; else fm++;
		}

		//ERROR TODO doesn't account for the 'inherent' density flag
		const char* dmark = "mixed";
		if (fm == 0) dmark = "MFM";
		else if (mfm == 0) dmark = "FM";

		sprintf(f, "track %02x %s len=0x%x\n", tracknum, dmark, tracklen); output(ctx, f);
	}

	int delta = 1; //TODO: based on default disk density
	int offset = 0;
	for (int sector = 0; sector < DMK_MAX_IDAM && offset < tracklen; sector++)
	{
		int new_delta, idam_offset, idam_len, dam_offset, dam_len;
		int density = FindDAM(sector, new_delta, idam_offset, idam_len, dam_offset, dam_len);

		if (idam_offset == 0)
		{
			// Tail-out of track
			dump_segment(ctx, output, NULL, sector, trackdata, offset, tracklen, delta, 0, -1);
			break;
		}

		// Lead-in of track
		int mark_extra = (density == 0 ? 0 : 3);
		dump_segment(ctx, output, NULL, sector, trackdata, offset, idam_offset - mark_extra, delta, 0, -1);

		offset = idam_offset - mark_extra;
		delta = stride[density];

		int idam_and_gap_end = offset + (idam_len + mark_extra) * delta;

		if (dam_offset != 0)
		{
			idam_and_gap_end = dam_offset;
		}
		dump_segment(ctx, output, "idam", sector, trackdata, offset, idam_and_gap_end, delta, 1 + mark_extra, idam_len + mark_extra - 2);
		offset = idam_and_gap_end;
		if (dam_offset != 0)
		{
			int dam_and_gap_end = offset + dam_len * delta;
			if (sector + 1 < DMK_MAX_IDAM && offset < tracklen)
			{
				int off2 = idamoffset[sector + 1] & 0x3fff;
				if (off2 != 0)
				{
					dam_and_gap_end = off2;
					if (density != 0)
					{
						dam_and_gap_end -= 3;
					}
				}
			}

			int mark_len = density == 0 ? 1 : 4;
			dump_segment(ctx, output, "dam ", sector, trackdata, offset, dam_and_gap_end, delta, mark_len, dam_len - 2);
			offset = dam_and_gap_end;
		}
	}
}

void DMKTrack::Dump(FILE* f, int tracknum)
{
	Dump(file_out, tracknum, f);
}

char* DMKTrack::Dump(int tracknum)
{
	size_t len = 0;
	Dump(char_size, tracknum, &len);
	char *res = new char[len + 1];
	res[len] = 0;
	char *ptr = res;
	Dump(char_out, tracknum, &ptr);

	return res;
}

