#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <ctype.h>

#ifdef _MSC_VER
#define stricmp _stricmp
#else
#define stricmp strcasecmp
#endif

#if defined(__APPLE__) || defined(__linux__)
#define _strnicmp strncasecmp
#endif

#include "fd_image.h"

// HFE flux format

// WIP
// HFE isn't really a block format, but we'll start with decoding it into blocks.

// A lot of the MFM code could be shared from HardDrive.cpp

#define HFE_Encoding_MFM (0)
#define HFE_Encoding_FM (2)

#define MFM_MARK_A1 (0x4489)

#define FM_CLOCK (1)	// 0C
#define FM_DATA_ONE (1)	// 0D
#define FM_MARK_FE (0x55111554)
#define FM_MARK_FB (0x55111455)
#define FM_MARK_FA (0x55111454)
#define FM_MARK_F8 (0x55111444)
#define FM_MARK_F9 (0x55111445)

static void next_mfm(const unsigned char *buf, int &bitpos, unsigned short &mfm);
static void sep_mfm(unsigned short mfm, unsigned char &data, unsigned char &clock);
static unsigned char read_byte_mfm(const unsigned char *buf, int &bitpos, unsigned short &mfm);
static void write_next_mfm(unsigned char *buf, int &bitpos, unsigned short &mfm);
static unsigned int mfm_encode(int prev_bit, int data, int skip_clock);
static unsigned char mfm_write(unsigned char* buf, int& bitpos, unsigned char data, unsigned char skip, unsigned char prev);
static void idam_mfm(unsigned int id, const unsigned char* buf, int& bitpos, unsigned short &mfm, SectorInfo& info, void (*output)(void *ctx, const char* text), void* ctx);
static int dam_mfm(unsigned int mark, const unsigned char* buf, int& bitpos, unsigned short &mfm, SectorInfo& info, unsigned char* dst, void (*output)(void *ctx, const char* text), void* ctx);
static unsigned int read_dam_data_mfm(int len, const unsigned char* buf, int& bitpos, unsigned short &mfm, unsigned int crc, unsigned char* dst);
static void mfm_align_mark(HFETrack &trk, int &bitpos);

static void next_fm(const unsigned char *buf, int &bitpos, unsigned int &fm);
static void sep_fm(unsigned int fm, unsigned char &data, unsigned char &clock);
static unsigned char grab2(unsigned int &fm);
static unsigned char read_byte_fm(const unsigned char *buf, int &bitpos, unsigned int &fm);
static void write_next_fm(unsigned char *buf, int &bitpos, unsigned int &fm);
static unsigned int fm_encode(int data, int clock);
static void fm_write(unsigned char* buf, int& bitpos, unsigned char data, unsigned char clock);
static void idam_fm(const unsigned char* buf, int& bitpos, unsigned int &fm, SectorInfo& info, void (*output)(void *ctx, const char* text), void* ctx);
static int dam_fm(const unsigned char* buf, int& bitpos, unsigned int &fm, SectorInfo& info, unsigned char* dst, void (*output)(void *ctx, const char* text), void* ctx);
static unsigned int read_dam_data_fm(int len, const unsigned char* buf, int& bitpos, unsigned int &fm, unsigned int crc, unsigned char* dst);
static void fm_align_mark(HFETrack &trk, int &bitpos);

static void flux_gap(int begin, int end, unsigned char* data, int encoding, void (*output)(void *ctx, const char* text), void* ctx);

HFEDiskDrive::HFEDiskDrive(const DriveConfiguration &config, const DiskImageMetadata* metadata)
	: DiskDrive(config, metadata)
{
	// Header and 1 block of track list (enough for 128 tracks)
	len = HFE_BLOCK_SIZE * 2;
	int size_ok = ensure_disk_size(len);
	assert(size_ok);

	// Header and track list 0xff filled by default

	memset(disk, 0xff, len);
	memcpy(&disk[0], HFE_MAGIC, strlen(HFE_MAGIC));

	// Blank header
	strcpy((char*)magic, HFE_MAGIC);
	version = 0;
	info.tracks = 0;
	info.sides = config.sides;
	encoding = HFE_Encoding_MFM;
	rate = 250;	// TODO: from metadata
	rpm = 300;	// TODO: metadata
	uimode = 7; // GENERIC_SHUGGART_DD_FLOPPYMODE
	// 0x11 unused
	track_list = 1;	// block #1
	writable = 0xff;
	single_step = 0xff;
	
	alt[0].use_alt = 0xff;
	alt[0].encode = 0xff;
	alt[1].use_alt = 0xff;
	alt[1].encode = 0xff;

	// Append one empty track
	AppendBlankTrack();

	// Keep in the in-memory header looking consistent
	update_header();

	stride[0] = 2;
	stride[1] = 1;

	radial_position = 0;
	radial_bit_position = 0;

	set_track(0, 0, -1);
    
	dirty = 0;
}

int HFEDiskDrive::Load(unsigned char *data, int count)
{
	// TODO: how can we 'throw' an error that Version 3 not supported?
	if (memcmp(&data[0], HFEv3_MAGIC, strlen(HFEv3_MAGIC)) == 0) {
		return 0;
	}

	int size_ok = ensure_disk_size(count);
	assert(size_ok);

	len = count;
	memcpy(disk, data, count);

	memcpy(magic, &disk[0x00], 8);
	magic[8] = 0;
	version = disk[0x08];
	info.tracks = disk[0x09];
	info.sides = disk[0x0a];
	encoding = disk[0x0b];
	rate = disk[0x0c] + disk[0x0d]*256;
	rpm = disk[0x0e] + disk[0x0f]*256;
	uimode = disk[0x10];	//TODO: expect 0x07 GENERIC_SHUGGART_DD_FLOPPYMODE, WARN on others (e.g. like an "AMIGA_DD" disk)
	// 0x11 unused
	track_list = disk[0x12] + disk[0x13]*256;
	writable = disk[0x14];
	single_step = disk[0x15];
	
	alt[0].use_alt = disk[0x16];
	alt[0].encode = disk[0x17];
	alt[1].use_alt = disk[0x18];
	alt[1].encode = disk[0x19];

	// TODO: file length checks
	for (int i = 0; i < info.tracks; i++) {
		track_pos[i].block = disk[track_list*512 + 4*i + 0] + disk[track_list*512 + 4*i + 1]*256;
		track_pos[i].len = disk[track_list*512 + 4*i + 2] + disk[track_list*512 + 4*i + 3]*256;
	}

	set_track(0, 0, -1);
    
	// TODO: other initialization
	dirty = 0;

	return 0; // LENGTH
}

void HFEDiskDrive::update_header()
{
	disk[0x08] = version;
	disk[0x09] = info.tracks;
	disk[0x0a] = info.sides;
	disk[0x0b] = encoding;
	disk[0x0c] = (rate & 0xff);
	disk[0x0d] = ((rate >> 8) & 0xff);
	disk[0x0e] = (rpm & 0xff);
	disk[0x0f] = ((rpm >> 8) & 0xff);
	disk[0x10] = uimode;
	// 0x11 unused
	disk[0x12] = (track_list & 0xff);
	disk[0x13] = ((track_list >> 8) & 0xff);
	disk[0x14] = writable;
	disk[0x15] = single_step;
	
	disk[0x16] = alt[0].use_alt;
	disk[0x17] = alt[0].encode;
	disk[0x18] = alt[1].use_alt;
	disk[0x19] = alt[1].encode;

	// TODO: file length checks
	for (int i = 0; i < info.tracks; i++) {
		int block = track_pos[i].block;
		disk[track_list*512 + 4*i + 0] = (block & 0xff);
		disk[track_list*512 + 4*i + 1] = ((block >> 8) & 0xff);

		int length = track_pos[i].len;
		disk[track_list*512 + 4*i + 2] = (length & 0xff);
		disk[track_list*512 + 4*i + 3] = ((length >> 8) & 0xff);
	}
}

const char* HFEDiskDrive::Save(FILE *fp)
{
	flush_track(-1); // ?

	// Write back header and track list
	update_header();

	int res = (int)fwrite(disk, 1, len, fp);

	if (res == 0)
	{
		return FileError();
	}

	if (res == len) {
		dirty = 0;
	}
	else
	{
		return "fwrite lost bytes";
	}

	return NULL; // No error
}

void HFEDiskDrive::go_track(int SIDE, int TRACK, int density)
{
	if (current_side != SIDE || current_track != TRACK) {
		//TODO: no error checking
		set_track(SIDE, TRACK, density);
	}
}

void HFEDiskDrive::GetDiskInfo(DiskInfo& _info)
{
	_info = info;
}

void HFEDiskDrive::GetTrackInfo(TrackInfo& info, int side, int track)
{
	// TODO: block interface STOP MESSING UP CURRENT TRACK
	go_track(side, track, -1);

	info.sectors = m_currentTrack.m_num_sector;
}

void HFEDiskDrive::GetSectorInfo(SectorInfo& info, int side, int track, int sector)
{
	go_track(side, track, -1);

	info = m_currentTrack.m_info[sector].m_info;
}

void HFEDiskDrive::GetSectorData(unsigned char* res, int side, int track, int sector)
{
	go_track(side, track, -1);

	HFETrackSectorInfo *info = &m_currentTrack.m_info[sector];

	memcpy(res, &m_currentTrack.m_data[info->m_data_offset], info->m_info.si_length);
}

int HFEDiskDrive::GetByte(int density, int* clock)
{
	if (m_currentTrack.m_state != HFETRACK_Read) {
		write_track(current_side, current_track, m_currentTrack);
		m_currentTrack.m_state = HFETRACK_Read;
	}

	// Scans in place (no precomputed marks)
	unsigned char dta, clk;
	if (density == 1) {
		int bitpos = radial_bit_position;
		unsigned short mfm = 0;
		read_byte_mfm(m_currentTrack.m_flux_data, bitpos, mfm);
		sep_mfm(mfm, dta, clk);
	}
	else {
		int bitpos = radial_bit_position;
		unsigned int fm = 0;
		read_byte_fm(m_currentTrack.m_flux_data, bitpos, fm);
		sep_fm(fm, dta, clk);
	}
	*clock = clk;

	return dta;
}

void HFEDiskDrive::AlignToMark(int density)
{
	if (density == 1) {
		mfm_align_mark(m_currentTrack, radial_bit_position);
	}
	else {
		fm_align_mark(m_currentTrack, radial_bit_position);
	}
}

void HFEDiskDrive::WriteByte(int density, int val, int clock)
{
	if (m_currentTrack.m_state != HFETRACK_Write) {
		m_currentTrack.m_prevWrite = 0;

		m_currentTrack.m_state = HFETRACK_Write;
		dirty = 1;
	}

	int bitpos = radial_bit_position;
	if (density == 1)
	{
		// HACK: clock values aren't correct from WD17xx
		unsigned char skip = 0;
		if (val == 0xa1 && (clock == 0xa0 || clock == 0x0a)) {
			skip = 0x04;
		}
		m_currentTrack.m_prevWrite = mfm_write(m_currentTrack.m_flux_data, bitpos, val, skip, m_currentTrack.m_prevWrite);
	}
	else {
		fm_write(m_currentTrack.m_flux_data, bitpos, val, clock);
	}
}

void HFEDiskDrive::AdvanceByte(int density)
{
	radial_bit_position += stride[density] * 16;
	if (radial_bit_position >= m_currentTrack.m_flux_len) {
		radial_bit_position = 0;
	}
	// radial_position used for index hole, etc.
	radial_position = radial_bit_position / 16;
}

void HFEDiskDrive::NotifyWriteTrack(int writeTrackOn)
{
	if (writeTrackOn)
	{
		CheckAllowTrackAppend();

		if (requested_side == 1 && info.sides == 1)
		{
			if (!(m_config.geometry_options & DRV_side_lock))
			{
				AppendBlankSide();
			}
		}
	}
}

static int hfe_block_count(int bytes)
{
	return (bytes + HFE_BLOCK_SIZE - 1) / HFE_BLOCK_SIZE;
}

int HFEDiskDrive::AppendBlankTrack()
{
	if (info.tracks >= 255)
	{
		return 0;
	}

	// Pad to block size (unlikely, but could occur if a file was truncated a little)
	if (len % HFE_BLOCK_SIZE != 0)
	{
		int extra = hfe_block_count(len) * HFE_BLOCK_SIZE - len;
		memset(&disk[len], 0, extra);	// zero new track
		len += extra;
	}

	// Find the track closest to the track list block
	// Find the minimum track block offset, maximum track_len
	// (unless len is really small)
	int min_block = len / HFE_BLOCK_SIZE;
	int max_track_len = 0;
	for (int i = 0; i < info.tracks; i++)
	{
		if (track_pos[i].block > track_list && track_pos[i].block < min_block) {
			min_block = track_pos[i].block;
		}
		if (i == 0 || track_pos[i].len > max_track_len) {
			max_track_len = track_pos[i].len;
		}
	}

	// Check if adding to the track list will overflow into the lowest allocated block
	int track_list_blocks = hfe_block_count((info.tracks + 1)*4);
	if (track_list + track_list_blocks > min_block) {
		if (!ensure_disk_size(len + HFE_BLOCK_SIZE)) {
			// Not enough space
			return 0;
		}
		int offset = (track_list + track_list_blocks - 1) * HFE_BLOCK_SIZE;
		memmove(&disk[offset + HFE_BLOCK_SIZE], &disk[offset], len - offset);
		// erase the new block
		memset(&disk[offset], 0xff, HFE_BLOCK_SIZE);
		// Adjust track positions
		for (int i = 0; i < info.tracks; i++) {
			track_pos[i].block++;
		}

		len += HFE_BLOCK_SIZE;
	}

	if (max_track_len == 0) {
		// make one up from the metadata
		int rpm = m_config.rpm;
		if (rpm == 0)
			rpm = 300;

		max_track_len = m_config.data_rate / 8 * 60 / rpm;
		// Double-sided and each byte takes two for clocks
		max_track_len *= 2 * 2;

		// FM requires double flux bits / data bit
		if (m_config.density == DISK_FM) {
			max_track_len *= 2;
		}

		// Typical is 25000 although (rounding to $6200 or $31 blocks)
	}
	int max_track_blocks = hfe_block_count(max_track_len);

	int extra_len = max_track_blocks * HFE_BLOCK_SIZE;

	if (!ensure_disk_size(len + extra_len)) {
		// No more space
		return 0;
	}

	// IGNORING the block list allocation
	// TODO: movemem() or whatever, then reduce

	int trk = info.tracks;
	info.tracks++;
	track_pos[trk].block = len / HFE_BLOCK_SIZE;
	track_pos[trk].len = max_track_len;
	memset(&disk[len], 0xaa, extra_len);	// 0xaa is "uniform" for FM/MFM

	len += extra_len;

	return 1;
}

int HFEDiskDrive::AppendBlankSide()
{
	// HFE always has space for double-sided
	if (info.sides < 2)
	{
		info.sides = 2;
	}

	return 1;
}

int HFEDiskDrive::GetRawByte(int offset)
{
	return 0;
}

int HFEDiskDrive::IsModified()
{
	return dirty;
}

void HFEDiskDrive::DebugDump()
{
}

void HFEDiskDrive::SetTrack(int track)
{
	int old_track = current_track;

	// BlockDrive avoid this because their track programs write back
	// but what do they do with format?
	// BlockDrive and others don't need this flush because the track programs
	// write back on the fly.
	flush_track(-1);

	// Note: base class will AppendBlankTrack() as necessary.
	// TODO: range checking.
	DiskDrive::SetTrack(track);

	if (old_track != current_track)
	{
		set_track(current_side, current_track, -1);
	}
}

void HFEDiskDrive::SetSide(int side)
{
	int old_side = current_side;

	DiskDrive::SetSide(side);

	if (old_side != current_side)
	{
		set_track(current_side, current_track, -1);
	}
}

void HFEDiskDrive::flush_track(int density)
{
	if (m_currentTrack.m_state == HFETRACK_Write) {
		write_track(current_side, current_track, m_currentTrack);
		m_currentTrack.m_state = HFETRACK_Read;
	}
}

int HFEDiskDrive::set_track(int SIDE, int TRACK, int density)
{
	flush_track(density);

	current_side = SIDE;
	current_track = TRACK;

	m_currentTrack.reset();

	if (scan_track(current_side, current_track, m_currentTrack, 0) < 0) {
		//TODO:
		return -1;
	}
	m_currentTrack.m_info = new HFETrackSectorInfo[m_currentTrack.m_num_sector];
	m_currentTrack.m_data = new unsigned char[m_currentTrack.m_data_len];
    scan_track(current_side, current_track, m_currentTrack, 1);

	tracklen = m_currentTrack.m_flux_len / 16;

	m_currentTrack.m_state = HFETRACK_Read;

	return 1;
}

int HFEDiskDrive::scan_track(int SIDE, int TRACK, HFETrack& track, int pass, void (*output)(void *ctx, const char* text), void* ctx)
{
	char f[256];

	if (track.m_info != NULL)
	{
		assert(pass >= 1);
	}
	else
	{
		assert(pass == 0);
	}

    // Range check
    if (SIDE >= 2 || TRACK >= info.tracks)
    {
        // TODO: creating a blank track if out of range.
        return -1;
    }
    
	if (pass == 0)
	{
		read_track(SIDE, TRACK, track);
	}

	int tail_decoding = 0;

	if (output)
	{
		const char* dmark = "MFM";

		// Only output when HFETrack has data.
		int mfm_count = 0, fm_count = 0;
		for (int i = 0; i < track.m_num_sector; i++)
		{
			if (track.m_info[i].m_info.si_flags & SECTOR_dden)
			{
				mfm_count++;
			}
			else
			{
				fm_count++;
			}
		}

		int tracklen = track.m_flux_len / 16;
		if (fm_count != 0 && mfm_count != 0)
		{
			dmark = "mixed";
		}
		else
		{
			if (fm_count == 0) {
				tail_decoding = 1;
				dmark = "MFM";
			}
			else {
				tail_decoding = 0;
				dmark = "FM";
				tracklen = track.m_flux_len / 32;
			}
		}

		sprintf(f, "\ntrack %02x %s side %d len=0x%x\n", TRACK, dmark, SIDE, tracklen); output(ctx, f);
	}
    
	unsigned short mfm = 0;
	unsigned int fm = 0;
    
	int gap_bitpos = 0;

	int bitpos = 0;
    
	int mfm_data_rate = m_config.data_rate / 1000;
	if (m_config.density == DISK_FM) {
		mfm_data_rate *= 2;
	}

	SectorInfo current_idam;
	current_idam.si_sector = -1;	// to mean: not yet current
	int sector_index = 0;
	int current_offset = 0;

	while (bitpos < track.m_flux_len)
	{
		next_fm(track.m_flux_data, bitpos, fm);
		mfm = (unsigned short)fm;		// Note: this is legitimate, same flux data
		//TODO: floppy has multiple A1 marks
		if (mfm == MFM_MARK_A1)
		{
			current_idam.set_rate(mfm_data_rate);
			tail_decoding = 1;/*MFM*/
			int mark_bitpos = bitpos;
            
			unsigned char mark = read_byte_mfm(track.m_flux_data, bitpos, mfm);
            
			// Hack: skipping A1 marks
			int mark_count = 0;
			while (mark == 0xa1 && mark_count < 3) {
				mark_count++;
				mark = read_byte_mfm(track.m_flux_data, bitpos, mfm);
			}
			// TODO: some tricky bits here
			// Only read the first DAM after an IDAM
			// A DAM might appear before the first IDAM
			// There could be two IDAM in a row, etc.
			if (mark_count == 3 && mark == 0xfe)
			{
				flux_gap(gap_bitpos, mark_bitpos - 16, track.m_flux_data, 1/*MFM*/, output, ctx);
				idam_mfm(mark, track.m_flux_data, bitpos, mfm, current_idam, output, ctx);
				gap_bitpos = bitpos - 16;
				if (pass == 0)
				{
					track.m_num_sector++;
				}
				else
				{
					track.m_info[sector_index].m_info = current_idam;
					track.m_info[sector_index].m_mark = mark;
					sector_index++;

				// Zero data in case DAM not found
				// More a TODO: if we provide zeros, need to update info.m_data_offset = current_offset;
				//memset(&track.m_data[current_offset], 0, info.si_length);
				}
			}
			// FA/F9 are possible via track write, but wouldn't be detected by the WD1793 FDC.
			else if (mark_count == 3 && (mark == 0xfb || mark == 0xf8))
			{
				// Check that DAM can be associated with IDAM:
				if (current_idam.si_sector >= 0 && // Have previous IDAM
					1 && // TODO: range check
					(current_idam.si_flags & SECTOR_unavail) // previous IDAM already has DAM
					)
				{
					flux_gap(gap_bitpos, mark_bitpos - 16, track.m_flux_data, 1/*MFM*/, output, ctx);
					(void)dam_mfm(mark, track.m_flux_data, bitpos, mfm, current_idam, track.m_data ? &track.m_data[current_offset] : NULL, output, ctx);
					gap_bitpos = bitpos - 16;
					if (pass == 0)
					{
						if (current_idam.si_sector >= 0)
						{
							track.m_data_len += current_idam.si_length;
						}
					}
					else
					{
						// TODO: fix is tricky because of the passes and what to show on output
						// There is also a corresponding FM case.
						// current_idam.si_sector < 0 is one check
						if (sector_index > 0)// Really: "if the previous sector doesn't already have a DAM
						{
							HFETrackSectorInfo& info = track.m_info[sector_index - 1];
							info.m_data_offset = current_offset;
							info.m_bitpos = mark_bitpos;
							// IDAM flags will change
							info.m_info = current_idam;

							current_offset += current_idam.si_length;
						}

					}
				}
				else
				{
					//TODO floating MFM DAM -- need to do currect output
				}
			}
			else
			{
				// Unknown mark or insufficent $A1 marks -- ignore it
				//printf("unknown mark %02x (mark_count=%d)\n", mark, mark_count);
			}

			// Fix up the reading state (mostly, really should have 16 extra bits)
			fm = mfm;
		}
		else if (fm == FM_MARK_FE)
		{
			current_idam.set_rate(mfm_data_rate / 2);
			tail_decoding = 0;/*FM*/

			unsigned char mark, clock;
			sep_fm(fm, mark, clock);

			flux_gap(gap_bitpos, bitpos - 32, track.m_flux_data, 0/*FM*/, output, ctx);

			idam_fm(track.m_flux_data, bitpos, fm, current_idam, output, ctx);
			gap_bitpos = bitpos - 32;

			if (pass == 0)
			{
				track.m_num_sector++;
			}
			else
			{
				track.m_info[sector_index].m_info = current_idam;
				track.m_info[sector_index].m_mark = mark;
				sector_index++;
			}
		}
		else if (fm == FM_MARK_FB || fm == FM_MARK_FA || fm == FM_MARK_F8 || fm == FM_MARK_F9)
		{
			current_idam.set_rate(mfm_data_rate / 2);
			tail_decoding = 0;/*FM*/

			int mark_bitpos = bitpos;

			unsigned char mark, clock;
			sep_fm(fm, mark, clock);

			// Check that DAM can be associated with IDAM:
			if (current_idam.si_sector >= 0 && // Have previous IDAM
				1 && // TODO: range check
				(current_idam.si_flags & SECTOR_unavail) // previous IDAM already has DAM
				)
			{
				flux_gap(gap_bitpos, bitpos - 32, track.m_flux_data, 0/*FM*/, output, ctx);
				(void)dam_fm(track.m_flux_data, bitpos, fm, current_idam, track.m_data ? &track.m_data[current_offset] : NULL, output, ctx);
				gap_bitpos = bitpos - 32;
				if (pass == 0)
				{
					track.m_data_len += current_idam.si_length;
				}
				else
				{
					HFETrackSectorInfo& info = track.m_info[sector_index - 1];
					info.m_data_offset = current_offset;
					info.m_bitpos = mark_bitpos;
					// IDAM flags will change
					info.m_info = current_idam;
		                
					current_offset += current_idam.si_length;
				}
			}
			else
			{
				//TODO floating FM DAM -- need to do currect output
			}
		}
	}

	// Use encoding of last sector seen to choose density
	flux_gap(gap_bitpos, bitpos, track.m_flux_data, tail_decoding, output, ctx);

	return 1;
}

void HFEDiskDrive::read_track(int SIDE, int TRACK, HFETrack& track)
{
    // Copy track data to a contiguous buffer
    int offset = track_pos[TRACK].block * HFE_BLOCK_SIZE;
    if (SIDE != 0) {
        offset += HFE_BLOCK_SIZE/2;
    }
    
    // Stored length is for both sides
    track.m_flux_len = track_pos[TRACK].len / 2;
    
    track.m_flux_data = new unsigned char[track.m_flux_len];
    
    for (int i = 0; i < track.m_flux_len; ) {
        int grab = track.m_flux_len - i;
        if (grab > HFE_BLOCK_SIZE/2) {
            grab = HFE_BLOCK_SIZE/2;
        }
        
        memcpy(&track.m_flux_data[i], &disk[offset], grab);
        
        i += grab;
        offset += HFE_BLOCK_SIZE/2 + grab;
    }
    
    // Convert to bit length
    track.m_flux_len *= 8;
}

void HFEDiskDrive::write_track(int SIDE, int TRACK, HFETrack& track)
{
    int offset = track_pos[TRACK].block * HFE_BLOCK_SIZE;
    if (SIDE != 0) {
        offset += HFE_BLOCK_SIZE/2;
    }

	int len = track.m_flux_len / 8;
    
    for (int i = 0; i < len; ) {
        int grab = len - i;
        if (grab > HFE_BLOCK_SIZE/2) {
            grab = HFE_BLOCK_SIZE/2;
        }
        
        memcpy(&disk[offset], &track.m_flux_data[i], grab);
        
        i += grab;
        offset += HFE_BLOCK_SIZE/2 + grab;
    }
}

void HFEDiskDrive::TrackScan(int side, int track, ScanTrack* callback)
{
	HFETrack trk;

    // Range check
    if (side >= 2 || track >= info.tracks)
    {
        //TODO: shold be a blank track...or an 'error' on callback
        return;
    }

	read_track(side, track, trk);

	callback->Scan_Begin();

#if 0
	unsigned short mfm = 0;
	unsigned int fm = 0;
	int fm_pos = 0;
	int mfm_pos = 0;

	int gap_pos = 0;

	while (fm_pos < trk.m_flux_len) {
		next_mfm(trk.m_flux_data, mfm_pos, mfm);
		next_fm(trk.m_flux_data, fm_pos, fm);
		if (mfm == MFM_MARK_A1) {
// this is just a cut/paste job.  I "don't care" about improvements or handling the "multiple A1" problem right now.
// there is one difference--preserving the 'missed clocks' (and presenting them, \ is better e.g. {a1\04}
			//GAP
			callback->Scan_Gap();
			// scan...

			int mark_bitpos = bitpos;
            
			unsigned char mark = read_byte_mfm(trk.m_flux_data, bitpos, mfm);
            
			// Hack: skipping A1 marks
			while (mark == 0xa1) {
				mark = read_byte_mfm(trk.m_flux_data, bitpos, mfm);
			}
			// TODO: some tricky bits here
			// Only read the first DAM after an IDAM
			// A DAM might appear before the first IDAM
			// There could be two IDAM in a row, etc.
			if (mark == 0xfe)
			{
				idam_mfm(mark, track.m_flux_data, bitpos, mfm, current_idam);
				if (pass == 0)
				{
					trk.m_num_sector++;
				}
				else
				{
					track.m_info[sector_index].m_info = current_idam;
					track.m_info[sector_index].m_mark = mark;
					sector_index++;

				// Zero data in case DAM not found
				// More a TODO: if we provide zeros, need to update info.m_data_offset = current_offset;
				//memset(&track.m_data[current_offset], 0, info.si_length);
				}
			}
			else if (mark == 0xfb || mark == 0xfa)
			{
				// TODO: if current_idam.si_sector < 0, we have to ignore this.
				// also, it we accept it, we should reset current_idam.si_sector to -1.
				int crc_check = dam_mfm(mark, track.m_flux_data, bitpos, mfm, current_idam, track.m_data ? &track.m_data[current_offset] : NULL);
				if (pass == 0)
				{
					track.m_data_len += current_idam.si_length;
				}
				else
				{
					assert(sector_index > 0);
					HFETrackSectorInfo& info = track.m_info[sector_index - 1];
					info.m_data_offset = current_offset;
					info.m_bitpos = mark_bitpos;
					// IDAM flags will change
					info.m_info = current_idam;

					current_offset += current_idam.si_length;
				}
			}
			else
			{
				//TODO: as above, any mark is important--although repeat A1s are a problem...
				printf("unknown mark %02x\n", mark);
			}

			gap_pos = fm_pos = mfm_pos;
		}
		else if (fm == FM_MARK_FE || fm == FM_MARK_FB || fm == FM_MARK_F8 || fm == FM_MARK_FA || fm == FM_MARK_F9) {
			//GAP
			callback->Scan_Gap();
			// scan...

			gap_pos = mfm_pos = fm_pos;
		}

		callback->Scan_Idam();
		callback->Scan_Gap();
		callback->Scan_Dam();
	}
#endif
	callback->Scan_End();

}

void HFEDiskDrive::TrackDump(FILE *f, int side, int track)
{
	HFETrack trk;

	if (scan_track(side, track, trk, 0) < 0) {
		// TODO: "blank" track
		return;
	}

	trk.m_info = new HFETrackSectorInfo[m_currentTrack.m_num_sector];
	trk.m_data = new unsigned char[m_currentTrack.m_data_len];

    scan_track(side, track, trk, 1);

	// Output on 3rd pass so we get gap / MFM information
    scan_track(side, track, trk, 2, file_out, f);
}

char *HFEDiskDrive::TrackDump(int side, int track)
{
	size_t len = 0;

	HFETrack trk;

	if (scan_track(side, track, trk, 0) < 0) {
		//TODO: "blank" track
		const char *msg = "[TODO]";

		char *res = new char[strlen(msg) + 1];

		strcpy(res, msg);

		return res;
	}

	trk.m_info = new HFETrackSectorInfo[m_currentTrack.m_num_sector];
	trk.m_data = new unsigned char[m_currentTrack.m_data_len];

	scan_track(side, track, trk, 1);

	// Scan finished, ready for output scan
	scan_track(side, track, trk, 2, char_size, &len);
	char *res = new char[len + 1];
	res[len] = 0;
	char *ptr = res;

	scan_track(side, track, trk, 2, char_out, &ptr);

	return res;
}


static void next_mfm(const unsigned char *buf, int &bitpos, unsigned short &mfm)
{
	int off = (bitpos >> 3);
	int bit = (bitpos & 7);
    
	mfm <<= 1;
	int b = !!(buf[off] & (1 << bit));
	mfm |= b;
    
	bitpos++;
}

static void next_fm(const unsigned char *buf, int &bitpos, unsigned int &fm)
{
	int off = (bitpos >> 3);
	int bit = (bitpos & 7);
    
	fm <<= 1;
	int b = !!(buf[off] & (1 << bit));
	fm |= b;
    
	bitpos++;
}

static void write_next_mfm(unsigned char *buf, int &bitpos, unsigned short &mfm)
{
	int off = (bitpos >> 3);
	int bit = (bitpos & 7);
 
	unsigned char mask = (1 << bit);
	if (mfm & 0x8000) {
		buf[off] |= mask;
	} else {
		buf[off] &= ~mask;
	}

	mfm <<= 1;
    
	bitpos++;
}

static void write_next_fm(unsigned char *buf, int &bitpos, unsigned int &fm)
{
	int off = (bitpos >> 3);
	int bit = (bitpos & 7);
 
	unsigned char mask = (1 << bit);
	if (fm & 0x80000000) {
		buf[off] |= mask;
	} else {
		buf[off] &= ~mask;
	}

	fm <<= 1;

	bitpos++;
}

static unsigned int mfm_encode(int prev_bit, int data, int skip_clock)
{
    unsigned int res = 0;
    int prev_zero = 0;
    if (!prev_bit) {
        prev_zero = 2;
    }
    for (int i = 7; i >= 0; i--)
    {
        res <<= 2;
        int bit = (data >> i) & 1;
        int clk = (((skip_clock >> i) & 1) << 1) ^ 2;
        if (bit) {
            res |= 1;           // always 01
            prev_zero = 0;
        } else {
            res |= (prev_zero & clk); // 00 unless previous zero then 10 (but skip_clock can override)
            prev_zero = 2;
        }
    }
    
    return res;
}

static unsigned int fm_encode(int data, int clock)
{
	unsigned int fm = 0;

	for (int i = 0; i < 8; i++) {
		fm <<= 2;
		if (clock & 0x80)
			fm |= FM_CLOCK;
		clock <<= 1;

		fm <<= 2;
		if (data & 0x80)
			fm |= FM_DATA_ONE;
		data <<= 1;
	}

	return fm;
}

static unsigned char mfm_write(unsigned char* buf, int& bitpos, unsigned char data, unsigned char skip, unsigned char prev)
{
	unsigned short mfm;

	mfm = mfm_encode(prev & 1, data, skip);
	
	for (int k = 0; k < 16; k++) {
		write_next_mfm(buf, bitpos, mfm);
	}

	return data;
}

static void fm_write(unsigned char* buf, int& bitpos, unsigned char data, unsigned char clock)
{
	unsigned int fm = fm_encode(data, clock);

	for (int k = 0; k < 32; k++) {
		write_next_fm(buf, bitpos, fm);
	}
}

static void sep_mfm(unsigned short mfm, unsigned char &data, unsigned char &clock)
{
	data = clock = 0;
	for (int i = 0; i < 8; i++) {
		clock = (clock << 1) | ((mfm >> 15) & 1);
		mfm <<= 1;
		data = (data << 1) | ((mfm>> 15)  & 1);
		mfm <<= 1;
	}
    //TODO: Not the real clock! That requires selective NOR
	// Maybe: I suspect it is the 'real' clock which is not just 0xff
	// The harder part is computing the 'real' clock on write.
}

static unsigned char grab2(unsigned int &fm)
{
	char res;
	res = (fm >> 30) & 3;
	fm <<= 2;
    
	return res;
}

static void sep_fm(unsigned int fm, unsigned char &data, unsigned char &clock)
{
    data = clock = 0;
	for (int i = 0; i < 8; i++) {
		unsigned char c, d;
		c = grab2(fm);
		clock <<= 1;
		if (c == FM_CLOCK) {
			clock |= 1;
		}
		d = grab2(fm);
		data <<= 1;
		if (d == FM_DATA_ONE) {
			data |= 1;
		}
	}
}

static unsigned char read_byte_mfm(const unsigned char *buf, int &bitpos, unsigned short &mfm)
{
    unsigned char data, clock;
    sep_mfm(mfm, data, clock);
    
    for (int i = 0; i < 16; i++) {
        next_mfm(buf, bitpos, mfm);
    }
    
    return data;
}

static unsigned char read_byte_fm(const unsigned char *buf, int &bitpos, unsigned int &fm)
{
    unsigned char data, clock;
    sep_fm(fm, data, clock);
    
    for (int i = 0; i < 32; i++) {
        next_fm(buf, bitpos, fm);
    }
    
    return data;
}

// MFM/FM aren't really different. Should pass in array of decoded values and initial CRC (FFFF for FM, FFFF + a1 a1 a1 for MFM)
// A complication is the location of the mark, with MFM it is always a 'prefix', FM the mark is clocked
static void idam_mfm(unsigned int id, const unsigned char* buf, int& bitpos, unsigned short &mfm, SectorInfo& info, void (*output)(void *ctx, const char* text), void* ctx)
{
	unsigned int trk, sid, sec, len, crc1, crc2;
	trk = read_byte_mfm(buf, bitpos, mfm);
	sid = read_byte_mfm(buf, bitpos, mfm);
	sec = read_byte_mfm(buf, bitpos, mfm);
	len = read_byte_mfm(buf, bitpos, mfm);
	crc1 = read_byte_mfm(buf, bitpos, mfm);
	crc2 = read_byte_mfm(buf, bitpos, mfm);
    
	unsigned int crc = 0xffff;
	crc = update_crc(crc, 0xa1);
	crc = update_crc(crc, 0xa1);
	crc = update_crc(crc, 0xa1);
	crc = update_crc(crc, id);
	crc = update_crc(crc, trk);
	crc = update_crc(crc, sid);
	crc = update_crc(crc, sec);
	crc = update_crc(crc, len);
	crc = update_crc(crc, crc1);
	crc = update_crc(crc, crc2);
    
	static int sector_size[4] = { 128, 256, 512, 1024 };
    
	info.si_flags = SECTOR_dden;		// MFM encoding

	// Mark as not available in case DAM not found
	info.si_flags |= SECTOR_unavail;

	if (crc != 0) {
		// TODO: crc error in IDAM really means block is 'invisible'
		//info.si_flags |= SECTOR_err;
	}
	info.si_length = sector_size[len & 3];
	info.si_sector = sec;
	info.si_track = trk;
	info.si_side = sid;

	if (output)
	{
		unsigned int data_int[] = { 0xa1, 0xa1, 0xa1, id, trk, sid, (unsigned int)info.si_sector, len, crc1, crc2 };
		const int len = int(sizeof data_int / sizeof data_int[0]);
		unsigned char data[len];
		for (int i = 0; i < len; i++)
			data[i] = data_int[i];

		dump_segment(ctx, output, "idam", sec, data, 0, len, 1, 4, len - 2);
	}
}

static int dam_mfm(unsigned int mark, const unsigned char* buf, int& bitpos, unsigned short &mfm, SectorInfo& info, unsigned char* dst, void (*output)(void *ctx, const char* text), void* ctx)
{
	unsigned int crc = 0xffff;
	crc = update_crc(crc, 0xa1);
	crc = update_crc(crc, 0xa1);
	crc = update_crc(crc, 0xa1);
	crc = update_crc(crc, mark);
    
	crc = read_dam_data_mfm(info.si_length, buf, bitpos, mfm, crc, dst);
	unsigned int crc1, crc2;
	crc1 = read_byte_mfm(buf, bitpos, mfm);
	crc2 = read_byte_mfm(buf, bitpos, mfm);
	crc = update_crc(crc, crc1);
	crc = update_crc(crc, crc2);
    
	// Mark as available
	info.si_flags &= ~SECTOR_unavail;

	//TODO SECTOR_delete marker
#if 0
	if (mark == 0xfa) {
		info.si_flags |= SECTOR_delete;		// Deleted sector mark
	}
#endif

	if (crc != 0) {
		info.si_flags |= SECTOR_err;	// CRC error in DAM
	}

	if (output)
	{
		int len = info.si_length + 3 + 1 + 2;
		unsigned char *data = new unsigned char[len];
		data[0] = data[1] = data[2] = 0xa1;
		data[3] = mark;
		if (dst != NULL)
		{
			memcpy(&data[4], dst, info.si_length);
		}
		else
		{
			memset(&data[4], 0x00, info.si_length);
		}
		data[len - 2] = crc1;
		data[len - 1] = crc2;
		dump_segment(ctx, output, "dam", info.si_sector, data, 0, len, 1, 4, len - 2);
		delete [] data;
	}

	return crc;
}

static unsigned int read_dam_data_mfm(int len, const unsigned char* buf, int& bitpos, unsigned short &mfm, unsigned int crc, unsigned char* dst)
{
	for (int i = 0; i < len; i++)
	{
		unsigned int data = read_byte_mfm(buf, bitpos, mfm);

		crc = update_crc(crc, data);
		if (dst) {
			*dst++ = data;
		}
	}

	return crc;
}
static void idam_fm(const unsigned char* buf, int& bitpos, unsigned int &fm, SectorInfo& info, void (*output)(void *ctx, const char* text), void* ctx)
{
	unsigned int mark, trk, sid, sec, len, crc1, crc2;
	mark = read_byte_fm(buf, bitpos, fm);
	trk = read_byte_fm(buf, bitpos, fm);
	sid = read_byte_fm(buf, bitpos, fm);
	sec = read_byte_fm(buf, bitpos, fm);
	len = read_byte_fm(buf, bitpos, fm);
	crc1 = read_byte_fm(buf, bitpos, fm);
	crc2 = read_byte_fm(buf, bitpos, fm);
    
	unsigned int crc = 0xffff;
	crc = update_crc(crc, mark);
	crc = update_crc(crc, trk);
	crc = update_crc(crc, sid);
	crc = update_crc(crc, sec);
	crc = update_crc(crc, len);
	crc = update_crc(crc, crc1);
	crc = update_crc(crc, crc2);
    
	static int sector_size[4] = { 128, 256, 512, 1024 };
    
	info.si_flags = 0;		// FM encoding

	// Mark as not available in case DAM not found
	info.si_flags |= SECTOR_unavail;

	//TODO: SECTOR_delete SECTOR_damhigh
	// Have to get these 'straight'
#if 0
	if (mark == 0xfa) {
		info.si_flags |= SECTOR_delete;		// Deleted sector mark
	}
	if (mark == 0xfb) {
		info.si_flags |= SECTOR_damhigh;	// Deleted sector mark
	}
#endif
	if (crc != 0) {
		// TODO: crc error in IDAM really means block is 'invisible'
		//info.si_flags |= SECTOR_err;
	}
	info.si_length = sector_size[len & 3];
	info.si_sector = sec;
	info.si_track = trk;
	info.si_side = sid;

	if (output)
	{
		unsigned int data_int[] = { mark, trk, sid, (unsigned int)info.si_sector, len, crc1, crc2 };
		const int len = int(sizeof data_int / sizeof data_int[0]);
		unsigned char data[len];
		for (int i = 0; i < len; i++)
			data[i] = data_int[i];

		dump_segment(ctx, output, "idam", sec, data, 0, len, 1, 1, len - 2);
	}
}

static int dam_fm(const unsigned char* buf, int& bitpos, unsigned int &fm, SectorInfo& info, unsigned char* dst, void (*output)(void *ctx, const char* text), void* ctx)
{
	unsigned char mark = read_byte_fm(buf, bitpos, fm);

	unsigned int crc = 0xffff;
	crc = update_crc(crc, mark);
    
	crc = read_dam_data_fm(info.si_length, buf, bitpos, fm, crc, dst);
	unsigned int crc1, crc2;
	crc1 = read_byte_fm(buf, bitpos, fm);
	crc2 = read_byte_fm(buf, bitpos, fm);
	crc = update_crc(crc, crc1);
	crc = update_crc(crc, crc2);
    
	// Mark as available
	info.si_flags &= ~SECTOR_unavail;

	if (crc != 0) {
		info.si_flags |= SECTOR_err;		// CRC error in DAM
	}

	if (output)
	{
		int len = info.si_length + 1 + 2;
		unsigned char *data = new unsigned char[len];
		data[0] = mark;
		if (dst != NULL)
		{
			memcpy(&data[1], dst, info.si_length);
		}
		else
		{
			memset(&data[1], 0x00, info.si_length);
		}
		data[len - 2] = crc1;
		data[len - 1] = crc2;
		dump_segment(ctx, output, "dam", info.si_sector, data, 0, len, 1, 1, len - 2);
		delete [] data;
	}

	return crc;
}

static unsigned int read_dam_data_fm(int len, const unsigned char* buf, int& bitpos, unsigned int &fm, unsigned int crc, unsigned char* dst)
{
	for (int i = 0; i < len; i++)
	{
		unsigned int data = read_byte_fm(buf, bitpos, fm);

		crc = update_crc(crc, data);
		if (dst) {
			*dst++ = data;
		}
	}

	return crc;
}

// Output text with potential compressed runs as "text*%d"
static int run_out(const char* text, char* prev, int count, void (*output)(void *ctx, const char* text), void* ctx, int& output_len)
{
	output_len = 0;

	if (count == 0) {
		strcpy(prev, text);
		count++;
	}
	else {
		if (strcmp(text, prev) != 0) {
			char f[256];
			if (count == 1) {
				sprintf(f, "%s", prev);
			}
			else {
				sprintf(f, "%s*%d", prev, count);
			}
			output_len = strlen(f);
			output(ctx, f);
			count = 1;
			strcpy(prev, text);
		}
		else {
			count++;
		}
	}

	return count;
}

// The "flux gap" is really irrelevant for emulated floppies
// However, it's important for debugging (and curiosity).  A gap
// with complicated data is a possible decoding bug or a bad image
// Parts of the gap (after format) are never read or written so in some ways
// displaying the data (or looking for sync) is wrong.
static void flux_gap(int begin, int end, unsigned char* buf, int encoding, void (*output)(void *ctx, const char* text), void* ctx)
{
	if (output)
	{
		char f[256];
		sprintf(f, "         %s:", encoding == 0 ? "fm" : "mfm"); output(ctx, f);

		char prev_text[256];
		int count = 0;
		int output_len;
		int chunk_out = 0;
		if (encoding == 0) // FM
		{
			unsigned int fm = 0;
			int bitpos = begin;
			while (bitpos + 32 <= end)
			{
				for (int i = 0; i < 32; i++) {
					next_fm(buf, bitpos, fm);
				}
				unsigned char data, clock;
				sep_fm(fm, data, clock);
				clock ^= 0xff;
				if (clock != 0) {
					sprintf(f, " %02x\\%02x", data, clock);
				}
				else {
					sprintf(f, " %02x", data);
				}
				count = run_out(f, prev_text, count, output, ctx, output_len);
				if (output_len) {
					chunk_out++;
					if (chunk_out >= 16) {
						chunk_out = 0;
						output(ctx, "\n         ");
					}
				}
			}
			run_out("", prev_text, count, output, ctx, output_len);

			if (bitpos < end)
			{
				unsigned int mask = 0;
				while (bitpos < end)
				{
					mask = 1 | (mask << 1);
					next_fm(buf, bitpos, fm);
				}
				sprintf(f, " rem: %x/%x", fm & mask, mask); output(ctx, f);
			}
			output(ctx, "\n");
		}
		else {
			// MFM
			unsigned short mfm = 0;

			int bitpos = begin;
			if (bitpos >= 16)
			{
				// Pull in the previous word to get the correct previous bit
				bitpos -= 16;
				for (int i = 0; i < 16; i++) {
					next_mfm(buf, bitpos, mfm);
				}
			}

			while (bitpos + 16 <= end)
			{
				int prev_bit = mfm & 1;
				for (int i = 0; i < 16; i++) {
					next_mfm(buf, bitpos, mfm);
				}
				unsigned char data, clock;
				sep_mfm(mfm, data, clock);
				// Translate clock into 'missing clocks'
				unsigned int pure_mfm = mfm_encode(prev_bit, data, 0x00);
				unsigned char pure_data, pure_clock;
				sep_mfm(pure_mfm, pure_data, pure_clock);
				clock ^= pure_clock;
				if (clock != 0) {
					sprintf(f, " %02x\\%02x", data, clock);
				}
				else {
					sprintf(f, " %02x", data);
				}
				count = run_out(f, prev_text, count, output, ctx, output_len);
				if (output_len) {
					chunk_out++;
					if (chunk_out >= 16) {
						chunk_out = 0;
						output(ctx, "\n         ");
					}
				}
			}
			run_out("", prev_text, count, output, ctx, output_len);

			if (bitpos < end)
			{
				unsigned int mask = 0;
				while (bitpos < end)
				{
					mask = 1 | (mask << 1);
					next_mfm(buf, bitpos, mfm);
				}
				sprintf(f, " rem: %x/%x", mfm & mask, mask); output(ctx, f);
			}
			output(ctx, "\n");
		}
	}
}

static void mfm_align_mark(HFETrack &trk, int &bitpos)
{
	unsigned short mfm = 0;

	int orig_bitpos = bitpos;
	// Retrieve previous state
	if (bitpos >= 16) {
		bitpos -= 16;
		for (int i = 0; i < 16; i++) {
			next_mfm(trk.m_flux_data, bitpos, mfm);
		}
	}

	// Sloppy, will got past end of track
	for (int i = 0; i < 16; i++) {
		next_mfm(trk.m_flux_data, bitpos, mfm);
		if (mfm == MFM_MARK_A1) {
			bitpos -= 16;
			return;
		}
	}

	bitpos = orig_bitpos;
}

static void fm_align_mark(HFETrack &trk, int &bitpos)
{
	int orig_bitpos = bitpos;

	unsigned int fm = 0;

	// Retrieve previous state
	if (bitpos >= 32) {
		bitpos -= 32;
		for (int i = 0; i < 32; i++) {
			next_fm(trk.m_flux_data, bitpos, fm);
		}
	}

	// OK, somewhat sloppy here as we'll go past the position
	for (int i = 0; i < 32; i++) {
		next_fm(trk.m_flux_data, bitpos, fm);
		if (fm == FM_MARK_FE || fm == FM_MARK_FB || fm == FM_MARK_FA || fm == FM_MARK_F8 || fm == FM_MARK_F9) {
			bitpos -= 32;
			return;
		}
	}

	bitpos = orig_bitpos;
}



DiskDrive* fd_create_hfe(const DriveConfiguration &config, const DiskImageMetadata* metadata)
{
	return new HFEDiskDrive(config, metadata);
}

