#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <ctype.h>

#ifdef _MSC_VER
#define stricmp _stricmp
#else
#define stricmp strcasecmp
#endif

#if defined(__APPLE__) || defined(__linux__)
#define _strnicmp strncasecmp
#endif

#include "fd_image.h"

// IMD format

IMDDiskDrive::IMDDiskDrive(const DriveConfiguration &config, const DiskImageMetadata* metadata)
	: BlockDrive(config, metadata)
{
	// Create a placeholder comment

	int size_ok = ensure_disk_size(1024);
	assert(size_ok);

	char *comment = (char*)disk;
	int day, month, year, hour, min, sec;
	if (metadata)
	{
		//TODO: metadata could/should override the comment
		day = metadata->dd;
		month = metadata->mm;
		year = metadata->yyyy;
		hour = metadata->hour;
		min = metadata->min;
		sec = metadata->sec;
	}
	else
	{
		day = month = 10;
		year = 1980;
		hour = min = sec = 10;
	}

	sprintf(comment, "IMD 1.18: %02d/%02d/%04d %02d:%02d:%02d\r\ntrs80gp blank disk %d sides %d tracks\n",
		day, month, year,
		hour, min, sec,
		info.sides, info.tracks);
	comment_len = strlen(comment);

	// Placeholder - make sure len isn't zero
	len = comment_len;
}

struct tmp
{
	int sector_number;
	int data_offset;
	int flag;
	int length;
} ;

// TODO:
// - probably the block-based interfaces should simply discard
//   the original source.  The "Save()" method would recreate it.
// This avoids having to maintain two copies of the information.
// (it also means the whole 'block' interface comes down to 
// initializing the pre-defined table structures)
int IMDDiskDrive::Load(unsigned char *data, int count)
{
// Sector Data Record decoding
#define FLAG_COMP (1)
#define FLAG_DEL (2)
#define FLAG_ERR (4)
#define FLAG_UNAVAIL (8)

	DeleteAllSectors();

	int size_ok = ensure_disk_size(count);
	assert(size_ok);

	len = count;
	memcpy(disk, data, count);

    //TODO: IMD really has a header, we could check version here
    //IMD v.vv: dd/mm/yyyy hh:mm:ss
    
	// skip ascii header
	int i;
	for (i = 0; i < count && disk[i] != 0x1a; i++)
	{
	}
    comment_len = i;
	i++;

	// TODO: members of the 'block implementation' class.
	// TODO: don't handle "head" (side) properly.
    info.sides = 1;
    info.tracks = 0;

    int start = i;
    for (int second = 0; second < 2; second++)
    {
        i = start;
        
        if (second) {
			side_info = new SideInfo[info.sides];
			for (int s = 0; s < info.sides; s++)
			{
				side_info[s].track_info = new TrackInfo[info.tracks];
				side_info[s].sector_info = new SectorInfo*[info.tracks];
				side_info[s].sector_data = new unsigned char**[info.tracks];
			}
        }
        
        int track = 0;
        for (;i + 4 < count; track++) {
            // sequence of tracks.
            int mode, /*cylinder,*/ head, sectors, size;
            mode = data[i+0];
            //[unused] cylinder = data[i+1];
            head = data[i+2];
            sectors = data[i+3];
            size = data[i+4];

            i += 5;

			int map = i;
            i += sectors;
			int cyl_map = -1;
			if (head & 0x80)
			{
				// Optional cylinder map
				cyl_map = i;
				i += sectors;
			}
			int head_map = -1;
			if (head & 0x40)
			{
				// Optional head map.
				head_map = i;
				i += sectors;
			}
			head &= 1; // only one bit valid.
			if (head)
			{
				info.sides = 2;
				track--;
			}
			else
			{
				if (!second) {
					info.tracks++;
				}
			}

			struct tmp sect_info[256];	// TODO: drop this.	//ERROR causes problems, too.

			int length_map = -1;
            int sector_length = 128;
            switch (size)
            {
            case 0: sector_length = 128; break;
            case 1: sector_length = 256; break;
            case 2: sector_length = 512; break;
            case 3: sector_length = 1024; break;
            case 4: sector_length = 2048; break;
            case 5: sector_length = 4096; break;
			case 255:
				// Variable sector length map
				length_map = i;
				i += sectors * 2;
				break;
			//OK, but, you need save support
			//Also, hard to test.  Variable sector per track is rare.
/*
             A sector size value of 0xFF indicates that a  table  of  sector
             sizes occurs  after  the  sector  numbering/cylinder/head  maps
             (immediately before the data records) which contains one 16-bit
             value  (in little endian format)  per sector which defines  the
             actual size of that sector.
*/
            }
            if (second)
            {
                side_info[head].track_info[track].sectors = sectors;
                side_info[head].sector_info[track] = new SectorInfo[sectors];
				side_info[head].sector_data[track] = new unsigned char*[sectors];
            }
            if (!second)
            {
                //printf("track %d has %d sectors\n", info.tracks, sectors);
            }
            for (int j = 0; j < sectors; j++)
            {
                int flag = data[i++];

				if (length_map >= 0)
				{
					sector_length = data[length_map + j*2] | (data[length_map + j*2 + 1] << 8);
				}

                sect_info[j].sector_number = data[map + j];
                sect_info[j].data_offset = i;	// either bytes or compressed byte
                sect_info[j].length = sector_length;
                sect_info[j].flag = 0;

				// Sector Data Record decoding
                switch (flag)
                {
                case 0: sect_info[j].flag =                                  FLAG_UNAVAIL; break;	// No data in image
                case 1: break;
                case 2: sect_info[j].flag = FLAG_COMP                      ; break;
                case 3: sect_info[j].flag =             FLAG_DEL           ; break;
                case 4: sect_info[j].flag = FLAG_COMP | FLAG_DEL           ; break;
                case 5: sect_info[j].flag =                        FLAG_ERR; break;
                case 6: sect_info[j].flag = FLAG_COMP            | FLAG_ERR; break;
                case 7: sect_info[j].flag =             FLAG_DEL | FLAG_ERR; break;
                case 8: sect_info[j].flag = FLAG_COMP | FLAG_DEL | FLAG_ERR; break;
                default: // invalid
                    break;
                }
                if (sect_info[j].flag & FLAG_COMP)
                {
                    i++;
                }
                else if (!(sect_info[j].flag & FLAG_UNAVAIL))
                {
                    i += sector_length;
                }
                // IMD has fixed sector length per track (unless length_map extension is used)
                if (second) {
                    side_info[head].sector_info[track][j].si_flags = 0;
                    side_info[head].sector_info[track][j].si_length = sector_length;
                    if (sect_info[j].flag & FLAG_UNAVAIL) {
                        side_info[head].sector_info[track][j].si_flags |= SECTOR_unavail;
                    }
                    if (sect_info[j].flag & FLAG_DEL) {
                        side_info[head].sector_info[track][j].si_flags |= SECTOR_delete;
                    }
                    if (sect_info[j].flag & FLAG_ERR) {
                        side_info[head].sector_info[track][j].si_flags |= SECTOR_err;
                    }
                    // 0,1,2=500,300,250 FM, 3,4,5=500,300,250 MFM
					switch (mode)
					{
						case 0: case 3: side_info[head].sector_info[track][j].set_rate(500); break;
						case 1: case 4: side_info[head].sector_info[track][j].set_rate(300); break;
						case 2: case 5: side_info[head].sector_info[track][j].set_rate(250); break;
					}
					//ERROR TODO: the bit rate may be not-expected.
                    if (mode > 2) {
                        side_info[head].sector_info[track][j].si_flags |= SECTOR_dden;
                    }
                    side_info[head].sector_info[track][j].si_side = head;
					if (head_map >= 0) {
						side_info[head].sector_info[track][j].si_track = data[head_map + j];
					}
                    side_info[head].sector_info[track][j].si_track = track;
					if (cyl_map >= 0) {
						side_info[head].sector_info[track][j].si_track = data[cyl_map + j];
					}
                    side_info[head].sector_info[track][j].si_sector = data[map + j];

					// Read actual sector data.
					//[unused] int sector = side_info[head].sector_info[track][j].si_sector;
					int offset = sect_info[j].data_offset;
					side_info[head].sector_data[track][j] = new unsigned char[sector_length];

					// Note: track conversion will still show a DAM here when it should not.
					if (sect_info[j].flag & FLAG_UNAVAIL)
					{
						// UNAVAIL means no data, let's put in zeros, but it should be ignored
						for (int k = 0; k < sector_length; k++)
						{
							side_info[head].sector_data[track][j][k] = 0;
						}
					}
					else if (sect_info[j].flag & FLAG_COMP)
					{
						int repeat = data[offset];
						for (int k = 0; k < sector_length; k++)
						{
							side_info[head].sector_data[track][j][k] = repeat;
						}
					}
					else
					{
						for (int k = 0; k < sector_length; k++)
						{
							side_info[head].sector_data[track][j][k] = data[offset + k];
						}
					}
                }
            }
        }
    }

	// Note: Often images have fewer tracks than
	// the actual drive.

	current_track = 0;
	radial_position = 0;
	tc_track = 0;

	CreateTrackProgram();

	return 0;
}

const char* IMDDiskDrive::Save(FILE *fp)
{
    // comment
    int wlen = fwrite(disk, 1, comment_len, fp);
	if (wlen == 0) return FileError();
	if (wlen != comment_len) return "fwrite lost bytes";

    if (fputc(0x1a, fp) == EOF) return FileError();

	for (int track = 0; track < info.tracks; track++)
	{
		for (int side = 0; side < info.sides; side++)
		{
			TrackInfo *track_info = side_info[side].track_info;
			SectorInfo** sector_info = side_info[side].sector_info;
			unsigned char ***sector_data = side_info[side].sector_data;

			int sectors =  track_info[track].sectors;
			if (sectors == 0)
			{
				continue;
			}

			int mode = 0, cylinder, head, size;
	        
			//ERROR TODO: must check all sectors on track are consistent encoding (if no, cannot save)
			int rate = sector_info[track][0].get_rate();
			if (rate == 0)
			{
				rate = m_config.data_rate / 1000;
			}
			if (rate < 275) { mode = 2; }		// 250 KB/s FM
			else if (rate < 400) { mode = 1; }	// 300 KB/s FM
			else if (rate < 600) { mode = 0; }	// 500 KB/s FM	//ERROR might be a lot higher
			if (sector_info[track][0].si_flags & SECTOR_dden) {
				mode += 3; // MFM rate
			}

			// Get sector size for this track
			int sector_length = sector_info[track][0].si_length;
			size = 1;
			switch (sector_length)
			{
			case 128: size = 0; break;
			case 256: size = 1; break;
			case 512: size = 2; break;
			case 1024: size = 3; break;
			default: 
				//TODO - recover contents
				return "Unsupported IMD sector size";
			}

			cylinder = track;
			head = side;

			// Flag if cylinder or head maps are required
			for (int sec = 0; sec < sectors; sec++)
			{
				if (sector_info[track][sec].si_track != track)
				{
					head |= 0x80;
				}
				if (sector_info[track][sec].si_side != side)
				{
					head |= 0x40;
				}
			}

			if (fputc(mode, fp) == EOF) return FileError();
			if (fputc(cylinder, fp) == EOF) return FileError();
			if (fputc(head, fp) == EOF) return FileError();
			if (fputc(sectors, fp) == EOF) return FileError();
			if (fputc(size, fp) == EOF) return FileError();
	        
			// sector numbering map
			for (int sec = 0; sec < sectors; sec++)
			{
				//TODO crude consistency check, all sectors must be same length
				if(	sector_info[track][sec].si_length != sector_length) return "IMD sectors differ in size on same track";
				if (fputc(sector_info[track][sec].si_sector, fp) == EOF) return FileError();
			}

			// cylinder map
			if (head & 0x80)
			{
				for (int sec = 0; sec < sectors; sec++)
				{
					if (fputc(sector_info[track][sec].si_track, fp) == EOF) return FileError();
				}
			}
	        
			// head map
			if (head & 0x40)
			{
				for (int sec = 0; sec < sectors; sec++)
				{
					if (fputc(sector_info[track][sec].si_side, fp) == EOF) return FileError();
				}
			}

			// sector data
			for (int sec = 0; sec < sectors; sec++)
			{
				int seclen = sector_info[track][sec].si_length;

				unsigned char* data = sector_data[track][sec];
				int mark = 0;

				// Sector Data Record encoding
				if (sector_info[track][sec].si_flags & SECTOR_delete) {
					mark |= FLAG_DEL;
				}

				if (sector_info[track][sec].si_flags & SECTOR_err) {
					mark |= FLAG_ERR;
				}
				// Check for repeated data
				int compress = 1;
				for (int j = 0; j < seclen; j++) {
					if (data[0] != data[j]) {
						compress = 0;
						break;
					}
				}
				if (compress) {
					mark |= FLAG_COMP;
					seclen = 1;
				}

				if (sector_info[track][sec].si_flags & SECTOR_unavail) {
					mark = 0;
					seclen = 0;
				}
				else {
					mark++;
				}
				if (fputc(mark, fp) == EOF) return FileError();
				if (seclen != 0) {
					int writelen = fwrite(data, 1, seclen, fp);
					if (writelen == 0) return FileError();
					if( writelen != seclen ) return "fwrite lost bytes";
				}
			}
		}
	}
    
    dirty = 0;
    
	return 0;
}

// Crude APIs added to make it easier for 'xtar' to create blank floppies

void IMDDiskDrive::MakeTrack(int track, int sector_count, int sector_len, int first_id, int density)
{
	current_track = track;

	if (track == 0) {
		info.tracks = 0;
	}
	AppendBlankTrack();//RETURN VALUE

	TrackInfo tinfo;
	tinfo.sectors = sector_count;

	SectorInfo* sinfo = new SectorInfo[sector_count];
	unsigned char ** tdata = new unsigned char*[sector_count];

	for (int i = 0; i < sector_count; i++)
	{
		sinfo[i].si_length = sector_len;
		sinfo[i].si_flags = density ? SECTOR_dden : 0;
		sinfo[i].si_side = 0;
		sinfo[i].si_track = track;
		sinfo[i].si_sector = first_id + i;

		// TODO: should set rate...

		//TODO: a JV3-specific hack here is to force sinfo[i].si_side to be 0 or 1 based on the current side.
		//But can IMD (or any other 'block' format) deal with the SIDE IDAM value being different from the logical side???
#ifdef JV3_SIDE_FIX
		sinfo[i].si_side = (current_side == 0) ? 0 : 1;
#endif

		tdata[i] = new unsigned char[sinfo[i].si_length];
		memset(tdata[i], 0, sinfo[i].si_length);
	}

	dirty = 1;
	// Update info
	//TODO: may need to change tracks/sides DiskInfo info;

	for (int i = 0; i < side_info[current_side].track_info[current_track].sectors; i++)
	{
		delete [] side_info[current_side].sector_data[current_track][i];
	}
	delete [] side_info[current_side].sector_data[current_track];
	side_info[current_side].sector_data[current_track] = tdata;

	delete [] side_info[current_side].sector_info[current_track];
	side_info[current_side].sector_info[current_track] = sinfo;

	side_info[current_side].track_info[current_track] = tinfo;
}

void IMDDiskDrive::WriteSector(int side, int track, int sector_index, int sector_len, const unsigned char* data)
{
	memcpy( side_info[side].sector_data[track][sector_index], data, sector_len);
}

DiskDrive* fd_create_imd(const DriveConfiguration &config, const DiskImageMetadata* metadata)
{
	return new IMDDiskDrive(config, metadata);
}
