#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <ctype.h>

#ifdef _MSC_VER
#define stricmp _stricmp
#else
#define stricmp strcasecmp
#endif

#if defined(__APPLE__) || defined(__linux__)
#define _strnicmp strncasecmp
#endif

#include "fd_image.h"

JV1DiskDrive::JV1DiskDrive(const DriveConfiguration &config, const DiskImageMetadata* metadata)
	: BlockDrive(config, metadata)
{
	// TODO: in factoring common code, the track count here could be invalid.
	// JV1 only allows 35 tracks etc.

	len = 1;// Placeholder length makes it act like a JV3/IMD blank
}

int JV1DiskDrive::Load(unsigned char *data, int count)
{
	DeleteAllSectors();

	int size_ok = ensure_disk_size(count);
	assert(size_ok);

	len = count;
	memcpy(disk, data, count);

	int sectors = 10;
	int sector_length = 256;
	int total_sectors = len / sector_length;
	// Create sector info
    info.sides = 1;
    info.tracks = (total_sectors + sectors - 1 ) / sectors;

	side_info = new SideInfo[1];
	TrackInfo *track_info = side_info[0].track_info = new TrackInfo[info.tracks];
	SectorInfo **sector_info = side_info[0].sector_info = new SectorInfo*[info.tracks];
	unsigned char ***sector_data = side_info[0].sector_data = new unsigned char**[info.tracks];

	for (int track = 0; track < info.tracks; track++)
	{
		track_info[track].sectors = total_sectors > sectors ? sectors : total_sectors;
		total_sectors -= sectors;

		sector_info[track] = new SectorInfo[sectors];
		sector_data[track] = new unsigned char*[sectors];
		for (int sec = 0; sec < sectors; sec++)
		{
			sector_info[track][sec].si_length = sector_length;
			sector_info[track][sec].si_side = 0;
			sector_info[track][sec].si_track = track;
			sector_info[track][sec].si_sector = sec;
			sector_info[track][sec].si_flags = 0;
			if (track == 17) {
				sector_info[track][sec].si_flags |= SECTOR_delete;
			}

			sector_data[track][sec] = new unsigned char[sector_length];
			memcpy(sector_data[track][sec], data + track * sectors * sector_length + sec * sector_length, sector_length);
		}
	}

    current_track = 0;
	radial_position = 0;
	tc_track = 0;

	CreateTrackProgram();

	return len;
}

const char* JV1DiskDrive::Save(FILE *fp)
{
	// TODO: must check that track/sector/dam/crc info is consistent.
	TrackInfo *track_info = side_info[0].track_info;
	SectorInfo **sector_info = side_info[0].sector_info;
	unsigned char ***sector_data = side_info[0].sector_data;
	for (int track = 0; track < info.tracks; track++)
	{
		int sectors =  track_info[track].sectors;
#if 0
		if (sectors != 10 && track + 1 < info.tracks)
		{
			return "JV1 image has non-10 sector track";
		}
#endif

		// Put sectors in strict 0 .. 9 order, no gaps, no extras
		int order[10];
		for (int i = 0; i < 10; i++) {
			order[i] = -1;
		}

		for (int sec = 0; sec < sectors; sec++) {
			int snum = sector_info[track][sec].si_sector;
			if (snum < 0 || snum > 9) {
				return "JV1 image has sector ID greater than 9";
			}
			if (order[snum] != -1) {
				return "JV1 image has duplicate sector ID";
			}

			order[snum] = sec;
		}
#if 0
		// This is OK if we pad...
		for (int i = 0; i < 10; i++) {
			if (order[i] == -1) {
				return "JV1 image missing sector";
			}
		}
#else
		sectors = 10;
#endif

		for (int s = 0; s < sectors; s++)
		{
			int writelen = 0;
			int seclen = 256;
			int sec = order[s];
			if (sec < 0) {
				// pad out the missing sector
				char buf[256];
				memset(buf, 0, sizeof(buf));
				writelen = fwrite(buf, 1, seclen, fp);
			}
			else {
				seclen = sector_info[track][sec].si_length;
				if (seclen != 256)
				{
					return "JV1 image has non-256 byte sector";
				}
				writelen = fwrite(sector_data[track][sec], 1, seclen, fp);
			}
			if (writelen == 0)
			{
				return FileError();
			}
			if ( writelen != seclen )
			{
				return "fwrite lost bytes";
			}
		}
	}

	dirty = 0;

	return NULL;
}


DiskDrive* fd_create_jv1(const DriveConfiguration &config, const DiskImageMetadata* metadata)
{
	return new JV1DiskDrive(config, metadata);
}
