#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <ctype.h>

#ifdef _MSC_VER
#define stricmp _stricmp
#else
#define stricmp strcasecmp
#endif

#if defined(__APPLE__) || defined(__linux__)
#define _strnicmp strncasecmp
#endif

#include "fd_image.h"

typedef struct {
  unsigned char track;
  unsigned char sector;
  unsigned char flags;
} SectorHeader;

// In one sense, this is simple with each header block mapping to a sector.
// However, the flags can change the sector size.
// Which means you must go through and create a mapping since
// the sectors are variable-length and packed tightly.
// Because of the sector header, you must keep the file well-truncated.
// Also, you have to read in the table and at least partially sort it
// so all the sectors are in some kind of order.
// Or, not; the track reader can simple scan the whole table.
typedef struct {
  SectorHeader headers1[JV3_RECORD_COUNT];
  unsigned char writeprot;
  //unsigned char data1[];
  // Could have a second set of headers, padding and data
} JV3;

#define JV3_DENSITY     0x80  /* 1=dden, 0=sden */
#define JV3_DAM         0x60  /* data address mark code; see below */
#define JV3_SIDE        0x10  /* 0=side 0, 1=side 1 */
#define JV3_ERROR       0x08  /* 0=ok, 1=CRC error */
#define JV3_NONIBM      0x04  /* 0=normal, 1=short */
#define JV3_SIZE        0x03  /* in used sectors: 0=256,1=128,2=1024,3=512
                                 in free sectors: 0=512,1=1024,2=128,3=256 */

#define JV3_FREE        0xFF  /* in track and sector fields of free sectors */
#define JV3_FREEF       0xFC  /* in flags field, or'd with size code */


JV3DiskDrive::JV3DiskDrive(const DriveConfiguration &config, const DiskImageMetadata* metadata)
	: BlockDrive(config, metadata)
{
	numSector = 0;

	// No entries in secInfo[] are legitimate
	secInfoValid = 0;
	memset(secInfo, 0xff, sizeof(secInfo));

	// Make sure len isn't zero otherwise no index pulse
	len = sizeof(secInfo);
	int size_ok = ensure_disk_size(len);
	assert(size_ok);
}

// Used by diskinfo for debugging
//#define JV3_VERBOSE

// secInfoValid code is used so that saving follows
// the same sector order (and free blocks) as loading.
// A bit complicated but makes it easy to verify that load/save
// works properly for JV3 by binary comparison.
int JV3DiskDrive::Load(unsigned char* data, int count)
{
	DeleteAllSectors();

	int eof_once = 1;

	int non_free_len[4] = { 256, 128, 1024, 512 };
	int size_ok = ensure_disk_size(count);
	assert(size_ok);

	len = count;
	memcpy(disk, data, count);

    int maxtrack = 0;
    int mintrack = 1000;
    int seccount[2][256];
	for(int side=0; side < 2; side++) {
	    for(int i=0;i<256;i++) {
	        seccount[side][i] = 0;
		}
    }
    
	// Scan through the sector headers.
	int alloc = JV3_RECORD_COUNT*3 + 1;// 0x2200
	int prevtrack=-1;
	numSector = 0;
	int numsides = 1;
	secInfoValid = 1;
#ifdef JV3_VERBOSE
	int badCrc=0,nonIbm=0,notInOrder=0,fancyDam=0,hitFree=0, freeSeg=0,non256=0;
#endif
	for (int i = 0; i < JV3_RECORD_COUNT*3; i += 3) {
		int trk, sec, flags;
		trk = disk[i];
		sec = disk[i + 1];
		flags = disk[i + 2];
		// Note: need to keep track of space consumed
		// and look for end of file.  There could be another
		// sector header there.
		JV3SectorInfo *info = &secInfo[i / 3];
		if (trk != 0xff && sec != 0xff) {
#ifdef JV3_VERBOSE
			if (hitFree != 0)
			{
				freeSeg += hitFree;
				hitFree = 0;
			}
			if (prevtrack > 0 && trk != prevtrack + 1 && trk != prevtrack) {
				notInOrder++;
			}
#endif
			if (prevtrack != trk) {
				prevtrack = trk;
				//printf("\n");
			}
            if (trk > maxtrack) {
                maxtrack = trk;
            }
            if (trk < mintrack) {
                mintrack = trk;
            }
			info->side = !!(flags & 0x10);
            seccount[info->side][trk]++;
			info->trk = trk;
			info->sec = sec;
			info->density = !!(flags & 0x80);
			info->dam = ((flags & 0x60) >> 5);
			if (info->side) numsides = 2;
			info->badCRC = !!(flags & 0x08);
			info->nonibm = !!(flags & 0x04);
			info->lenCode = (flags & 0x3);
			info->len = non_free_len[(flags & 0x3)];
			info->data = &disk[alloc];
			info->sec_index = -1;

			// sec_index has to be based on "track_info[track].sectors" calculated below
			info->raw_flags = flags;
			//printf("%02d %02d %02x (%s) dam=%d, side=%d, err=%d, !ibm=%d, len=%d @%05x\n",
			//	info->trk, info->sec, flags,
			//	info->density ? "DD" : "SD",
			//	info->dam,
			//	info->side,
			//	info->crcOK,
			//	info->nonibm,
			//	info->len,
			//	alloc);
#ifdef JV3_VERBOSE
			if (info->dam > 1) {fancyDam++;}
			badCrc += info->badCRC;
			nonIbm+=info->nonibm;
			non256 += (info->len != 256);
#endif

			numSector++;
			if (alloc > count)
			{
				if (eof_once)
				{
					eof_once = 0;
					printf("ERROR: jv3 track data past end of file\n");
					assert(0);//ERROR: TODO: no way to have errors on load? (or warnings)
				}
			}
		}
		else
		{
			// Add as free sector
			info->trk = info->sec = 0xff;
			info->lenCode = ((flags & 0x3)  ^ 0x3);
			info->len = non_free_len[info->lenCode];
			info->data = NULL;
			info->sec_index = -2;
			info->raw_flags = flags;
			if (alloc < count) {
				info->data = &disk[alloc];
			}
#ifdef JV3_VERBOSE
			hitFree++;
#endif
		}
		if (alloc < count) {
			alloc += info->len;
		}

	}
	//printf("\n");
	writable = disk[JV3_RECORD_COUNT*3];
	//printf("writeable %02x\n", writable);
	//printf("alloc = %05x, file length=%05x\n", alloc, count);
	if (writable != 0x00 && writable != 0xff)
		printf("bad jv3 writable flag: 0x%02x\n", writable);
#ifdef JV3_VERBOSE
	if (/*badCrc != 0 || */nonIbm != 0 || notInOrder != 0 || fancyDam != 0 || freeSeg != 0 || non256 != 0)
	{
		printf("jv3: sectors=%d", numSector);
		if (badCrc != 0) printf(" badCrc=%d", badCrc);
		if (nonIbm != 0) printf(" nonIbm=%d", nonIbm);
		if (fancyDam != 0) printf(" fancyDam=%d", fancyDam);
		if (notInOrder != 0) printf(" notInorder=%d", notInOrder);
		if (freeSeg != 0) printf(" freeSeg=%d", freeSeg);
		if (non256 != 0) printf(" non256=%d", non256);
		printf("\n");
	}
#endif
    //printf("track range=%d,%d\n", mintrack, maxtrack);
    //for (int i = mintrack; i < maxtrack; i++) {
    //    printf(" %2d", seccount[i]);
    //}
    //printf("\n");
    
	// Build track info
	if (mintrack != 0) {
		printf("JV3: mintrack=%d\n", mintrack);
	}
	assert(mintrack == 0);//ERROR

    info.sides = numsides;
    info.tracks = maxtrack + 1;

	side_info = new SideInfo[info.sides];
	for (int side = 0; side < info.sides; side++)
	{
		TrackInfo *track_info = side_info[side].track_info = new TrackInfo[info.tracks];
		SectorInfo** sector_info = side_info[side].sector_info = new SectorInfo*[info.tracks];
		unsigned char ***sector_data = side_info[side].sector_data = new unsigned char**[info.tracks];

		for (int track = 0; track < info.tracks; track++)
		{
			int sectors = seccount[side][track];
			track_info[track].sectors = 0;
			sector_info[track] = new SectorInfo[sectors];
			sector_data[track] = new unsigned char*[sectors];
		}

		for (int i = 0; i < JV3_RECORD_COUNT; i++)
		{
			JV3SectorInfo *info = &secInfo[i];
			if (info->trk == 0xff && info->sec == 0xff)
			{
				// free sector, ignore
				continue;
			}
			if (info->side != side)
			{
				continue;
			}

			int track = info->trk;
			int sec = track_info[track].sectors;
			info->sec_index = sec;
			track_info[track].sectors++;

			sector_info[track][sec].si_length = info->len;
			sector_info[track][sec].si_side = info->side;
			sector_info[track][sec].si_track = track;
			sector_info[track][sec].si_sector = info->sec;
			sector_info[track][sec].si_flags = 0;
#ifdef JV3_SIDE_FIX
			// Not really a side problem.  data_rate is a bit of mess
			sector_info[track][sec].set_rate(m_config.data_rate / 1000);
#endif
			if (info->density) {
				sector_info[track][sec].si_flags |= SECTOR_dden;
			}
			if (info->badCRC) {
				sector_info[track][sec].si_flags |= SECTOR_err;
			}
			// TODO: info->nonIBM (maybe that is covered by length)
			if (info->dam & 1) {
				// TODO: SECTOR_DELETE is incomplete
				sector_info[track][sec].si_flags |= SECTOR_delete;
			}
			if (info->dam & 2) {
				sector_info[track][sec].si_flags |= SECTOR_damhigh;
			}

			sector_data[track][sec] = new unsigned char[info->len];
			memcpy(sector_data[track][sec], info->data, info->len);
		}
	}

    current_track = 0;
	radial_position = 0;
	tc_track = 0;
    
	CreateTrackProgram();

	return len;
}

void JV3DiskDrive::NotifyWriteTrack(int writeTrackOn)
{
	if (writeTrackOn)
		secInfoValid = 0;

	BlockDrive::NotifyWriteTrack(writeTrackOn);
}

#if 0
 typedef struct {
 unsigned char track;
 unsigned char sector;
 unsigned char flags;
 } SectorHeader;
 
#define JV3_DENSITY     0x80  /* 1=dden, 0=sden */
#define JV3_DAM         0x60  /* data address mark code; see below */
#define JV3_SIDE        0x10  /* 0=side 0, 1=side 1 */
#define JV3_ERROR       0x08  /* 0=ok, 1=CRC error */
#define JV3_NONIBM      0x04  /* 0=normal, 1=short */
#define JV3_SIZE        0x03  /* in used sectors: 0=256,1=128,2=1024,3=512
in free sectors: 0=512,1=1024,2=128,3=256 */
#endif

const char* JV3DiskDrive::WriteSecInfo(FILE *fp, const SectorInfo* info)
{
    int jv3 = 0;
    int flags = info->si_flags;
    if (flags & SECTOR_dden) {
        jv3 |= 0x80;
    }
    if (flags & SECTOR_delete) {
        jv3 |= 0x20;
    }
	if (flags & SECTOR_damhigh) {
		jv3 |= 0x40;
	}
    if (info->si_side != 0) {
        jv3 |= 0x10;
    }
    if (flags & SECTOR_err) {
        jv3 |= 0x08;
    }

	switch (info->si_length)
	{
		case 256:  jv3 |= 0; break;
		case 128:  jv3 |= 1; break;
		case 1024: jv3 |= 2; break;
		case 512:  jv3 |= 3; break;
		default:
			return "JV3 unsupported sector length";
	}

    if (fputc(info->si_track, fp) == EOF) return FileError();
    if (fputc(info->si_sector, fp) == EOF) return FileError();
    if (fputc(jv3, fp) == EOF) return FileError();

	return NULL;
}

const char* JV3DiskDrive::WriteSecData(FILE* fp, int seclen, const unsigned char* data)
{
	int writelen = fwrite(data, 1, seclen, fp);
	if (writelen == 0)
		return FileError();
	if ( writelen != seclen )
		return "fwrite lost bytes";

	return NULL;
}

const char* JV3DiskDrive::Save(FILE *fp)
{
    // TODO: 2901 sector limit must be obeyed (start second header)
	int recordCount = 0;
	for (int side = 0; side < info.sides; side++)
	{
		// Mark all sectors as not saved
		for (int track = 0; track < info.tracks; track++)
		{
			int sectors =  side_info[side].track_info[track].sectors;
			for (int sec = 0; sec < sectors; sec++)
			{
				side_info[side].sector_info[track][sec].si_flags |= SECTOR_savemarker;
			}
		}
	}

	int total = 0;
	if (secInfoValid)
	{
		// Output sectorInfo in secInfo order
		for (int i = 0; i < JV3_RECORD_COUNT && total < numSector; i++)
		{
			JV3SectorInfo *info = &secInfo[i];
			if (info->trk == 0xff && info->sec == 0xff)
			{
				if (fputc(info->trk, fp) == EOF) return FileError();
				if (fputc(info->sec, fp) == EOF) return FileError();
				if (fputc(info->raw_flags, fp) == EOF) return FileError();
			}
			else
			{
				const char* err = WriteSecInfo(fp, &side_info[info->side].sector_info[info->trk][info->sec_index]);
				if (err) return err;
				side_info[info->side].sector_info[info->trk][info->sec_index].si_flags &= ~SECTOR_savemarker;
				total++;
			}
			recordCount++;
		}
	}

	// Output header info if not already saved
	for (int side = 0; side < info.sides; side++)
	{
		for (int track = 0; track < info.tracks; track++)
		{
			int sectors =  side_info[side].track_info[track].sectors;
			for (int sec = 0; sec < sectors; sec++)
			{
				if (side_info[side].sector_info[track][sec].si_flags & SECTOR_savemarker)
				{
					const char* err = WriteSecInfo(fp, &side_info[side].sector_info[track][sec]);
					if (err) return err;
					total++;
					recordCount++;
				}
			}
		}
	}
    // pad out to 2901 entries.
    for (; recordCount < JV3_RECORD_COUNT; recordCount++) {
        if (fputc(0xff, fp) == EOF) return FileError();
        if (fputc(0xff, fp) == EOF) return FileError();
        if (fputc(0xff, fp) == EOF) return FileError();
    }

	// Be strict about writing either 0xff for writable, 0x00 for read-only
	if (fputc(writable ? 0xff : 0x00, fp) == EOF) return FileError();
    
    // Now data
	total = 0;
	if (secInfoValid)
	{
		// Output sector data in secInfo order
		for (int i = 0; i < JV3_RECORD_COUNT && total < numSector; i++)
		{
			JV3SectorInfo *info = &secInfo[i];
			if (info->trk == 0xff && info->sec == 0xff)
			{
				if (info->data != NULL)
				{
					// Write free data
					const char* err = WriteSecData(fp, info->len, info->data);
					if (err) return err;
				}
			}
			else
			{
				const char* err = WriteSecData(fp, side_info[info->side].sector_info[info->trk][info->sec_index].si_length, side_info[info->side].sector_data[info->trk][info->sec_index]);
				if (err) return err;
				side_info[info->side].sector_info[info->trk][info->sec_index].si_flags &= ~SECTOR_savemarker;
				total++;
			}
		}
	}

	for (int side = 0; side < info.sides; side++)
	{
		for (int track = 0; track < info.tracks; track++)
		{
			int sectors =  side_info[side].track_info[track].sectors;
			for (int sec = 0; sec < sectors; sec++)
			{
				if (side_info[side].sector_info[track][sec].si_flags & SECTOR_savemarker)
				{
					/*const char* err =*/ WriteSecData(fp, side_info[side].sector_info[track][sec].si_length, side_info[side].sector_data[track][sec]);
					side_info[side].sector_info[track][sec].si_flags &= ~SECTOR_savemarker;
				}
			}
		}
	}
    
	dirty = 0;

	return NULL;
}

DiskDrive* fd_create_jv3(const DriveConfiguration &config, const DiskImageMetadata* metadata)
{
	return new JV3DiskDrive(config, metadata);
}
