// Reusable utilities
FILE* OpenOrCreateFile(const char* path, bool& read_only, Notifier* report);
FILE* OpenFileOnly(const char* path, bool& read_only);
int HasExtension(const char* name, const char* ext);
char *ReplaceExtension(const char* path, const char *ext);

enum HardDiskImageFormat
{
	DSK_HDV1,
	DSK_EMU,
	DSK_DREM,
} ;

#define ENABLE_HDV_UPGRADE

#ifdef ENABLE_HDV_UPGRADE
char* hdv1_model2_upgrade(const char *filename, FILE* fp, const char* op);
#endif

class HardDriveImage;
class HardDrive;

class HardDrive
{
public:
	HardDrive()
		: attached(0)
		, current_cylinder(0)
		, dsk(NULL)
	{
	}

	int attached;
	int current_cylinder;
	// Geometry info...?
	DriveConfiguration m_config;

	HardDriveImage* dsk;
} ;

// Deviated a bit from BlockInterface
class HardDriveBlockInterface
{
public:
	virtual int IsModified() { return 1;} // TODO: pure virtual = 0;

	virtual void GetDiskInfo(DiskInfo& info) {};
	virtual void GetTrackInfo(TrackInfo& info, int head, int cylinder) {};

    virtual void GetSectorInfo(SectorInfo& info, int head, int cylinder, int sector_id, int expected_length = 0) = 0;
    virtual void GetSectorData(unsigned char* res, const SectorInfo& info) = 0;
	virtual void PutSectorData(const unsigned char* res, const SectorInfo& info) = 0;

	virtual void NotifyFormat(int cylinder, int head, int sector_count, int sector_size, unsigned char* buffer, int gap3) {}

	// MinSectorID is a placeholder until SectorInfo and physical sector indices are handled properly
	virtual int GetMinSectorID(int head, int cylinder) = 0;
} ;

class HardDriveImage : public HardDriveBlockInterface
{
public:
	Notifier* report;

	HardDriveImage(Notifier* err)
		: report(err)
	{
		// Placeholder values (Original 8MB Tandy drive)
		m_config.rpm = 3125;
		m_config.data_rate = 4340000;
		m_config.sides = 4;
		m_config.tracks = 256;
		m_config.sectors_per_track = 17;
		drive = 0;
	}
	// struct DiskInfo is almost enough--but we need sectors per track
	// We need to interpret: tracks => cylinders, sides => heads

	// You want to design to allow other formats.  BlockInterface is a good start.

	// Also plan for detection code.  One difference with hard drives
	// as we'll leave the file open.

	// Add an error/warning interface.

	// And: the "drive" state needs to be separate.  At least
	// the current track should not be managed by the image.

	virtual const DriveConfiguration* GetConfig() { return &m_config; }
	virtual HardDiskImageFormat ImageFormat() = 0;

	void Fail(const char* message);
	void Warn(const char* message);
	virtual void Load(FILE* fp, bool read_only, const DiskImageMetadata *stamp = NULL){};
	virtual void SaveToFile(){};
	virtual int Flush(){return 0;}

	HardDrive* GetDrive(void) { return drive; }

protected:
	DriveConfiguration m_config;
	HardDrive* drive;
} ;


// Disk image of raw blocks--geometry must be specified.
class BlockHardDrive : public HardDriveImage
{
public:
	BlockHardDrive(Notifier* err);

	virtual HardDiskImageFormat ImageFormat() { return DSK_DREM;  }

	void SetGeometry(int heads, int cylinders, int sectors_per_track, int sector_size, int interleave, int min_id);

	virtual void GetTrackInfo(TrackInfo& info, int head, int cylinder);

    virtual void GetSectorInfo(SectorInfo& info, int head, int cylinder, int sector_id, int expected_length = 0);
    virtual void GetSectorData(unsigned char* res, const SectorInfo& info);
	virtual void PutSectorData(const unsigned char* res, const SectorInfo& info);

	virtual void NotifyFormat(int cylinder, int head, int sector_count, int sector_size, unsigned char* buffer, int gap3);

	virtual int GetMinSectorID(int head, int cylinder);

	virtual void Load(FILE* fp, bool read_only, const DiskImageMetadata *stamp = NULL);
	virtual void SaveToFile();

protected:

	int SeekToSector(int head, int cylinder, int sector);

	int cylinders;
	int heads;
	int sectors_per_head;
	int sector_length;

	int min_sector_id;

	int interleave;

	int m_file_len;

	FILE* m_file;
} ;

HardDriveImage* OpenHardDriveImage(FILE *fp, bool read_only, const char* path, const char* format_name, const unsigned char *data_in, int len, Notifier* report, const DriveConfiguration& config, const DiskImageMetadata *stamp);

#define EMU_DRIVE_MAGIC 0xee, 'M', 'F', 'M', 0x0d, 0x0a, 0x1a, 0x00

HardDriveImage* hd_create_hdv1(Notifier* report);
HardDriveImage* hd_create_emu(Notifier* report);

// DREM creation/loading is awkward (two files)
HardDriveImage* hd_create_drem(Notifier* report, const char* path, bool& read_only, FILE*& fp);
void hd_image_create_blank_drem(FILE *fp, Notifier* report);
