#include <stdio.h>
#include <string.h>
#include <math.h>

#include <assert.h>

#include "fd_image.h"
#include "hd_image.h"

#ifdef _MSC_VER
#define stricmp _stricmp
#define strnicmp _strnicmp
#else
#define stricmp strcasecmp
#define strnicmp strncasecmp
#endif

class IniFile
{
public:
	IniFile(Notifier* err);
	~IniFile();

	void Load(FILE* fp, bool readonly);
	int ReadParameter(const char* section, const char* id, int &value);
	int WriteParameter(const char* section, const char* id, int value);

	bool HasSection(const char* section);
private:
	Notifier* report;
	FILE *m_file;
	bool m_read_only;
	struct Line {
		char* m_line;
		Line* m_next;
	} ;
	Line* m_list;

	const char* SkipSpace(const char* str);
	Line* FindSection(const char* section);
	Line* FindParameter(const char* section, const char* id, int &offset);
} ;

class DREMHardDrive : public BlockHardDrive
{
public:
	DREMHardDrive(Notifier* err);
	IniFile& GetIni() {
		return m_ini;
	}

	virtual ~DREMHardDrive(){}

	virtual HardDiskImageFormat ImageFormat() { return DSK_DREM; }

	//void InitBlank(const DiskImageMetadata *stamp);
	void InitBlankConfig(FILE *fp);
	virtual void NotifyFormat(int cylinder, int head, int sector_count, int sector_size, unsigned char* buffer, int gap3);

private:
	IniFile m_ini;
} ;

// Simple INI file parser, e.g.
// [Section]
// # comment
// parameter=value
// [Section2]
// etc.
IniFile::IniFile(Notifier* err)
	: report(err)
	, m_file(NULL)
	, m_read_only(false)
	, m_list(NULL)
{
}

IniFile::~IniFile()
{
	for (Line* pHead = m_list; pHead != NULL; )
	{
		Line* remove = pHead;
		pHead = remove->m_next;

		delete [] remove->m_line;
		delete remove;
	}

	m_list = NULL;
}

void IniFile::Load(FILE* fp, bool readonly)
{
	m_read_only = readonly;
	m_file = fp;

	if (fseek(fp, 0, SEEK_SET) != 0)
	{
		report->Notify(Notifier::NOTIFY_Fail, "Seek Error");

		return;
	}

	// Read lines into linked list.
	// Don't care much if they are truncated.
	char buf[4096];
	Line** pHead = &m_list;
	while (fgets(buf, sizeof(buf), m_file))
	{
		Line* next = new Line;
		next->m_line = new char[strlen(buf)+ 1];
		strcpy(next->m_line, buf);
		next->m_next = NULL;

		*pHead = next;
		pHead = &next->m_next;
	}
}

const char* IniFile::SkipSpace(const char* str)
{
	while (str[0] != 0 && (str[0] == ' ' || str[0] == '\t')) {
		str++;
	}

	return str;
}

//TODO: strstr() logic should really be an optional parameter.
IniFile::Line* IniFile::FindSection(const char* section)
{
	// Find section
	Line* pHead;
	for (pHead = m_list; pHead != NULL; pHead = pHead->m_next)
	{
		// The rule here is a bit weak for compatibilty with older DREM files
		const char* line = SkipSpace(pHead->m_line);
		if (line[0] == '[' && strstr(line + 1, section) != NULL)
		{
			// Skip the header
			pHead = pHead->m_next;
			break;
		}
	}

	return pHead;
}

IniFile::Line* IniFile::FindParameter(const char* section, const char* id, int &offset)
{
	// Find section
	Line* pHead = FindSection(section);

	if (pHead != NULL)
	{
		// Find parameter
		int idlen = strlen(id);
		for ( ; pHead != NULL; pHead = pHead->m_next)
		{
			const char* line = SkipSpace(pHead->m_line);
			if (line[0] == '[')
				break;
			if (line[0] != '#')
			{
				if (strnicmp(line, id, idlen) == 0)
				{
					line = SkipSpace(line + idlen);
					if (line[0] == '=')
					{
						line = SkipSpace(line + 1);
						offset = line - pHead->m_line;

						return pHead;
					}
				}
			}
		}
	}

	return NULL;
}

int IniFile::ReadParameter(const char* section, const char* id, int &value)
{
	int offset;
	Line* pHead = FindParameter(section, id, offset);
	if (pHead != NULL)
	{
		const char* line = pHead->m_line + offset;
		sscanf(line, "%d", &value);
		return 1;
	}

	return 0;
}

int IniFile::WriteParameter(const char* section, const char* id, int value)
{
	int offset;
	Line* pHead = FindParameter(section, id, offset);
	if (pHead != NULL)
	{
		// Preserve a bit of the formatting, but not any trailing comment
		char text[256];
		sprintf(text, "%d", value);

		char *new_line = new char[offset + strlen(text) + 1 + 1];
		memcpy(new_line, pHead->m_line, offset);
		sprintf(new_line + offset, "%s\n", text);
		delete [] pHead->m_line;
		pHead->m_line = new_line;

		//TODO: "DIRTY" flag
		return 1;
	}

	return 0;
}


bool IniFile::HasSection(const char* section)
{
	return FindSection(section) != NULL;
}

DREMHardDrive::DREMHardDrive(Notifier* err)
	: BlockHardDrive(err)
	, m_ini(err)
{
}

//void DREMHardDrive::InitBlank(const DiskImageMetadata *stamp)
void DREMHardDrive::InitBlankConfig(FILE *fp)
{
	// Tricky: InitBlank() is called from Load
	// The block drive image itself is just an empty file.
	// For the Ini file, we need to fopen(), fprintf(), and read back.

// TODO: should be using meta-data and meta-data passed in: const DiskImageMetadata *stamp

	fprintf(fp,
		"# DREM config file created by trs80gp DATE\n"			// Supposed to put "DATE" in there.
		"[DSK]\n"	// Version 2.  Version 1 is "DREM DSK Config"
		"Name=TRS-80 Hard Drive\n"
		"Encoding=MFM\n"
		"RPM=3600\n"
		"Tracks=%d\n"
		"Sides=%d\n"
		"Sectors=%d\n"
		"First Sector ID=%d\n"
		"Sector Size=%d\n"
		"Format=WDC\n"
		"Step Signal Hold=8\n"
		"Interleave=%d\n"
		"Data ECC=NO\n",
		cylinders,
		heads,
		sectors_per_head,
		min_sector_id,
		sector_length,
		interleave
		);
}

void DREMHardDrive::NotifyFormat(int cylinder, int head, int sector_count, int sector_size, unsigned char* buffer, int gap3)
{
	int min = -1;
	int max = -1;
	for (int i = 0; i < sector_count; i++)
	{
		assert(buffer[i*2] == 0);		// if 0x80, a bad block, we don't support it.
		int snum = buffer[i*2+1];
		if (min == -1 || snum < min) min = snum;
		if (max == -1 || snum > max) max = snum;
	}
	if (min == -1 || max == -1 || (max - min + 1) != sector_count)
	{
		Warn("we will have some problems\n");
	}
	if (min_sector_id != min)
	{
		min_sector_id = min;
		(void)m_ini.WriteParameter("DSK", "First Sector ID", min_sector_id);
	}

	if (sectors_per_head != sector_count)
	{
		sectors_per_head = sector_count;
		(void)m_ini.WriteParameter("DSK", "Sectors", sectors_per_head);
	}
	if (sector_length != sector_size)
	{
		sector_length = sector_size;
		(void)m_ini.WriteParameter("DSK", "Sector Size", sector_length);
	}
	if (cylinder > cylinders)
	{
		cylinders = cylinder;
		(void)m_ini.WriteParameter("DSK", "Tracks", cylinders);
	}
	if (head > heads)
	{
		heads = head;
		(void)m_ini.WriteParameter("DSK", "Sides", heads);
	}

	// Detect interleave of the sector buffer.
	// too complicated.  Just look at first and second sector.
	int computed_interleave = -1;
	int base_snum = buffer[1];
	for (int test_interleave = 1; test_interleave <= sector_count; test_interleave++) {
		for (int i = 1; i < sector_count; i++)
		{
			int j;
			for (j = 1; j < sector_count; j++)
			{
				int offset = (test_interleave * j) % sector_count;


				// Should be the right idea
				if (buffer[offset*2+1] != base_snum + j)
				{
					break;
				}
			}
			if (j >= sector_count)
			{
				computed_interleave = test_interleave;
				break;
			}
		}
	}

	if (computed_interleave > 0 && computed_interleave != interleave)
	{
		interleave = computed_interleave;
		(void)m_ini.WriteParameter("DSK", "Interleave", interleave);
	}
}

HardDriveImage* hd_create_drem(Notifier* report, const char* path, bool& read_only, FILE*& fp)
{
	// Try to match the case of the extension, defaulting to lower case
	const char *ext_order[2] = { "dsk", "DSK" };
	if (strlen(path) > 4 && strcmp(path + strlen(path) - 4, ".CFG") == 0)
	{
		const char* tmp = ext_order[0];
		ext_order[0] = ext_order[1];
		ext_order[1] = tmp;
	}

	FILE *dsk = NULL;
	for (int i = 0; i < 2; i++)
	{
		char* dsk_path = ReplaceExtension(path, ext_order[i]);

		dsk = OpenFileOnly(dsk_path, read_only);
		delete [] dsk_path;
		if (dsk != NULL)
		{
			break;
		}
	}

	if (dsk == NULL)
	{
		// Will use upper case if original ends in .CFG
		char* dsk_path = ReplaceExtension(path, ext_order[0]);

		dsk = OpenOrCreateFile(dsk_path, read_only, report);
		delete [] dsk_path;
		if (report->m_note_count[Notifier::NOTIFY_Fail] > 0)
		{
			return NULL;
		}
	}

	// TODO: Push code into DREMHardDrive

	IniFile config(report);
	config.Load(fp, true);

	// Give a bit more context if we can't find a section
	int heads, tracks, sectors, sector_size, min_id, interleave = 0;
	const char* drem_1 = "DREM DSK Config";	// DREM version 1
	const char* drem_2 = "DSK";	// DREM version 2 (but will match any [ ... DSK ... ] header)
	const char* section = NULL;

	if (config.HasSection(drem_2))
	{
		section = drem_2;
	}
	if (config.HasSection(drem_1))
	{
		section = drem_1;
	}

	if (section == NULL)
	{
		report->Notify(Notifier::NOTIFY_Fail, "Could not find [%s] or [%s] section header", drem_2, drem_1);
	}

	if (!config.ReadParameter(section, "sides", heads))
	{
		report->Notify(Notifier::NOTIFY_Fail, "Missing Sides= specification");
	}
	if (!config.ReadParameter(section, "tracks", tracks))
	{
		report->Notify(Notifier::NOTIFY_Fail, "Missing Tracks= specification");
	}
	if (!config.ReadParameter(section, "sectors", sectors))
	{
		report->Notify(Notifier::NOTIFY_Fail, "Missing Sectors= specification");
	}
	if (!config.ReadParameter(section, "sector size", sector_size))
	{
		report->Notify(Notifier::NOTIFY_Fail, "Missing Sector Size= specification");
	}
	if (!config.ReadParameter(section, "first sector id", min_id))
	{
		report->Notify(Notifier::NOTIFY_Fail, "Missing First Sector ID= specification");
	}
	// OK if interleave is missing
	(void)config.ReadParameter(section, "interleave", interleave);

	if (report->m_note_count[Notifier::NOTIFY_Fail] > 0)
	{
		if (dsk)
		{
			fclose(dsk);
		}

		return NULL;
	}

	fclose(fp);

	// Give caller new file handle
	fp = dsk;

	// replace filename .cfg extension with .dsk
	// look up Tracks, Sectors, Sides, Interleave (opt), "First Sector ID"
	// Can return the block drive--ultimately, format will require the .cfg
	// But...we need to find the .dsk
	BlockHardDrive *bhd = new BlockHardDrive(report);

	bhd->SetGeometry(heads, tracks, sectors, sector_size, 10, min_id);

	return bhd;
}

void hd_image_create_blank_drem(FILE *fp, Notifier* report)
{
	// TODO: HACK--create the 'blank' drive (will be 'normal' when DREMHardDrive replaces BlockDrive here)
	// This should become the "normal" path.

	//TODO: DREMHardDrive needs to stop inheriting from BlockHardDrive
	// and implement blocks itself (different m_file, too, not the .cfg one)
	DREMHardDrive *tmp = new DREMHardDrive(report);

	tmp->InitBlankConfig(fp);
	fseek(fp, 0, SEEK_SET);

	delete tmp; // warning: non-virtual destructors present
}

#if 0
// DREM code to pull out

// Add error delegate, produce errors if not HDV file (e.g. bad magic)
// replaces *&error at least.
HardDriveImage* OpenHardDriveImage(FILE *fp, bool read_only, const char* path, const char* format_name, const unsigned char *data_in, int len, Notifier* report, const DriveConfiguration& config, const DiskImageMetadata *stamp)
{
	//TODO: data means a pre-arranged image
	assert(data_in == NULL);

	HardDriveImage* drive_image = NULL;

	unsigned char buf[4096];
	if (fseek(fp, 0, SEEK_END) != 0)
	{
		report->Notify(Notifier::NOTIFY_Fail, "Seek Error");

		return NULL;
	}

	//[unused] long file_len = ftell(fp);
	if (fseek(fp, 0, SEEK_SET) != 0)
	{
		report->Notify(Notifier::NOTIFY_Fail, "Seek Error");

		return NULL;
	}

	int buf_len = fread(buf, 1, sizeof(buf), fp);

	// Seek back to start (EMU loading at least required it)
	if (fseek(fp, 0, SEEK_SET) != 0)
	{
		report->Notify(Notifier::NOTIFY_Fail, "Seek Error");

		return NULL;
	}

	if (strcmp(format_name, "emu") == 0)
	{
		extern void make_crc_table();

		make_crc_table();

		if (buf_len > 0 && !Detect_EMU(buf, buf_len))
		{
			report->Notify(Notifier::NOTIFY_Fail, "Not a DG EMU hard drive image");

			return NULL;
		}

		drive_image = hd_create_emu(report);
		drive_image->Load(fp, read_only, stamp);

		return drive_image;
	}
	else if (strcmp(format_name, "drem") == 0)
	{
		if (buf_len == 0)
		{
			// TODO: HACK--create the 'blank' drive (will be 'normal' when DREMHardDrive replaces BlockDrive here)
			// This should become the "normal" path.

			//TODO: DREMHardDrive needs to stop inheriting from BlockHardDrive
			// and implement blocks itself (different m_file, too, not the .cfg one)
			DREMHardDrive *tmp = new DREMHardDrive(report);

			tmp->InitBlankConfig(fp);
			fseek(fp, 0, SEEK_SET);

			delete tmp; // warning: non-virtual destructors present
		}

		if (buf_len > 0 && !Detect_DREM(buf, buf_len))
		{
			report->Notify(Notifier::NOTIFY_Fail, "Not a DREM hard drive config file");
			return NULL;
		}

		// Try to match the case of the extension, defaulting to lower case
		const char *ext_order[2] = { "dsk", "DSK" };
		if (strlen(path) > 4 && strcmp(path + strlen(path) - 4, ".CFG") == 0)
		{
			const char* tmp = ext_order[0];
			ext_order[0] = ext_order[1];
			ext_order[1] = tmp;
		}

		FILE *dsk = NULL;
		for (int i = 0; i < 2; i++)
		{
			char* dsk_path = ReplaceExtension(path, ext_order[i]);

			dsk = OpenFileOnly(dsk_path, read_only);
			delete [] dsk_path;
			if (dsk != NULL)
			{
				break;
			}
		}

		if (dsk == NULL)
		{
			// Will use upper case if original ends in .CFG
			char* dsk_path = ReplaceExtension(path, ext_order[0]);

			dsk = OpenOrCreateFile(dsk_path, read_only, report);
			delete [] dsk_path;
			if (report->m_note_count[Notifier::NOTIFY_Fail] > 0)
			{
				return NULL;
			}
		}

		// TODO: Push code into DREMHardDrive

		IniFile config(report);
		config.Load(fp, true);

		// Give a bit more context if we can't find a section
		int heads, tracks, sectors, sector_size, min_id, interleave = 0;
		const char* drem_1 = "DREM DSK Config";	// DREM version 1
		const char* drem_2 = "DSK";	// DREM version 2 (but will match any [ ... DSK ... ] header)
		const char* section = NULL;

		if (config.HasSection(drem_2))
		{
			section = drem_2;
		}
		if (config.HasSection(drem_1))
		{
			section = drem_1;
		}

		if (section == NULL)
		{
			report->Notify(Notifier::NOTIFY_Fail, "Could not find [%s] or [%s] section header", drem_2, drem_1);
		}

		if (!config.ReadParameter(section, "sides", heads))
		{
			report->Notify(Notifier::NOTIFY_Fail, "Missing Sides= specification");
		}
		if (!config.ReadParameter(section, "tracks", tracks))
		{
			report->Notify(Notifier::NOTIFY_Fail, "Missing Tracks= specification");
		}
		if (!config.ReadParameter(section, "sectors", sectors))
		{
			report->Notify(Notifier::NOTIFY_Fail, "Missing Sectors= specification");
		}
		if (!config.ReadParameter(section, "sector size", sector_size))
		{
			report->Notify(Notifier::NOTIFY_Fail, "Missing Sector Size= specification");
		}
		if (!config.ReadParameter(section, "first sector id", min_id))
		{
			report->Notify(Notifier::NOTIFY_Fail, "Missing First Sector ID= specification");
		}
		// OK if interleave is missing
		(void)config.ReadParameter(section, "interleave", interleave);

		if (report->m_note_count[Notifier::NOTIFY_Fail] > 0)
		{
			if (dsk)
			{
				fclose(dsk);
			}

			return NULL;
		}

		fclose(fp);

		// replace filename .cfg extension with .dsk
		// look up Tracks, Sectors, Sides, Interleave (opt), "First Sector ID"
		// Can return the block drive--ultimately, format will require the .cfg
		// But...we need to find the .dsk
		BlockHardDrive *bhd;
		drive_image = bhd = new BlockHardDrive(report);

		bhd->SetGeometry(heads, tracks, sectors, sector_size, 10, min_id);

		drive_image->Load(dsk, read_only, stamp);
	}
	else
	{
		if (buf_len > 0 && !Detect_HDV1(buf, buf_len))
		{
			report->Notify(Notifier::NOTIFY_Fail, "Not an HDV1 hard drive image");
			return NULL;
		}

		drive_image = hd_create_hdv1(report);
		drive_image->Load(fp, read_only, stamp);
	}

	if (drive_image)
	{
		//TODO: fetch geometry
	}

	return drive_image;
}

#include <errno.h>

FILE* OpenOrCreateFile(const char* filename, bool& read_only, Notifier* report)
{
	read_only = false;
	FILE *fp;

	bool readonly_ok = false;
	fp = fopen(filename, "rb");
	if (fp != NULL)
	{
		readonly_ok = true;
		fclose(fp);
	}

	fp = fopen(filename, "r+b");	// existing file, read/write
	if (fp == NULL)
	{
		fp = fopen(filename, "w+b");	// read/write but create or truncate
	}
	if (fp == NULL)
	{
		if (!readonly_ok)
		{
			report->Notify(Notifier::NOTIFY_Fail, "Could not create file %s: %s", filename, strerror(errno));

			return NULL;
		}

		fp = fopen(filename, "rb");		// read only
		if (fp == NULL)
		{
			report->Notify(Notifier::NOTIFY_Fail, "Could not open file %s: %s", filename, strerror(errno));

			return NULL;
		}
		read_only = true;
	}
	
	return fp;
}

FILE* OpenFileOnly(const char* filename, bool& read_only)
{
	read_only = false;
	FILE *fp;

	fp = fopen(filename, "r+b");	// existing file, read/write
	if (fp == NULL)
	{
		fp = fopen(filename, "rb");		// read only existing file.
		if (fp != NULL)
		{
			read_only = true;
		}
	}

	return fp;
}

char *ReplaceExtension(const char* path, const char *ext)
{
	int len = strlen(path);
	const char *p = path ? strrchr(path, '.') : 0;
	if (p)
	{
		len = p - path;
	}

	char* res = new char[len + 1 + strlen(ext) + 1];
	memcpy(res, path, len);
	res[len] = '.';
	strcpy(&res[len + 1], ext);

	return res;
}
#endif

