#include <stdio.h>
#include <string.h>
#include <math.h>

#include <assert.h>

#include "fd_image.h"
#include "hd_image.h"

#ifdef _MSC_VER
#define stricmp _stricmp
#define strnicmp _strnicmp
#else
#define stricmp strcasecmp
#define strnicmp strncasecmp
#endif

// David Gesswein hard drive emulator format.

struct EmuTrackSectorInfo
{
	EmuTrackSectorInfo()
		: m_data_offset(-1)
	{
	}

	SectorInfo m_info;
	unsigned int m_mark;
	int m_data_offset;

	// Position of data in MFM stream
	int m_mfm_index;
	int m_mfm_bit;
} ;

struct EmuTrack
{
	EmuTrack()
		: m_num_sector(0)
		, m_info(NULL)
		, m_data_len(0)
		, m_data(NULL)
		, m_mfm_len(0)
		, m_mfm_data(NULL)
	{
	}
	~EmuTrack()
	{
		delete [] m_info;
		delete [] m_data;
		delete [] m_mfm_data;
	}

	void reset()
	{
		m_num_sector = 0;
		m_data_len = 0;
		delete [] m_info;
		delete [] m_data;
		delete [] m_mfm_data;
		m_info = NULL;
		m_data = NULL;
		m_mfm_data = NULL;
	}

	int m_num_sector;
	EmuTrackSectorInfo *m_info;
	int m_data_len;
	unsigned char* m_data;

	int m_mfm_len;
	unsigned int* m_mfm_data;
} ;

class EmuHardDrive : public HardDriveImage
{
public:
	EmuHardDrive(Notifier* err);

	virtual const DriveConfiguration* GetConfig();
	virtual HardDiskImageFormat ImageFormat() { return DSK_EMU; }

	virtual void GetDiskInfo(DiskInfo& info);
    virtual void GetTrackInfo(TrackInfo& info, int head, int cylinder);

    virtual void GetSectorInfo(SectorInfo& info, int head, int cylinder, int sector_id, int expected_length = 0);
    virtual void GetSectorData(unsigned char* res, const SectorInfo& info);
	virtual void PutSectorData(const unsigned char* res, const SectorInfo& info);

	virtual void NotifyFormat(int cylinder, int head, int sector_count, int sector_size, unsigned char* buffer, int gap3);

	virtual int GetMinSectorID(int head, int cylinder);

	virtual void Load(FILE* fp, bool read_only, const DiskImageMetadata *stamp = NULL);
	virtual void SaveToFile();

	static unsigned char magic[8];

protected:
	void InitBlank(const DiskImageMetadata *stamp);
	void WriteHeader();
	void WriteEofMarker();

	int SeekToSector(int head, int cylinder, int sector);
	int scan_track(FILE *fp, int CYL, int HEAD, EmuTrack& track, int pass);

	unsigned char m_header[16];
	int m_header_length;
	// Version data
	unsigned char m_minor, m_major, m_file_type;
	int m_tracklen, m_track_header_len, m_cylinders, m_heads, m_bit_rate;
	int m_comment_length;
	char *m_comment;
	unsigned int m_note_length;
	char *m_note;
	unsigned int m_track_time;

	int m_current_track, m_current_head;

	EmuTrack m_currentTrack;

	FILE* m_file;
	int m_file_size;
	int m_modified;
	bool m_readonly;

	void SetComment(const char* comment);
	void SetNote(const char* note);

	int detected_sectors_per_track;
} ;

extern unsigned int update_crc(unsigned int crc, unsigned int data);

unsigned int mfm_encode(int prev_bit, int data, int skip_clock)
{
    unsigned int res = 0;
    int prev_zero = 0;
    if (!prev_bit) {
        prev_zero = 2;
    }
    for (int i = 7; i >= 0; i--)
    {
        res <<= 2;
        int bit = (data >> i) & 1;
        int clk = (((skip_clock >> i) & 1) << 1) ^ 2;
        if (bit) {
            res |= 1;           // always 01
            prev_zero = 0;
        } else {
            res |= (prev_zero & clk); // 00 unless previous zero then 10 (but skip_clock can override
            prev_zero = 2;
        }
    }
    
    return res;
}

//UGH: I should just have a single bit index
static void read_bit(const unsigned int* buf, int& index, int& bit, unsigned int& res)
{
	res <<= 1;
	res |= ((buf[index] >> bit) & 1);
	bit--;
	if (bit < 0) {
		bit = 31;
		index++;
	}
}

static void write_mfm_bit(unsigned int* buf, int& index, int& bit, unsigned int data)
{
	unsigned int mask = (1 << bit);
	if (data) {
		buf[index] |= mask;
	} else {
		buf[index] &= ~mask;
	}
	bit--;
	if (bit < 0) {
		bit = 31;
		index++;
	}
}

static void write_mfm_byte(unsigned int* buf, int& index, int& bit, unsigned int data)
{
	for (int i = 15; i >= 0; i--)
	{
		write_mfm_bit(buf, index, bit, (data >> i) & 1);
	}
}

static unsigned int sync(const unsigned int* buf, int& index, int& bit)
{
	unsigned int clock = 0;
	unsigned int data = 0;
	for (int i = 0; i < 8; i++)
	{
		read_bit(buf, index, bit, clock);
		read_bit(buf, index, bit, data);
	}

	data &= 0xff;
	clock &= 0xff;

	return data;
}

//#define NOISY
static unsigned int dump(int len, const unsigned int* buf, int& index, int& bit, unsigned int crc, unsigned char* dst)
{
#ifdef NOISY
	int lindex = 0;
	unsigned int line[16];
#endif

	for (int i = 0; i < len; i++)
	{
		unsigned int data = sync(buf, index, bit);
#ifdef NOISY
		char res[256];
		sprintf(res, " %02x", data);
		ods(res);
#endif
		crc = update_crc(crc, data);
		if (dst) {
			*dst++ = data;
		}

#ifdef NOISY
		line[lindex++] = data;
		//if ((i + 1) % 16 == 0)
		if (lindex >= 16)
		{
			lindex = 0;
			ods("   ");
			for (int j = 0; j < 16; j++)
			{
				if (line[j] >= ' ' && line[j] <= 127)
				{
					res[j] = (char)line[j];
				}
				else
				{
					res[j] = '.';
				}
			}
			res[16] = 0;
			ods(res);
			ods("\n");
		}
#endif
	}
	return crc;
}

static void idam(unsigned int id, const unsigned int* buf, int& index, int& bit, SectorInfo& info)
{
#ifdef NOISY
	char res[1024];
#endif

	unsigned int clo, sdh, sec, crc1, crc2;
	clo = sync(buf, index, bit);
	sdh = sync(buf, index, bit);
	sec = sync(buf, index, bit);
	crc1 = sync(buf, index, bit);
	crc2 = sync(buf, index, bit);

//FE=0..255, FF=256..511,FC=512-767,FD=768-1023
//WD2010	F6=1024...1279 F7=1280...1535 F4=1536...1791 F5=1792...2047
	int high_cyl[4] = { 512, 768, 0, 256 };
	int high_cyl2[4] = { 1536, 1792, 1024, 1280 };	// WD2010 2048 track support
	if (id & 0x08)
		clo += high_cyl[id & 3];
	else
		clo += high_cyl2[id & 3];

	unsigned int crc = 0xffff;
	crc = update_crc(crc, 0xa1);
	crc = update_crc(crc, id);
	crc = update_crc(crc, clo);
	crc = update_crc(crc, sdh);
	crc = update_crc(crc, sec);
	crc = update_crc(crc, crc1);
	crc = update_crc(crc, crc2);

#ifdef NOISY
	sprintf(res, "IDAM: %02x cyl=%03x sdh=%02x sec=%02x crc=%02x%02x [%04x]%s\n", id, clo, sdh, sec, crc1, crc2, crc, crc == 0 ? "" : " *WRONG*");
	//sprintf(res, "IDAM: %02x : ", id);
	ods(res);
	//dump(6, buf, index, bit);
	//ods("\n");
#endif

	static int sector_size[4] = { 256, 512, 1024, 128 };

	info.si_flags = SECTOR_dden;		// MFM encoding
	info.si_length = sector_size[(sdh >> 5) & 3];;
	info.si_sector = sec;
	info.si_track = clo;
	info.si_side = (sdh & 7);
}

static int dam(unsigned int mark, const unsigned int* buf, int& index, int& bit, SectorInfo& info, unsigned char* dst)
{
#ifdef NOISY
	char res[1024];
#endif

	unsigned int crc = 0xffff;
	crc = update_crc(crc, 0xa1);
	crc = update_crc(crc, mark);

#ifdef NOISY
	sprintf(res, "DAM: id=%02x len=%d\n", mark, info.si_length);
	ods(res);
#endif

	crc = dump(info.si_length, buf, index, bit, crc, dst);
	unsigned int crc1, crc2;
	crc1 = sync(buf, index, bit);
	crc2 = sync(buf, index, bit);
	crc = update_crc(crc, crc1);
	crc = update_crc(crc, crc2);
#ifdef NOISY
	sprintf(res, "   crc=%02x%02x [%04x]%s\n\n", crc1, crc2, crc, crc == 0 ? "" : " *WRONG*");
	ods(res);
#endif

	return crc;
}

unsigned char EmuHardDrive::magic[8] = { EMU_DRIVE_MAGIC };

EmuHardDrive::EmuHardDrive(Notifier* err)
	: HardDriveImage(err)
	, m_comment_length(0)
	, m_comment(NULL)
	, m_note_length(0)
	, m_note(NULL)
	, m_current_track(-1)
	, m_current_head(-1)
	, m_file(NULL)
	, detected_sectors_per_track(-1)
{
	m_config.data_rate = 5000000;
	m_config.rpm = 3600;
}

const DriveConfiguration* EmuHardDrive::GetConfig()
{
	// Update with current setting
	m_config.sectors_per_track = detected_sectors_per_track > 0 ? detected_sectors_per_track : 32;
	m_config.sides = m_heads;
	m_config.tracks = m_cylinders;

	return &m_config;
}

void EmuHardDrive::GetDiskInfo(DiskInfo& info)
{
	info.tracks = m_cylinders;
	info.sides = m_heads;
}

void EmuHardDrive::GetTrackInfo(TrackInfo& info, int head, int cylinder)
{
	// Use arbitrary sector id, ignore result
	(void)SeekToSector(head, cylinder, 0);

	info.sectors = m_currentTrack.m_num_sector;
}

void EmuHardDrive::GetSectorInfo(SectorInfo& info, int head, int cylinder, int sector_id, int expected_length)
{
	int sector_index = SeekToSector(head, cylinder, sector_id);
	if (sector_index < 0)
	{
		info.si_flags = SECTOR_unavail;

		return;
	}
	assert(sector_index >= 0);

	info = m_currentTrack.m_info[sector_index].m_info;
}

void EmuHardDrive::GetSectorData(unsigned char* res, const SectorInfo& info)
{
	int sector_index = SeekToSector(info.si_side, info.si_track, info.si_sector);
	assert(sector_index >= 0);
	assert(info.si_length == m_currentTrack.m_info[sector_index].m_info.si_length);

	memcpy(res, &m_currentTrack.m_data[ m_currentTrack.m_info[sector_index].m_data_offset ], m_currentTrack.m_info[sector_index].m_info.si_length);
}

unsigned int write_byte(unsigned int& prev, unsigned int data, unsigned int clk_skip, unsigned int* hack, int& hack_index, int& hack_half, unsigned int crc)
{
	crc = update_crc(crc, data);

	unsigned int encode = mfm_encode(prev & 1, data, clk_skip);

	if (hack_half == 0)
	{
		hack[hack_index] &= 0x0000ffff;
		hack[hack_index] |= (encode << 16);
		hack_half = 1;
	}
	else
	{
		hack[hack_index] &= 0xffff0000;
		hack[hack_index] |= (encode & 0xffff);
		hack_index++;
		hack_half = 0;
	}

	return crc;
}


void EmuHardDrive::PutSectorData(const unsigned char* src, const SectorInfo& info)
{
	int sector_index = SeekToSector(info.si_side, info.si_track, info.si_sector);
	assert(sector_index >= 0);
	EmuTrackSectorInfo esec = m_currentTrack.m_info[sector_index];

	assert(info.si_length == esec.m_info.si_length);

	// Copy to track buffer.
	memcpy(&m_currentTrack.m_data[ esec.m_data_offset ], src, esec.m_info.si_length);

	// Create MFM data

	// UGH: can just have one integer bit index
	// m_mfm_index, m_mfm_bit point at $A1 F8 <data> CRC1 CRC2

	int hack_index = 0;
	int hack_half = 0;		// "high" or "low" half
	unsigned int hack[1024+16];

	unsigned int crc = 0xffff;
	unsigned int prev = 0;	// Doesn't matter for "a1" leading byte
	unsigned int data;

	crc = write_byte(prev, data = 0xa1, 0x04, hack, hack_index, hack_half, crc);
	crc = write_byte(prev, data = 0xf8, 0x00, hack, hack_index, hack_half, crc);
	for (int i = 0; i < esec.m_info.si_length; i++)
	{
		data = src[i];
		crc = write_byte(prev, data, 0x00, hack, hack_index, hack_half, crc);
		prev = data;
	}
	unsigned int crc1 = ((crc >> 8) & 0xff);
	unsigned int crc2 = (crc & 0xff);
	crc = write_byte(prev, data = crc1, 0x00, hack, hack_index, hack_half, crc);
	crc = write_byte(prev, data = crc2, 0x00, hack, hack_index, hack_half, crc);
	assert(crc == 0);

	// Hardware would write a certain # of zeros.  We write zeros to make
	// sure the 'clock' bit is right on the first zero.
	(void)write_byte(prev, data = 0x00, 0x00, hack, hack_index, hack_half, crc);
	(void)write_byte(prev, data = 0x00, 0x00, hack, hack_index, hack_half, crc);

	// Write back to MFM track
	int index = esec.m_mfm_index;
	int bit = esec.m_mfm_bit;
	for (int i = 0; i < hack_index; i++)
	{
		write_mfm_byte(m_currentTrack.m_mfm_data, index, bit, hack[i] >> 16);
		write_mfm_byte(m_currentTrack.m_mfm_data, index, bit, hack[i] & 0xffff);
	}

	// UGH terrible condition on last piece
	assert(hack_half == 0);//this will be OK, so testing will be hard.

	// Actually, you have to write-back since scan_track reads.  Fair enough.
	EmuTrack temp;

	//Write MFM track
	{
		FILE* fp = m_file;

		// Grab the little header first
		int file_offset = m_header_length + (m_tracklen + m_track_header_len)*(m_heads*info.si_track + info.si_side);

		// Sort of but a sanity check.
		// Maybe a good idea regardless.
		fseek(fp, file_offset, SEEK_SET);
		unsigned int track_header[3];
		int len = fread(track_header, 1, sizeof(track_header), fp);
		assert(len == sizeof(track_header));
		assert(track_header[0] == 0x12345678);
		assert(int(track_header[1]) == info.si_track);
		assert(int(track_header[2]) == info.si_side);

		// Write MFM track data
		fseek(fp, file_offset + m_track_header_len, SEEK_SET);
		assert( ftell(fp) == file_offset + m_track_header_len);

		int r = fwrite(m_currentTrack.m_mfm_data, 1, m_tracklen, fp);

		assert(r == m_tracklen);
	}

#ifdef EMUDRIVE_VERIFY
	// Test code--read back the track and verify the change.

	scan_track(m_file, info.si_track, info.si_side, temp, 0);
	temp.m_info = new EmuTrackSectorInfo[temp.m_num_sector];
	temp.m_data = new unsigned char[temp.m_data_len];
	scan_track(m_file, info.si_track, info.si_side, temp, 1);

	assert(temp.m_num_sector == m_currentTrack.m_num_sector);
	assert(temp.m_data_len == m_currentTrack.m_data_len);
	assert(temp.m_mfm_len == m_currentTrack.m_mfm_len);

	assert(memcmp(temp.m_info, m_currentTrack.m_info, temp.m_num_sector * sizeof(temp.m_info[0])) == 0);
	assert(memcmp(temp.m_data, m_currentTrack.m_data, temp.m_data_len) == 0);
	assert(memcmp(temp.m_mfm_data, m_currentTrack.m_mfm_data, m_tracklen) == 0);
#endif
}

#ifdef WANT_DUMPTRACK
static void DumpTrack(int CYL, int HEAD, unsigned int* m_mfm_data, int m_mfm_len, FILE *fp);
#endif

void EmuHardDrive::NotifyFormat(int cylinder, int head, int sector_count, int sector_size, unsigned char* buffer, int gap3)
{
	// I need blank images to 'force' format to work.
	//TODO:
	// For non-blank, I have to match the existing geometry.

	// Abbreviated track:
	// GAP3*4E (GAP1 = ?)
	// Repeated sectors:
	//    14*00 IDAM 3*00 12*00 DAM 3*00 GAP3*4E
	// 4E until index pulse

	assert(head < 8);

	bool geometry_changed = false;
	if (head >= m_heads)
	{
		m_heads = head + 1;
		geometry_changed = true;
	}
	if (cylinder >= m_cylinders)
	{
		m_cylinders = cylinder + 1;
		geometry_changed = true;
	}

	if (geometry_changed)
	{
		char comment[256];
		sprintf(comment, "--heads %d --cylinders %-4d --rate 10000000", m_heads, m_cylinders);
		// Bit of hack -- update the command, but only if there is sufficient space
		if (int(strlen(comment)) + 1 <= m_comment_length)
		{
			memset(m_comment, 0, m_comment_length);
			strcpy(m_comment, comment);
		}

		WriteHeader();
	}

	gap3 += 3;

	// Load track as side effect
	(void)SeekToSector(head, cylinder, 0);

	int hack_index = 0;
	int hack_half = 0;		// "high" or "low" half
	unsigned int *hack = m_currentTrack.m_mfm_data;
	int max_index = m_currentTrack.m_mfm_len;

	unsigned int prev = 0;	// Doesn't matter for "a1" leading byte
	unsigned int data;

	// Lead in 4E
	for (int i = 0; i < gap3; i++) {
		(void)write_byte(prev, data = 0x4e, 0x00, hack, hack_index, hack_half, 0xffff);
	}

	// Sectors
	for (int sector = 0; sector < sector_count; sector++)
	{
		unsigned int crc;
		unsigned int crc1, crc2;
		unsigned int prev = 0;
		unsigned int data;

		static int sector_len[4] = { 256, 512, 1024, 128 };
		int size_mark = 0;
		for (int i = 0; i < 4; i++) {
			if (sector_len[i] == sector_size) {
				break;
			}
			size_mark += 0x20;
		}

		// 14 zeros
		for (int i = 0; i < 14; i++) {
			(void)write_byte(prev, data = 0x00, 0x00, hack, hack_index, hack_half, 0xffff);
		}

		// IDAM
		unsigned int idam[4];
		
		int cyl_high = (cylinder >> 8);
		idam[0] = 0xfc | ((cyl_high & 3) ^ 2);	// mark and cylinder MSB
		if (cylinder >= 1024) {
			idam[0] = 0xf0 | ((cyl_high & 7) ^ 2); //WD2010 support for 2048 tracks
		}

		idam[1] = (cylinder & 0xff);				// cylinder LSB
		idam[2] = (buffer[sector*2] & 0x80) | size_mark | head;	// "SDH"
		idam[3] = buffer[sector*2 + 1];	// sector ID
		crc = 0xffff;
		crc = write_byte(prev, data = 0xa1, 0x04, hack, hack_index, hack_half, crc);		// A1 sync
		for (int i = 0; i < 4; i++)
		{
			crc = write_byte(prev, data = idam[i], 0x00, hack, hack_index, hack_half, crc);
		}
		crc1 = ((crc >> 8) & 0xff);
		crc2 = (crc & 0xff);
		crc = write_byte(prev, data = crc1, 0x00, hack, hack_index, hack_half, crc);
		crc = write_byte(prev, data = crc2, 0x00, hack, hack_index, hack_half, crc);
		assert(crc == 0);

		// DAM to IDAM gap zeros
		for (int i = 0; i < 15; i++) {
			(void)write_byte(prev, data = 0x00, 0x00, hack, hack_index, hack_half, 0xffff);
		}

		// DAM
		crc = 0xffff;

		crc = write_byte(prev, data = 0xa1, 0x04, hack, hack_index, hack_half, crc);
		crc = write_byte(prev, data = 0xf8, 0x00, hack, hack_index, hack_half, crc);
		// Fill data with 0xff
		for (int i = 0; i < sector_size; i++)
		{
			crc = write_byte(prev, data = 0xff, 0x00, hack, hack_index, hack_half, crc);
			prev = data;
		}
		crc1 = ((crc >> 8) & 0xff);
		crc2 = (crc & 0xff);
		crc = write_byte(prev, data = crc1, 0x00, hack, hack_index, hack_half, crc);
		crc = write_byte(prev, data = crc2, 0x00, hack, hack_index, hack_half, crc);
		assert(crc == 0);

		// 3 zeros
		for (int i = 0; i < 3; i++) {
			(void)write_byte(prev, data = 0x00, 0x00, hack, hack_index, hack_half, 0xffff);
		}

		// 4E trailer
		for (int i = 0; i < gap3; i++) {
			(void)write_byte(prev, data = 0x4e, 0x00, hack, hack_index, hack_half, 0xffff);
		}
	}
	assert(hack_index < max_index);	// should have checked earlier, rotating buffer, I suppose

	// Fill remainder with 4E
	while (hack_index < max_index)
	{
		(void)write_byte(prev, data = 0x4e, 0x00, hack, hack_index, hack_half, 0xffff);
	}

	//TODO assert(hack_half == 1);	//TODO: edge case to fill it up

	// Write track

	int file_offset = m_header_length + (m_tracklen + m_track_header_len)*(m_heads*cylinder + head);

	(void)fseek(m_file, file_offset, SEEK_SET);
	assert(ftell(m_file) == file_offset);

	unsigned int track_header[3];
	track_header[0] = 0x12345678;
	track_header[1] = cylinder;
	track_header[2] = head;

	int len = fwrite(track_header, 1, sizeof(track_header), m_file);
	assert(len == sizeof(track_header));
	len = fwrite(m_currentTrack.m_mfm_data, 1, m_tracklen, m_file);
	assert(len == m_tracklen);

#ifdef WANT_DUMPTRACK
	// And also test with a dump
	{
		FILE *fp = fopen("dump.txt", "w");
		DumpTrack(cylinder, head, hack, hack_index, fp);
		fclose(fp);
	}
#endif

	// if we're at EOF, write an eof marker
	int pos = ftell(m_file);
	if (fseek(m_file, 0, SEEK_END) == 0)
	{
		int file_end = ftell(m_file);
		if (pos == file_end)
		{
			WriteEofMarker();
		}
	}

	// Read it back as it is now the current track.
	m_currentTrack.m_num_sector = sector_count;
	m_currentTrack.m_data_len = sector_count * sector_size;
	m_currentTrack.m_info = new EmuTrackSectorInfo[m_currentTrack.m_num_sector];
	m_currentTrack.m_data = new unsigned char[m_currentTrack.m_data_len];
	scan_track(m_file, cylinder, head, m_currentTrack, 1);

	m_current_track = cylinder;
	m_current_head = head;

	// Change the notion of sectors per track
	detected_sectors_per_track = sector_count;
}

int EmuHardDrive::GetMinSectorID(int head, int cylinder)
{
	// Load track as side effect
	(void)SeekToSector(head, cylinder, 0);

	int min = 0;
	int num_sector = m_currentTrack.m_num_sector;
	if (num_sector > 0) {
		min = m_currentTrack.m_info[0].m_info.si_sector;
	}
	for (int i = 1; i < num_sector; i++) {
		if (m_currentTrack.m_info[i].m_info.si_sector < min) {
			min = m_currentTrack.m_info[i].m_info.si_sector;
		}
	}

	return min;
}

void EmuHardDrive::SetComment(const char* comment)
{
	// Warning: header length will change
	delete [] m_comment;
	m_comment_length = strlen(comment) + 1;
	m_comment = new char[m_comment_length];
	strcpy(m_comment, comment);
}

void EmuHardDrive::SetNote(const char* note)
{
	// Warning: header length will change
	delete [] m_note;
	m_note_length = strlen(note) + 1;
	m_note = new char[m_note_length];
	strcpy(m_note, note);
}

void EmuHardDrive::InitBlank(const DiskImageMetadata *stamp)
{
// http://www.pdp8online.com/mfm/code/mfm/emu_tran_file.c

// First 16: magic + version (00 02 02 02) + header len
// Next  16: m_tracklen, m_track_header_len (0xc), m_cylinders, m_heads
// Next   8: data_rate, command_line_length -- then command (length includes \0)
// command may as well duplicate this: --heads 6 --cylinders 230  --rate 10000000
// which should be length 2b  (called the "note" field)

// After that:
// 4 byte start of track time from index in nanoseconds (value 01 00 00 00 = 1)
// First track seems to align on 4-byte boundary.
// Then tracks: 78 56 34 12 CYL HEAD [m_tracklen_bytes]
// ...
// ends with a 78 56 34 12 ff ff ff ff ff ff ff ff marker

	memset(m_header, 0, sizeof(m_header));
	memcpy(m_header, magic, sizeof(magic));

	// Version
	m_minor = 2;
	m_major = 2;
	m_file_type = 2;
	m_header[8] = 0;
	m_header[9] = m_minor;
	m_header[10] = m_major;
	m_header[11] = m_file_type;

	// Calculation: 2 * data rate / revolutions/second = stored bits, divide by 8 for bytes.
	// Data rate doubled to get the MFM bit rate.
	int tracklen_bits = int(ceil(2 * m_config.data_rate / (m_config.rpm / 60.)));
	// Round to even 32-bit word length
	tracklen_bits = (tracklen_bits + 31) & ~31;
	// Convert to bytes.
	m_tracklen = tracklen_bits / 8;

	// Note, the above tracklen calculation is a problem since the ext2emu utility doesn't
	// use rate/RPM--instead it uses controller info with # 32-bit words/track.
	// The ceil/rouding is close enough, but the actual value is simply 5209*4=20836
	// It only becomes a problem when trying mixing ext2emu generated files with
	// trs80gp ones on the DGEMU hardware--the driver complains that the track lengths differ.
	// See https://www.pdp8online.com/mfm/code/mfm/inc/mfm_decoder.h for sample controller data.

	m_track_header_len = 0xc;
	m_cylinders = 0;
	m_heads = 0;
	m_bit_rate = m_config.data_rate * 2;

	char comment[256];
	sprintf(comment, "--heads %d --cylinders %-4d --rate 10000000", m_heads, m_cylinders);
	SetComment(comment);
	SetNote("");
	m_track_time = 0;

	m_header_length = 
		sizeof(m_header) +
		5*sizeof(int) +
		2*sizeof(int) + m_note_length + m_comment_length +
		sizeof(int);

	// Want to round this out
	if (m_header_length % 4 != 0)
	{
		m_header_length += 4 - m_header_length % 4;
		assert(m_header_length % 4 == 0);
	}

	WriteHeader();
	// Write EOF track
	WriteEofMarker();

	fflush(m_file);
}

void EmuHardDrive::WriteHeader()
{
	(void)fseek(m_file, 0, SEEK_SET);

	// strict aliasing breaking means we can't do it the easy way
	//*(int*)&m_header[12] = m_header_length;
	m_header[12] = m_header_length;
	m_header[12 + 1] = m_header_length >> 8;
	m_header[12 + 2] = m_header_length >> 16;
	m_header[12 + 3] = m_header_length >> 24;

	(void)fwrite(m_header, 1, sizeof(m_header), m_file);

	unsigned int param[6];

	param[0] = m_tracklen;
	param[1] = m_track_header_len;
	param[2] = m_cylinders;
	param[3] = m_heads;
	param[4] = m_bit_rate;
	param[5] = m_comment_length;
	(void)fwrite(param, 1, sizeof(param), m_file);

	(void)fwrite(m_comment, 1, m_comment_length, m_file);
	param[0] = m_note_length;
	(void)fwrite(param, 1, sizeof(int), m_file);
	(void)fwrite(m_note, 1, m_note_length, m_file);
	param[0] = m_track_time;
	(void)fwrite(param, 1, sizeof(int), m_file);
	for (int offset = ftell(m_file); offset < m_header_length; offset++)
	{
		(void)fputc(0, m_file);
	}

	int final = ftell(m_file);
	// Might fail if header has extra data (e.g. minor version bump)
	assert(final == m_header_length);
}

void EmuHardDrive::WriteEofMarker()
{
	unsigned int track_header[3];
	track_header[0] = 0x12345678;
	track_header[1] = 0xffffffff;
	track_header[2] = 0xffffffff;

	int len = fwrite(track_header, 1, sizeof(track_header), m_file);
	assert(len == sizeof(track_header));
}

void EmuHardDrive::Load(FILE* fp, bool read_only, const DiskImageMetadata *stamp)
{
	m_readonly = read_only;

	if (fseek(fp, 0, SEEK_END) != 0)
	{
		Fail("Seek error");
		return;
	}

	m_file_size = ftell(fp);
	fseek(fp, 0, SEEK_SET);

	if (m_file_size == 0)
	{
		// New file
		m_file = fp;

		InitBlank(stamp);
		return;
	}

	m_file = fp;

	int len = fread(m_header, 1, sizeof(m_header), m_file);
	assert(len == sizeof(m_header));

	m_minor = m_header[9];
	m_major = m_header[10];
	m_file_type = m_header[11];

	assert(memcmp(m_header, magic, sizeof(magic)) == 0); 
	assert(m_file_type == 2);

	assert(m_major > 2 || (m_major == 2 && m_minor >= 2));

	// sizeof of full header
	// strict aliasing says we must to this the hard way
	//m_header_length = *(int*)&m_header[12];
	m_header_length = (m_header[12] & 0xff) |
		((m_header[12 + 1] & 0xff) << 8) |
		((m_header[12 + 2] & 0xff) << 16) |
		((m_header[12 + 3] & 0xff) << 24);

	unsigned int param[6];
	len = fread(param, 1, sizeof(param), m_file);
	assert(len == sizeof(param));

	m_tracklen = param[0];
	m_track_header_len = param[1];
	m_cylinders = param[2];
	m_heads = param[3];
	m_bit_rate = param[4];
	m_comment_length = param[5];

	delete [] m_comment;
	m_comment = new char[m_comment_length];
	len = fread(m_comment, 1, m_comment_length, m_file);
	assert(len == m_comment_length);

	len = fread(&m_note_length, 1, sizeof(int), m_file);
	assert(len == sizeof(int));
	delete [] m_note;
	m_note = new char[m_note_length];
	len = fread(m_note, 1, m_note_length, m_file);
	assert(len == int(m_note_length));

	len = fread(&m_track_time, 1, sizeof(int), m_file);
	assert(len == sizeof(int));
}

void EmuHardDrive::SaveToFile()
{
	int err;

	err = fclose(m_file);
	m_file = NULL;
	assert(err == 0);
}

void my_dump_segment(FILE *f, const char* name, const SectorInfo* info, unsigned char* data, int end, int mark_len = 0, int crc_offset = -1)
{
	extern void dump_segment(FILE *f, const char* name, int sector, unsigned char* data, int offset, int end, int delta, int mark_len, int crc_offset);

	// TODO: want to dump the info.si_side (cylinder), too
	dump_segment(f, name, info ? info->si_sector : 0, data, 0, end, 1, mark_len, crc_offset);
}

#ifdef WANT_DUMPTRACK

// ASCII dump of a track
// - gaps are quite noisy (probably realistic but could be a bug)
static void DumpTrack(int CYL, int HEAD, unsigned int* m_mfm_data, int m_mfm_len, FILE *fp)
{
	unsigned char *temp_buf = new unsigned char[m_mfm_len];

	fprintf(fp, "cylinder %02x head %x len=0x%0x\n", CYL, HEAD, m_mfm_len * 2);

	unsigned int clockdata = 0;

	int index = 0;
	int bit = 31;

	int buf_len = 0;
	int gap_bit = 0;

	int gap_prev_index = index, gap_prev_bit = bit;
	SectorInfo current_idam;
	current_idam.si_sector = -1;	// to mean: not yet current
	current_idam.si_length = 512;	//TODO: no proper way to default this
	while (index < m_mfm_len)
	{
		read_bit(m_mfm_data, index, bit, clockdata); clockdata &= 0xffff;
		if (clockdata != 0x4489)	// data 0xa1 clock 0x0a
		{
			// accumulate gap
			gap_bit++;
			if (gap_bit == 16)
			{
				temp_buf[buf_len++] = sync(m_mfm_data, gap_prev_index, gap_prev_bit);
				gap_bit = 0;
				//TODO: we'll miss partial bytes here (my_dump_segment needs extra byte + bit count
			}
		}
		else
		{
			// Found mark, output the gap information
			my_dump_segment(fp, NULL, NULL/*no sector*/, temp_buf, buf_len);
			buf_len = 0;

			temp_buf[buf_len++] = 0xa1;

			unsigned int mark = sync(m_mfm_data, index, bit);
			temp_buf[buf_len++] = mark;

			int saved_bit = bit, saved_index = index;
			if (mark >= 0xfc)//TODO: missing 2048 sector support (WIP)
			{
				idam(mark, m_mfm_data, index, bit, current_idam);

				// Grab IDAM bytes
				for (int i = 0; i < 5; i++) {
					temp_buf[buf_len++] = sync(m_mfm_data, saved_index, saved_bit);
				}
				// Dump IDAM
				my_dump_segment(fp, "idam", &current_idam, temp_buf, buf_len, 1, buf_len - 2);
				buf_len = gap_bit = 0;
				gap_prev_index = index;
				gap_prev_bit = bit;
			}
			else if (mark == 0xf8)
			{
				// TODO: current_idam.si_sector < 0 logic here
				// Grab DAM byte and CRC
				for (int i = 0; i < current_idam.si_length + 2; i++) {
					temp_buf[buf_len++] = sync(m_mfm_data, index, bit);
				}
				// Dump DAM
				my_dump_segment(fp, "dam ", &current_idam, temp_buf, buf_len, 1, buf_len - 2);
				buf_len = gap_bit = 0;
				gap_prev_index = index;
				gap_prev_bit = bit;
			}
			else
			{
				// Unknown mark
				my_dump_segment(fp, "???", &current_idam, temp_buf, buf_len, 1);
				buf_len = gap_bit = 0;
				gap_prev_index = index;
				gap_prev_bit = bit;
			}
			clockdata = 0;
		}
	}
	if (buf_len > 0)
	{
		my_dump_segment(fp, NULL, 0/*no sector*/, temp_buf, buf_len);
	}

	delete [] temp_buf;
}

#endif

static void alloc_blank_track(EmuTrack& track, int tracklen)
{
	// Make the track blank
	track.m_mfm_len = tracklen / 4;
	track.m_mfm_data = new unsigned int[track.m_mfm_len];
	memset(track.m_mfm_data, 0, tracklen);
}

//#define CORVUS_DECODE

#ifdef CORVUS_DECODE
//#define CRC_POLY 0x1021
#define XCRC_POLY 0x8005
static unsigned int Xupdate_crc_notable(unsigned int crc, unsigned int data)
{
	for (int b = 0; b < 8; b++) {
		if ( (crc ^ (data << (8 + b))) & 0x8000) {
			crc = ((crc << 1) ^ 0x8005) & 0xffff;
		} else {
			crc = ((crc << 1) & 0xffff);
		}
	}

	return crc;
}

unsigned long Xcrc_table[256];
void Xmake_crc_table()
{
	for (unsigned int i = 0; i < 256; i++) {
		Xcrc_table[i] = Xupdate_crc_notable(0, i);
	}
}

unsigned int Xupdate_crc(unsigned int crc, unsigned int data)
{
	return ((crc << 8) ^ Xcrc_table[(crc >> 8) ^ data]) & 0xffff;
}
#endif

int EmuHardDrive::scan_track(FILE *fp, int CYL, int HEAD, EmuTrack& track, int pass)
{
#ifdef CORVUS_DECODE
	int is_corvus = strstr(m_comment, "Corvus") != NULL;

	if (is_corvus) {
		//HEAD = 1;
		Xmake_crc_table();

		//HEAD = 1;
		//CYL = 18;
	}
#endif
	if (track.m_info != NULL)
	{
		assert(pass == 1);
	}
	else
	{
		assert(pass == 0);
	}

	if (pass == 0)
	{
		// Grab the little header first
		int file_offset = m_header_length + (m_tracklen + m_track_header_len)*(m_heads*CYL + HEAD);

		fseek(fp, file_offset, SEEK_SET);
		unsigned int track_header[3];
		int len = fread(track_header, 1, sizeof(track_header), fp);
		if (len != sizeof(track_header))
		{
			// Make the track blank
			alloc_blank_track(track, m_tracklen);

			return -1;
		}

		assert(len == sizeof(track_header));
		assert(track_header[0] == 0x12345678);
		if (track_header[1] == 0xffffffff || track_header[2] == 0xffffffff)
		{
			// EOF marker, use a blank track
			alloc_blank_track(track, m_tracklen);

			return -1;
		}

		assert(int(track_header[1]) == CYL);
		assert(int(track_header[2]) == HEAD);

		// Read MFM track data
		fseek(fp, file_offset + m_track_header_len, SEEK_SET);

		assert(m_tracklen % 4 == 0);
		track.m_mfm_len = m_tracklen / 4;
		track.m_mfm_data = new unsigned int[track.m_mfm_len];
		int r = fread(track.m_mfm_data, 1, m_tracklen, fp);

		assert(r == m_tracklen);
	}

	unsigned int clockdata = 0;

	int index = 0;
	int bit = 31;
#ifdef CORVUS_DECODE
	if (is_corvus) {

		SectorInfo current_idam;
		current_idam.si_sector = -1;	// to mean: not yet current
		int sector_index = 0;
		int current_offset = 0;

#ifdef NOISY
		extern void ods(char *str);
		char buf[4096];
		sprintf(buf, "CORVUS track dump H%d CYL%3d\n", HEAD, CYL);
		ods(buf);
#endif
		int run = 0;
		while (index < track.m_mfm_len)
		{
			read_bit(track.m_mfm_data, index, bit, clockdata);

			if (clockdata == 0x55555555 || clockdata == 0xaaaaaaaa) {
				run ++;
				if (run > 20) {
//					sprintf(buf, "%08x at %d/%d run=%d\n", clockdata, index, bit, run);
//					ods(buf);
				}
			}
			else {
				if (run > 20) {
					if ((clockdata & 0xf) == 0x9) {
#ifdef NOISY
						sprintf(buf, "%08x hit a sync?\n", clockdata, index, bit);
#endif

						unsigned int crc = 0xffff;

						// Ignore next two bits, I think it matches this code:
						//if (drive_params->controller == CONTROLLER_CORVUS_H) {
						//	next_header_time += 164900;
						//	raw_bit_cntr = -2;
						//	// May be better to time from current track time in
						//} else if (drive_params->controller == CONTROLLER_VECTOR4 ||
						// The 164900 time is in units of 200MHz
						// Multiplying it out by 20 (sectors) and coverting to RPM gets 3638 RPM
						// which makes sense.

						// Toss "data" bit?
						read_bit(track.m_mfm_data, index, bit, clockdata);
						read_bit(track.m_mfm_data, index, bit, clockdata);

						// pull some bytes?
						unsigned char data[640];
						int i;
						for (i = 0; i < 20; i++) {
							data[i] = sync(track.m_mfm_data, index, bit);
							crc = Xupdate_crc(crc, data[i]);
						}
						if ((data[0] >> 5) == HEAD && data[1] == (CYL & 0xff) && data[2] == ((CYL >> 8) & 0xff)) {
#ifdef NOISY
							sprintf(buf, "H%d CYL%03d SEC%02d\n", data[0] >> 5, data[1] + (data[2] << 8), data[0] & 0x1f);
							ods(buf);
#endif
							// Pull in lots
							for ( ; i < 512 + 3 + 2; i++) {
								data[i] = sync(track.m_mfm_data, index, bit);
								crc = Xupdate_crc(crc, data[i]);
							}
#ifdef NOISY
							char *p = buf;
							for (int j = 0; j < i; j++) {
								sprintf(p, " %02x", data[j]);
								p += strlen(p);
								if (j % 16 == 15) {
									strcat(p, "\n");
									ods(buf);
									p = buf;
								}
							}
							sprintf(p, "   [%04x]", crc);
							p += strlen(p);
#endif
							if (crc != 0) {
#ifdef NOISY
								sprintf(p, "  -- BAD CRC");
#endif
							}
							else {
								current_idam.si_length = 512;
								current_idam.si_sector = (data[0] & 0x1f);
								current_idam.si_flags = SECTOR_dden;
								current_idam.si_side = (data[0] >> 5);
								current_idam.si_track = data[1] + (data[2] << 8);
								if (pass == 0) {
									track.m_num_sector++;

									track.m_data_len += current_idam.si_length;
								}
								else {
									track.m_info[sector_index].m_info = current_idam;
									track.m_info[sector_index].m_mark = 0xf8;		// FAKE mark
									sector_index++;

									memcpy(&track.m_data[current_offset], data + 3, current_idam.si_length);

									EmuTrackSectorInfo& info = track.m_info[sector_index - 1];
									info.m_data_offset = current_offset;
									info.m_mfm_index = 0;// TODO a1_index;
									info.m_mfm_bit = 0; // TODO a1_bit;

									current_offset += current_idam.si_length;
								}

							}
#ifdef NOISY
							strcat(p, "\n");
							ods(buf);
#endif
						}
						else {
#ifdef NOISY
							ods(buf);
#endif
						}
						run = 0;
					}
					else {
						//sprintf(buf, "%08x end run %d/%d\n", clockdata, index, bit);
					}
					//ods(buf);
				}
				else {
					run = 0;	// I dunno, kills "sort runs" or something.
				}
			}
		}

		return 1;
	}
#endif

	// Scan through bits
	//for (int i = 0; i < 1000; i++) // 1000 clock/data pairs
	SectorInfo current_idam;
	current_idam.si_sector = -1;	// to mean: not yet current
	int sector_index = 0;
	int current_offset = 0;
#ifdef NOISY
	int prev_bit_index = 0;
#endif
	while (index < track.m_mfm_len)
	{
		read_bit(track.m_mfm_data, index, bit, clockdata); clockdata &= 0xffff;
		if (clockdata == 0x4489)	// data 0xa1 clock 0x0a
		{
			// Have to back up 16-bits
			int a1_index = index;
			int a1_bit = bit + 16;
			while (a1_bit > 31)
			{
				a1_bit -= 32;
				a1_index--;
			}

#ifdef NOISY
			char res[256];
			int bit_index = a1_index*32 + (31-a1_bit);
			sprintf(res, "MARK at %10.2f (+ %10.2f)\n", bit_index / 16.0f, (bit_index - prev_bit_index) / 16.0f);
			prev_bit_index = bit_index;
			ods(res);
#endif
			unsigned int mark = sync(track.m_mfm_data, index, bit);

//			sprintf(res, "index,bit=%d,%d clock=%02x data=%02x next=%02x\n", index, bit, clock, data, next);
//			ods(res);

			// TODO: some tricky bits here
			// Only read the first DAM after an IDAM
			// A DAM might appear before the first IDAM
			// There could be two IDAM in a row, etc.
			if (mark >= 0xfc || (mark >= 0xf4 && mark <= 0xf7))	//WD2010 support for 2048 tracks
			{
				idam(mark, track.m_mfm_data, index, bit, current_idam);
				if (pass == 0)
				{
					track.m_num_sector++;
					if (current_idam.si_side != HEAD || current_idam.si_track != CYL)
					{
#ifdef NOISY
						ods("off track\n");
#endif
					}
				}
				else
				{
					track.m_info[sector_index].m_info = current_idam;
					track.m_info[sector_index].m_mark = mark;
					sector_index++;
				}
			}
			else if (mark == 0xf8)
			{
				// TODO: if current_idam.si_sector < 0, we have to ignore this.
				// also, it we accept it, we should reset current_idam.si_sector to -1.
				int crc_check = dam(mark, track.m_mfm_data, index, bit, current_idam, track.m_data ? &track.m_data[current_offset] : NULL);
				if (pass == 0)
				{
					track.m_data_len += current_idam.si_length;
				}
				else
				{
					assert(sector_index > 0);
					EmuTrackSectorInfo& info = track.m_info[sector_index - 1];
					info.m_data_offset = current_offset;
					info.m_mfm_index = a1_index;
					info.m_mfm_bit = a1_bit;

					current_offset += current_idam.si_length;
				}
				if (crc_check != 0)
				{
#ifdef NOISY
					ods("bad dam crc\n");
#endif
				}
			}
			else
			{
#ifdef NOISY
				sprintf(res, "unknown mark %02x\n", mark);
				ods(res);
#endif
			}
			clockdata = 0;
		}
	}

#ifdef NOISY
	ods("done\n");
#endif

	return 1;
}

int EmuHardDrive::SeekToSector(int head, int cylinder, int sector_id)
{
	//TODO: technically, I think you may have to scan the track list.
	if (m_current_track != cylinder || m_current_head != head)
	{
		m_currentTrack.reset();
		if (scan_track(m_file, cylinder, head, m_currentTrack, 0) < 0)
		{
			return -1;
		}
		m_currentTrack.m_info = new EmuTrackSectorInfo[m_currentTrack.m_num_sector];
		m_currentTrack.m_data = new unsigned char[m_currentTrack.m_data_len];
		scan_track(m_file, cylinder, head, m_currentTrack, 1);

#ifdef WANT_DUMPTRACK
		//{
		//	FILE *fp = fopen("dump.txt", "w");
		//	DumpTrack(cylinder, head, m_currentTrack.m_mfm_data, m_currentTrack.m_mfm_len, fp);
		//	fclose(fp);
		//}
#endif

		m_current_track = cylinder;
		m_current_head = head;
	}

	if (detected_sectors_per_track <= 0)
	{
		detected_sectors_per_track = m_currentTrack.m_num_sector;
	}

	// Find sector
	for (int i = 0; i < m_currentTrack.m_num_sector; i++)
	{
		if (m_currentTrack.m_info[i].m_info.si_sector == sector_id)
		{
			return i;
		}
	}

	return -1;
}

HardDriveImage* hd_create_emu(Notifier* report)
{
	return new EmuHardDrive(report);
}

