#include <stdio.h>
#include <string.h>
#include <math.h>

#include <assert.h>

#include "fd_image.h"
#include "hd_image.h"

#ifdef _MSC_VER
#define stricmp _stricmp
#define strnicmp _strnicmp
#else
#define stricmp strcasecmp
#define strnicmp strncasecmp
#endif

class HDV1HardDriveImage : public HardDriveImage
{
public:
	HDV1HardDriveImage(Notifier* err) 
		: HardDriveImage(err)
	{
	}

	virtual int IsModified();

	virtual const DriveConfiguration* GetConfig();
	virtual HardDiskImageFormat ImageFormat() { return DSK_HDV1; }

	virtual void Load(FILE* fp, bool read_only, const DiskImageMetadata *stamp = NULL);
	virtual void SaveToFile();
	virtual int Flush();

    virtual void GetDiskInfo(DiskInfo& info);
    virtual void GetTrackInfo(TrackInfo& info, int head, int cylinder);

    virtual void GetSectorInfo(SectorInfo& info, int head, int cylinder, int sector_id, int expected_length = 0);
    virtual void GetSectorData(unsigned char* res, const SectorInfo& info);
	virtual void PutSectorData(const unsigned char* res, const SectorInfo& info);

	virtual void NotifyFormat(int cylinder, int head, int sector_count, int sector_size, unsigned char* buffer, int gap3);

	virtual int GetMinSectorID(int head, int cylinder);

protected:
	void TODO(const char* message)
	{
#if !defined(SUPPRESS_TODO)
		printf("TODO: %s\n", message);
#endif
	}
	void SeekToSector(int head, int cylinder, int sector, int sector_size);

	void InitBlank(const DiskImageMetadata *stamp);

	void WriteHeader();

	FILE* m_file;
	int m_file_size;
	int m_modified;
	bool m_readonly;

	// Header data to be written back.
	unsigned char m_header[256];

	// Some parts of header decoded;
	int header_version;
	int write_protect;
	DiskImageMetadata hdv1_metadata;

	// TODO: Geometry information should reflect into HardDrive
	int heads;			// will be calculated if zero
	int cylinders;
	int sectors_per_cylinder;
	int directory_cylinder;

	// Calculated
	int sectors_per_head;			// F48 format stores this instead of sectors_per_cylinder
	int expected_sector_length;		// ultimately may not be needed.
} ;


const DriveConfiguration* HDV1HardDriveImage::GetConfig()
{
	// Update with current setting
	m_config.data_rate = 5000000;
	m_config.rpm = 3600;
	m_config.sectors_per_track = sectors_per_head;
	m_config.sides = heads;
	m_config.tracks = cylinders;

	return &m_config;
}

void HDV1HardDriveImage::InitBlank(const DiskImageMetadata *stamp)
{
	// Blank image
	memset(m_header, 0, sizeof(m_header));
	m_header[0] = 0x56; m_header[1] = 0xcb; m_header[2] = header_version = 0x10;

	m_header[4] = 1; m_header[5] = 4;

	write_protect = 0;

	if (stamp)
	{
		hdv1_metadata = *stamp;
	}
	else
	{
		hdv1_metadata.yyyy = 1977;
		hdv1_metadata.mm = 1;
		hdv1_metadata.dd = 1;
		hdv1_metadata.hour = hdv1_metadata.min = hdv1_metadata.sec = 0;
		hdv1_metadata.trs80_model_family = 0;
	}

	m_header[14] = hdv1_metadata.yyyy - 1900;
	m_header[12] = hdv1_metadata.mm;
	m_header[13] = hdv1_metadata.dd;

	heads = 2;
	cylinders = 1;
	// Maximum cylinder limit
	cylinders &= 0x7ff;

	m_header[26] = heads;
	m_header[27] = ((cylinders >> 8) & 7);
	m_header[28] = (cylinders & 255);

	sectors_per_head = 32;
	expected_sector_length = 256;

	if (hdv1_metadata.trs80_model_family != 2)
	{
		// Documented HDV1 format
		m_header[29] = sectors_per_cylinder = sectors_per_head * heads;
	}
	else
	{
		// HDV1 format used by F48  (Model II image), m_header[29] is sectors_per_head
		// Reduce sectors per head to a realistic value (can't have 32 512-byte sectors)
		sectors_per_head = 17;
		sectors_per_cylinder = sectors_per_head * heads;
		m_header[29] = sectors_per_cylinder = sectors_per_head;
		expected_sector_length = 512;

		m_header[8] |= 2;
		m_header[8] |= 8;	// 512 bytes/sector
	}

	m_header[31] = directory_cylinder = 1;

	m_file_size = sizeof(m_header) + sectors_per_cylinder * cylinders * expected_sector_length;

	WriteHeader();
}

int HDV1HardDriveImage::IsModified()
{
	return m_modified;
}

void HDV1HardDriveImage::SaveToFile()
{
	fclose(m_file);
	m_file = NULL;
}

int HDV1HardDriveImage::Flush()
{
	return 0;
}

void HDV1HardDriveImage::Load(FILE* fp, bool read_only, const DiskImageMetadata *stamp)
{
	m_modified = 0;
	m_readonly = read_only;

	if (fseek(fp, 0, SEEK_END) != 0)
	{
		Fail("Seek error");
		return;
	}

	m_file_size = ftell(fp);
	fseek(fp, 0, SEEK_SET);

	if (m_file_size == 0)
	{
		// New file
		m_file = fp;

		InitBlank(stamp);
		return;
	}

	int len = fread(m_header, 1, sizeof(m_header), fp);
	if (len != sizeof(m_header))
	{
		Fail("header too small");
		return;
	}

	m_file = fp;

	if (fseek(fp, 0, SEEK_END) != 0)
	{
		Fail("Seek error");
		return;
	}
	m_file_size = ftell(fp);

	//HDV1 http://www.trs-80emulators.com/hdv1-format/
	//HDV1 marked up some changes. Should ignore more bytes.
	// The big change is 26: #heads.  I can probably fake
	// the sector size based on file size.

	//  0-3  | 0x56 | 0xcb | VER  | csum |
	//  4-7  | blks | mp4  |media |flag1 |
	//  8-11 | blks |      reserved      |
	// 12-15 |  mm  |  dd  |  yy  |      |
	// 24-27 |             | h/cyl| cylH |
	// 28-31 | cylL | sec  | gran | dcyl |
	if (m_header[0] != 0x56 && m_header[1] != 0xcb)
	{
		Fail("Bad magic");
		return;
	}

	header_version = m_header[2];
	if (!(header_version == 0x10 || header_version == 0x11))
	{
		Fail("Unsupported version");
		return;
	}

	if (!(m_header[4] == 1 && m_header[5] == 4 && m_header[6] == 0))
	{
		Fail("Bad 'magic' settings");
		return;
	}

	write_protect = 0;
	if (m_header[7] & 0x80)
	{
		write_protect = 1;
	}

	if (m_header[8] & 0x01)
	{
		Warn("Don't understand auto-boot");
	}

	// xtrs says there is a date here.  I'll read for potential future use
	TODO("correct interpretation of month");
	hdv1_metadata.yyyy = m_header[14] + 1900;
	hdv1_metadata.mm = m_header[12];
	hdv1_metadata.dd = m_header[13];
	hdv1_metadata.hour = hdv1_metadata.min = hdv1_metadata.sec = 0;
	hdv1_metadata.trs80_model_family = 0;

	if (m_header[8] & 0x02)
	{
		hdv1_metadata.trs80_model_family = 2;
	}
	else
	{
		// Machine model unknown.
		hdv1_metadata.trs80_model_family = 0;
	}

	heads = m_header[26];
	cylinders = ((m_header[27] & 7) << 8) | m_header[28];
	if (cylinders == 0)
	{
		cylinders = 256;
	}
	sectors_per_cylinder = m_header[29];

	expected_sector_length = 256;
	if (hdv1_metadata.trs80_model_family == 2)
	{
		sectors_per_cylinder = sectors_per_cylinder * heads;
		switch ((m_header[8] >> 2) & 0x3)
		{
			case 0: expected_sector_length = 128; break;
			case 1: expected_sector_length = 256; break;
			case 2: expected_sector_length = 512; break;
			case 3: expected_sector_length = 1024; break;
		}
	}

	if (sectors_per_cylinder == 0) 
		sectors_per_cylinder = 256;	// 0 => 256 according is xtrs

	// gran[ules per track] and dcyl[directory cylinder] deprecated.
	// directory_cylinder should be 1
	directory_cylinder = m_header[31];

	// Now some calculations
	if (heads == 0)
	{
		sectors_per_head = 32;
		heads = sectors_per_cylinder / 32;
	}
	else
	{
		sectors_per_head = sectors_per_cylinder / heads;
	}

	if (heads > 8)
	{
		Warn("Unusually high head count");
	}
}

void HDV1HardDriveImage::GetDiskInfo(DiskInfo& info)
{
	//TODO: the geometry information goes into the drive.
	info.tracks = cylinders;
	info.sides = heads;
}

void HDV1HardDriveImage::GetTrackInfo(TrackInfo& info, int head, int cylinder)
{
	(void)head;
	(void)cylinder;

	info.sectors = sectors_per_head;
}

void HDV1HardDriveImage::GetSectorInfo(SectorInfo& info, int head, int cylinder, int sector_id, int expected_length)
{
	// Help utilities which don't know the sector length (if the HDV1 image declared the sector length)
	if (expected_length == 0)
		expected_length = expected_sector_length;

	info.si_flags = SECTOR_dden;		// MFM encoding
	info.si_length = expected_length;
	info.si_sector = sector_id;
	info.si_track = cylinder;
	info.si_side = head;
	//TODO
	TODO("need to set_rate() for sector info");
}

void HDV1HardDriveImage::SeekToSector(int head, int cylinder, int sector, int sector_size)
{
	int file_offset = cylinder * sectors_per_cylinder + head * sectors_per_head + (sector % sectors_per_head);
	file_offset = sizeof(m_header) + file_offset * sector_size;
	fseek(m_file, file_offset, SEEK_SET);

	TODO("assert on ftell()");
}

void HDV1HardDriveImage::GetSectorData(unsigned char* res, const SectorInfo& info)
{
	// Initially result to zero so empty drives look empty
	memset(res, 0, info.si_length);

	SeekToSector(info.si_side, info.si_track, info.si_sector, info.si_length);
	if (int(fread(res, 1, info.si_length, m_file)) != int(info.si_length))
	{
		Warn("failed reading of full sector...only a warning? Error, or a diagnostic?");
	}
}

void HDV1HardDriveImage::PutSectorData(const unsigned char* res, const SectorInfo& info)
{
	if (write_protect)
	{
		TODO("must note error on write protect");
		return;
	}

	SeekToSector(info.si_side, info.si_track, info.si_sector, info.si_length);
	if (int(fwrite(res, 1, info.si_length, m_file)) != info.si_length)
	{
		Warn("failed reading of full sector...only a warning? Error, or a diagnostic?");
	}

	m_modified = 1;
}

void HDV1HardDriveImage::WriteHeader()
{
	fseek(m_file, 0, SEEK_SET);
	if (fwrite(m_header, 1, sizeof(m_header), m_file) != sizeof(m_header))
	{
		Warn("failed to write header");
	}

	m_modified = 1;
}

void HDV1HardDriveImage::NotifyFormat(int cylinder, int head, int sector_count, int sector_size, unsigned char* buffer, int gap3)
{
	// Format can override the sectors_per_head (and even sector_length)
	expected_sector_length = sector_size;
	sectors_per_head = sector_count;

	// "header is dirty" if changed.
	bool changed = false;
	if (head >= heads)
	{
		heads = head + 1;
		changed = true;
	}
	if (cylinder >= cylinders)
	{
		cylinders = cylinder + 1;
		changed = true;
	}
	sectors_per_cylinder = heads * sectors_per_head;

	if (changed)
	{
		m_header[26] = heads;
		m_header[27] = ((cylinders >> 8) & 7);
		m_header[28] = (cylinders & 255);
		assert(sectors_per_cylinder <= 256);

		if (hdv1_metadata.trs80_model_family != 2)
		{
			// Documented HDV1 format
			m_header[29] = sectors_per_cylinder == 256 ? 0 : sectors_per_cylinder;
		}
		else
		{
			// F48 format (Model II image)
			m_header[29] = sectors_per_head;

			if (expected_sector_length != 512)
			{
				//TODO: should really set the correct sector size
				Warn("Model II image using non-512 byte sectors");
			}
		}

		WriteHeader();
	}

	// File might have to get bigger. SeekToSector() needs to deal with that.
	// OTOH, I'm half-way convinced FORMAT should truncate the file (if pre-existing)
	// _chsize() for Win32, used in DiskDrive.cpp but not ported.
}

int HDV1HardDriveImage::GetMinSectorID(int head, int cylinder)
{
	if (hdv1_metadata.trs80_model_family == 2) {
		return 1;
	}

	return 0;
}

HardDriveImage* hd_create_hdv1(Notifier* report)
{
	return new HDV1HardDriveImage(report);
}

