//
// xfile -- Utility to read write files from Xenix hard drive images.
//

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <utime.h>
#include <errno.h>
#include <sys/stat.h>

#include "fd_image.h"
#include "hd_image.h"

int gCylinders;	// Cylinders per drive
int gHeads;		// Heads per cylinder
int gSectors;	// Sectors per track
int gBytes;		// Bytes per sector

int gMaxBadTrack = 0;
int gNumBadTrack = 0;

// I don't know if the Xenix filesystem always starts on track 2 of cylinder 2.
// Might be on track 1 or 0 if there is no boot track.
int gStartHead = 2;

// Start with a reasonable minimums so we may get the actual values from the
// superblock at block 1.
int gNumInodeBlock = 1;
int gNumDataBlock = 1;

static HardDriveImage *hd;

char* floppy;
int floppy_len;

char* read_floppy(char* path, int &length, int &try_hd)
{
	DiskImageMetadata meta;
	DriveConfiguration drive_configuration[1] = { 360,	500*1024, DISK_MFM, 1, 77, 26 };
	Notifier report;

	const char* error = NULL;

	DiskDrive* dsk = LoadDisk(path, NULL, 0, error, drive_configuration[0]);
	if (error != NULL) { 
		try_hd = 1;
		//fprintf(stderr, "xfile: error loading %s: %s\n", path, error);
		return 0;
	}
	
	try_hd = 0;
	
	const DriveConfiguration *dc = dsk->GetConfig();
	
	// Compute size
	length = 0;
	gCylinders = dc->tracks;
	gHeads = dsk->GetNumSides();
	gSectors = 0;
	gBytes = 512;	// bytes/sector

	if (gHeads != 1) {
		fprintf(stderr, "xfile: warning: floppy disk with %d sides not tested\n", gHeads);
	}	

	for (int track = 0; track < dc->tracks; track++) {
		for (int side = 0; side < dc->sides; side++) {
			if (track == 0 && side == 0)
				continue;

			TrackInfo trk;
			dsk->GetTrackInfo(trk, side, track);
			if (trk.sectors > gSectors) {
				if (gSectors > 0) {
					fprintf(stderr, "xfile: warning: track %d has %d sectors (more than previous max %d)\n", track, trk.sectors, gSectors);
				}
				gSectors = trk.sectors;
			}
	
			for (int s = 0; s < trk.sectors; s++) {
				SectorInfo sec;
				dsk->GetSectorInfo(sec, side, track, s);
				length += sec.si_length;
			}
		}
	}
	//printf("floppy size = %d (%d blocks)\n", length, length / 512);

	// Read all blocks in order
	char *blocks = new char[length];
	int index = 0;

	// iterate, skipping track 0
	for (int track = 1; track < dc->tracks; track++) {
		for (int side = 0; side < dc->sides; side++) {
			TrackInfo trk;
			dsk->GetTrackInfo(trk, side, track);
			SectorInfo *sec = new SectorInfo[trk.sectors];
			int *order = new int[trk.sectors];
			for (int s = 0; s < trk.sectors; s++) {
				dsk->GetSectorInfo(sec[s], side, track, s);
				int snum = sec[s].si_sector;
				// assumed 1..sectors labels
				if (snum < 1 || snum > trk.sectors) {
					fprintf(stderr, "xfile: sector id %d is too large\n", snum);
					delete [] blocks;
					return NULL;
				}
				order[snum - 1] = s;
			}
			// Go in order.
			for (int s = 0; s < trk.sectors; s++) {
				unsigned char sdata[16384];
				dsk->GetSectorData(sdata, side, track, order[s]);
				SectorInfo *sinfo = &sec[ order[s] ];
	
				memcpy(&blocks[index], sdata, sinfo->si_length);
				index += sinfo->si_length;
			}
	
			delete [] sec;
			delete [] order;
		}
	}
	delete dsk;

	return blocks;
}

int be_short(const unsigned char *data);
int be_long(const unsigned char *data);

#define SUPERBLOCK	(1)
#define ROOT_INODE	(2)

void ReadBlock(int blknum, unsigned char *data)
{
	if (floppy != NULL) {
		if (blknum < 0 || blknum >= floppy_len / 512) {
			fprintf(stderr, "xfile: error: block %d out of range; aborting.\n", blknum);
			exit(1);
		}

		memcpy(data, &floppy[blknum * 512], 512);

		return;
	}

	int head = ((blknum / gSectors) + gStartHead) % gHeads;
	int cylinder = ((blknum / gSectors + gStartHead) / gHeads);
	int sector = (blknum % gSectors) + 1;

	if (blknum < 0 || blknum >= gNumInodeBlock + gNumDataBlock) {
		// We could return an empty block but let's just blow up for now.
		fprintf(stderr, "xfile: error: block %d out of range; aborting.\n", blknum);
		exit(1);
	}

	//printf("Reading block %d at (%d,%d,%d) [%02x,%02x,%02x]\n",
	//	blknum,
	//	head, cylinder, sector,
	//	head, cylinder, sector);

	SectorInfo info;
	hd->GetSectorInfo(info, head, cylinder, sector, 512);
	hd->GetSectorData(data, info);
}

#define NUM_DIRECT_BLOCK			10
#define IDX_INDIRECT_BLOCK			10
#define IDX_DOUBLE_INDIRECT_BLOCK	11
#define IDX_TRIPLE_INDIRECT_BLOCK	12
#define NUM_INODE_BLOCK				13

struct Inode {
	int typePerm;
	int numLink;
	int size;
	int uid;
	int gid;
	int block[NUM_INODE_BLOCK];
	time_t atime, mtime, ctime;
	// Devices don't have blocks but put major + minor device numbers in there.
	int major;
	int minor;
};

const char *known_uid_gid[] = {
	"root", "cron", "sys", "bin", "uucp", NULL, "asg"
};

#define I_TYPE(typeperm) (((typeperm) >> 12) & 0xf)
#define I_IS_FILE(typeperm) (I_TYPE(typeperm) == 8)
#define I_IS_DIR(typeperm) (I_TYPE(typeperm) == 4)
#define I_IS_CHAR(typeperm) (I_TYPE(typeperm) == 2)
#define I_IS_FIFO(typeperm) (I_TYPE(typeperm) == 1)
#define I_IS_BLOCK(typeperm) (I_TYPE(typeperm) == 6)
#define I_IS_NAME(typeperm) (I_TYPE(typeperm) == 5)
#define I_SETUID(typeperm) (((typeperm) >> 11) & 1)
#define I_SETGID(typeperm) (((typeperm) >> 10) & 1)
#define I_STICKY(typeperm) (((typeperm) >> 9) & 1)
#define I_OWNER(typeperm) (((typeperm) >> 6) & 7)
#define I_GROUP(typeperm) (((typeperm) >> 3) & 7)
#define I_WORLD(typeperm) ((typeperm) & 7)

void ReadInode(int num, Inode *node)
{
	memset(node, 0, sizeof *node);

	if (num == 0)
		return;

	int block = (num - 1) / 8 + 2;
	int offset = ((num - 1) % 8) * 64;

	unsigned char data[512];

	ReadBlock(block, data);

	unsigned char *id = data + offset;

	node->typePerm = be_short(id + 0);
	node->numLink = be_short(id + 2);
	node->uid = be_short(id + 4);
	node->gid = be_short(id + 6);
	node->size = be_long(id + 8);
	node->atime = be_long(id + 52);
	node->mtime = be_long(id + 56);
	node->ctime = be_long(id + 60);

	if (I_IS_DIR(node->typePerm) || I_IS_FILE(node->typePerm)) {
		for (int i = 0; i < NUM_INODE_BLOCK; i++) {
			unsigned char *blk = id + 12 + i * 3;
			node->block[i] = (blk[0] << 16) | (blk[1] << 8) | blk[2];
		}
	}
	else {
		node->major = id[13];
		node->minor = id[14];
	}
}

void DumpBlock(unsigned char *data)
{
	for (int row = 0; row < 32; row++) {
		printf("%3x:", row * 16);
		for (int col = 0; col < 16; col++) {
			printf(" %02x", data[row * 16 + col]);
		}
		printf("   ");
		for (int col = 0; col < 16; col++) {
			int ch = data[row * 16 + col];
			printf("%c", ch >= ' ' && ch < 127 ? ch : '.');
		}
		printf("\n");
	}
}

struct BlockPos {
	int blknum;
	int offset;
};

BlockPos *gBlockList;

int ListBlock(int level, int blk, int *blkcnt, int pos)
{
	if (blk == 0) {
		int step = 512;
		for (; level > 0; level--)
			step *= 512 / 4;

		return pos + step;
	}

	if (level == 0) {
		gBlockList[*blkcnt].blknum = blk;
		gBlockList[*blkcnt].offset = pos;
		(*blkcnt)++;

		return pos + 512;
	}

	unsigned char data[512];
	ReadBlock(blk, data);

	for (int i = 0; i < 512; i += 4)
		pos = ListBlock(level - 1, be_long(data + i), blkcnt, pos);

	return pos;
}

void ListFileBlock(int inum, int *blkcnt, int *size)
{
	Inode file;
	ReadInode(inum, &file);

	*blkcnt = 0;
	int pos = 0;

	for (int i = 0; i < NUM_DIRECT_BLOCK; i++)
		pos = ListBlock(0, file.block[i], blkcnt, pos);

	pos = ListBlock(1, file.block[IDX_INDIRECT_BLOCK], blkcnt, pos);
	pos = ListBlock(2, file.block[IDX_DOUBLE_INDIRECT_BLOCK], blkcnt, pos);
	pos = ListBlock(3, file.block[IDX_TRIPLE_INDIRECT_BLOCK], blkcnt, pos);

	*size = file.size;
}

int SaveFile(int inum, const char *name)
{
	int blkcnt, size;

	ListFileBlock(inum, &blkcnt, &size);

	FILE *fp = fopen(name, "wb");
	if (!fp) {
		fprintf(stderr, "xfile: could not open '%s' for writing\n", name);
		return 0;
	}

	unsigned char data[512];
	int pos = 0;
	for (int i = 0; i < blkcnt; i++) {
		ReadBlock(gBlockList[i].blknum, data);
		if (gBlockList[i].offset != pos) {
			pos = gBlockList[i].offset;
			if (fseek(fp, pos, SEEK_SET) != 0) {
				fprintf(stderr, "xfile: seek error on '%s'\n", name);
				fclose(fp);
				return 0;
			}
		}
		if (fwrite(data, 512, 1, fp) != 1) {
			fprintf(stderr, "xfile: write error on '%s'\n", name);
			fclose(fp);
			return 0;
		}
		pos += 512;
	}

	if (fclose(fp) == EOF) {
		fprintf(stderr, "xfile: error closing '%s'\n", name);
		return 0;
	}

	// A bit of a lazy way to handle rouding the size.  But it
	// takes explicit action if the block list is much longer than the
	// file size.

	if (truncate(name, size) != 0) {
		fprintf(stderr, "xfile: error setting size of file '%s'\n", name);
		return 0;
	}

	Inode file;
	ReadInode(inum, &file);

	struct utimbuf accmod = { file.atime, file.mtime };
	if (utime(name, &accmod)) {
		fprintf(stderr, "xfile: error setting times of file '%s'\n", name);
		return 0;
	}

	return 1;
}

unsigned char *LoadDir(int inum, int *len);
void ListDir(int inum);
void ClearVisited(void);
void ListTree(int inum, const char *name);
int SaveDir(int inum, const char *name, int depth = 1);
void StatFile(int inum);
int GetInode(const char *path);

void usage(const char *msg)
{
	printf("xfile: %s\n", msg);
	printf("\n");
	printf("xfile version 1.0     http://48k.ca/xfile.html\n");
	printf("Lists and extracts files from Model 16 Xenix hard and floppy drive images.\n");
	printf("\n");
	printf("usage: xfile [-e] [-r] [-v] drive-image-file {files directories}\n");
	printf("\t-r\tRecursively list or extract subdirectories.\n");
	printf("\t-e\tExtract each listed file or directory contents.\n");
	printf("\t\tOtherwise files and directories are only listed.\n");
	printf("\t-v\tPrint volume information.  Repeat for even more info.\n");
	printf("\n");
	printf("Supports .hdv1, DREM and Gesswein MFM emulator .emu format and various\n");
	printf("floppy drive image formats.\n");
	exit(1);
}

int main(int argc, char *argv[])
{
	int verbose = 0;
	int extract = 0;
	int recurseDepth = 1;

	for (; argc > 2 && argv[1][0] == '-'; argc--, argv++) {
		if (strcmp(argv[1], "-v") == 0)
			verbose++;
		else if (strcmp(argv[1], "-e") == 0 || strcmp(argv[1], "-x") == 0)
			extract = 1;
		else if (strcmp(argv[1], "-r") == 0)
			recurseDepth = 10000;
		else
			usage("unknown argument");
	}

	if (argc < 2)
		usage("no hard drive image file given");

	FILE* fp = NULL;

	int try_hd;
	floppy = read_floppy(argv[1], floppy_len, try_hd);
	if (floppy == NULL) {
		if (try_hd) {
			fp = fopen(argv[1], "rb");
			if (!fp) {
				printf("xfile: could not open '%s'\n", argv[1]);
				return 2;
			}
		}
		else
			return 9;
	}

	SectorInfo info;

if (fp != NULL) {
	const char *format_name = "hdv1";
	const char *ext = strrchr(argv[1], '.');

#if defined(__APPLE__) || defined(__linux__)
#define stricmp strcasecmp
#endif

	if (ext && stricmp(ext, ".cfg") == 0)
		format_name = "drem";
	else if (ext && stricmp(ext, ".emu") == 0)
			format_name = "emu";

	DiskImageMetadata meta;
	DriveConfiguration drive_configuration;
	Notifier report;
	bool read_only = true;

	hd = OpenHardDriveImage(fp, read_only, argv[1],
		format_name, NULL, 0, &report, drive_configuration, &meta);

	if (!hd) {
		printf("xfile: %s\n", report.m_notifications[Notifier::NOTIFY_Fail]);
		return 3;
	}

	const DriveConfiguration *dc = hd->GetConfig();
	if (verbose > 0) {
		printf("Disk Image:    %d cylinders, %d heads, %d sectors\n",
			dc->tracks, dc->sides, dc->sectors_per_track);
	}

	unsigned char sector[512];

	// .pvh - Physical Volume Header.
	hd->GetSectorInfo(info, 0, 0, 1, 512);
	hd->GetSectorData(sector, info);

	if (verbose > 2) {
		printf("Physical Volume Header sector:\n");
		DumpBlock(sector);
	}

	if (sector[0] != '.' || sector[1] != 'p' || sector[2] != 'v' || sector[3] != 'h') {
		fprintf(stderr, "xfile: first block is not the physical volume header\n");
		return 1;
	}

	int pvh_size = be_short(sector + 4);

	if (pvh_size < 16) {
		fprintf(stderr, "xfile: physical volume header size < 16 bytes; giving up.\n");
		return 2;
	}

	gCylinders = be_short(sector + 6);
	gHeads = be_short(sector + 8);
	gSectors = be_short(sector + 10);
	gBytes = be_short(sector + 12);

	int precomp_cylinder = (sector[14] << 8) | sector[15];
}

	unsigned char sector[512];

	if (verbose > 0) {
		printf("Volume Header: %d cylinders, %d heads, %d sectors, %d bytes/sector\n",
			gCylinders, gHeads, gSectors, gBytes);
	}

	if (gBytes != 512) {
		fprintf(stderr, "xfile: Can only handle 512 byte sectors.\n");
		return 5;
	}

if (fp != NULL) {
	// .bad - bad track remapping table
	hd->GetSectorInfo(info, 0, 0, 3, 512);
	hd->GetSectorData(sector, info);

	if (verbose > 2) {
		printf("Bad Track Remapping sector:\n");
		DumpBlock(sector);
	}

	if (sector[0] != '.' || sector[1] != 'b' || sector[2] != 'a' || sector[3] != 'd') {
		fprintf(stderr, "xfile: first block is not the physical volume header\n");
		return 1;
	}

	gMaxBadTrack = be_short(sector + 4);
	gNumBadTrack = be_short(sector + 6);

	if (verbose > 0) {
		printf("Bad Tracks   : %d bad tracks out of maximum %d\n", gNumBadTrack, gMaxBadTrack);
		if (verbose > 1 && gNumBadTrack > 0) {
			unsigned char *bt = sector + 8;
			for (int i = 0; i < gNumBadTrack; i++, bt += 3) {
				int cylinder = (bt[0] << 8) | bt[1];
				int head = bt[2];
				printf(" %d/%d", cylinder, head);
			}
			printf("\n");
		}
		printf("\n");
	}

	// It seems fairly obvious that each bad track in the list is mapped in
	// order to the gMaxBadTracks reserved at the end of the disk.
	// But I'm loathe to code that up without an example.
	// Which I could generate with diskinit.

	if (gNumBadTrack > 0) {
		fprintf(stderr, "xfile: Sorry, bad track re-mapping unimplented.\n");
		return 6;
	}
}

	// Block 1 is the superblock, all right, but I don't have a lot of
	// justification for knowing where block 0 starts.
	ReadBlock(SUPERBLOCK, sector);

	if (verbose > 2) {
		printf("Superblock:\n");
		DumpBlock(sector);
	}

	gNumInodeBlock = be_short(sector);
	gNumDataBlock = be_long(sector + 2);
	time_t lastUpdate = be_long(sector + 0x19E);
	int freeBlock = be_long(sector + 0x1A2);
	int freeInode = be_short(sector + 0x1A6);
	int clean = sector[0x1BC] == 70;
	int magic = be_long(sector + 0x1F0);
	int type = be_long(sector + 0x1F4);

	if (verbose > 1) {
		printf("Data blocks: %d, %d free\n", gNumDataBlock, freeBlock);
		printf("Inode blocks: %d\n", gNumInodeBlock);
		printf("Inodes: %d, %d free\n", gNumInodeBlock * 8, freeInode);
		printf("Last Update: %s", ctime(&lastUpdate));
		printf("Clean: %s\n", clean ? "yes" : "no");
		printf("\n");
	}

	if (magic != 0x2b5544) {
		fprintf(stderr, "xfile: superblock not system 3 type (magic=0x%x)\n", magic);
		if (!floppy) {
			fprintf(stderr, "xfile: format may not be understood; aborting.\n");
			return 1;
		}
		else {
			fprintf(stderr, "xfile: floppy disk, so proceeding anyways.\n");
		}
	}

	if (type != 1) {
		fprintf(stderr, "xfile: superblock indicates block size #%d\n", type);
		if (!floppy) {
			fprintf(stderr, "xfile: Only block size #1 (512 bytes) is supported;  aborting.\n");
			return 1;
		}
		else {
			fprintf(stderr, "xfile: floppy disk, so proceeding anyways.\n");
		}
	}

	gBlockList = (BlockPos *)malloc(gNumDataBlock * sizeof *gBlockList);
	if (!gBlockList) {
		fprintf(stderr, "xfile: out of memory\n");
		return 1;
	}

	if (argc == 2) {
		if (recurseDepth > 1) {
			ClearVisited();
			ListTree(2, "/");
		}
		else
			ListDir(ROOT_INODE);

		return 0;
	}

	for (; argc > 2; argc--, argv++) {
		int inum = GetInode(argv[2]);
		if (inum == 0) {
			fprintf(stderr, "%s not found\n", argv[2]);
			continue;
		}
		Inode node;
		ReadInode(inum, &node);

		char *last_slash = strrchr(argv[2], '/');
		char *base_name = last_slash ? last_slash + 1 : argv[2];

		if (extract) {
			if (I_IS_FILE(node.typePerm))
				SaveFile(inum, base_name);
			else if (I_IS_DIR(node.typePerm))
				printf("%d files and directories extracted\n",
					SaveDir(inum, base_name, recurseDepth));
			else
				printf("%s is a special file so not extracted\n", argv[2]);
		}
		else if (I_IS_DIR(node.typePerm)) {
			if (recurseDepth > 1) {
				ClearVisited();
				ListTree(inum, argv[2]);
			}
			else
				ListDir(inum);
		}
		else {
			StatFile(inum);
			printf("%s\n", argv[2]);
		}
	}

	delete [] floppy;

	return 0;
}

int FindInode(int inum, int namelen, const char *name);

int GetInode(const char *path)
{
	int inum = ROOT_INODE;

	while (inum && *path) {
		while (*path == '/')
			path++;

		if (!*path)
			return inum;

		const char *slash = strchr(path, '/');
		int len = slash ? slash - path : strlen(path);

		inum = FindInode(inum, len, path);

		path += len;
	}

	return inum;
}

int FindInode(int inum, int namelen, const char *name)
{
	if (namelen > 14)
		return 0;

	char padname[14];
	memset(padname, 0, sizeof padname);
	memcpy(padname, name, namelen);

	int blkcnt, size;
	ListFileBlock(inum, &blkcnt, &size);

	for (int i = 0; i < blkcnt; i++) {
		unsigned char sector[512];
		ReadBlock(gBlockList[i].blknum, sector);

		for (int ent = 0; ent < 512; ent += 16) {
			int n = be_short(sector + ent);
			if (n && memcmp(padname, (char *)sector + ent + 2, 14) == 0)
				return n;
		}
	}

	return 0;
}

unsigned char *LoadDir(int inum, int *len)
{
	int blkcnt, size;
	ListFileBlock(inum, &blkcnt, &size);

	unsigned char *dir;
	int dircnt = 0;

	for (int pass = 0; pass < 2; pass++) {
		for (int i = 0; i < blkcnt; i++) {
			unsigned char sector[512];
			ReadBlock(gBlockList[i].blknum, sector);

			for (int ent = 0; ent < 512; ent += 16) {
				int n = be_short(sector + ent);
				if (n) {
					if (pass == 1) {
						memcpy(dir + dircnt * 16, sector + ent, 16);
					}
					dircnt++;
				}
			}
		}
		if (pass == 0 && dircnt > 0) {
			dir = (unsigned char *)malloc(dircnt * 16);
			if (!dir) {
				fprintf(stderr, "xfile: out of memory\n");
				exit(1);
			}
			dircnt = 0;
		}
	}

	*len = dircnt;

	return dir;
}

int cmpent(const void *a, const void *b)
{
	// memcmp() compares too much, but everything beyond the first nul
	// should be 0.

	return memcmp((const char *)a + 2, (const char *)b + 2, 14);
}

void ListDir(int inum)
{
	int dircnt;
	unsigned char *dir = LoadDir(inum, &dircnt);

	qsort(dir, dircnt, 16, cmpent);

	for (int i = 0; i < dircnt; i++) {
		StatFile(be_short(dir + i * 16));
		printf("%-14s\n", dir + i * 16 + 2);
	}

	free(dir);
}

// inodes are at most 16 bits so no reason to worry about dynamically
// allocating an array to record which we've seen.
static char visited[0x10000];

void ClearVisited(void)
{
	memset(visited, '\0', sizeof visited);
}

void ListTree(int inum, const char *dirname)
{
	if (visited[inum])
		return;

	if (*dirname == '\0')
		dirname = "/";

	printf("%s\n", dirname);

	if (strcmp(dirname, "/") == 0)
		dirname = "";

	int dircnt;
	unsigned char *dir = LoadDir(inum, &dircnt);

	qsort(dir, dircnt, 16, cmpent);

	for (int i = 0; i < dircnt; i++) {
		Inode file;
		int n = be_short(dir + i * 16);
		ReadInode(n, &file);
		char name[15];
		name[14] = '\0';
		strncpy(name, (char *)dir + i * 16 + 2, 14);
		if (strcmp(name, ".") == 0 || strcmp(name, "..") == 0)
			continue;

		if (I_IS_FILE(file.typePerm))
			printf("%s/%s\n", dirname, name);
		else if (I_IS_DIR(file.typePerm)) {
			char *path = (char *)malloc(strlen(dirname) + 1 + strlen(name) + 1);
			if (!path) {
				fprintf(stderr, "xfile: out of memory\n");
				exit(1);
			}
			sprintf(path, "%s/%s", dirname, name);
			ListTree(n, path);
			free(path);
		}
	}

	free(dir);
}

int SaveTree(int inum, const char *dirname, int depth)
{
	int savecnt = 0;

	if (depth <= 0)
		return savecnt;

	if (visited[inum]) {
		fprintf(stderr, "xfile: warning: '%s' is a directory loop.\n", dirname);
		return savecnt;
	}

	visited[inum] = 1;

#ifdef _WIN32
	#define DIRMODE
#else
	#define DIRMODE , 0777
#endif

	if (*dirname) {
		if (mkdir(dirname DIRMODE) != 0 && errno != EEXIST) {
			fprintf(stderr, "xfile: cannot create directory '%s'\n", dirname);
			return savecnt;
		}

		savecnt++;

		if (chdir(dirname) != 0) {
			fprintf(stderr, "xfile: cannot enter directory '%s'\n", dirname);
			return savecnt;
		}
	}

	int dircnt;
	unsigned char *dir = LoadDir(inum, &dircnt);

	for (int i = 0; i < dircnt; i++) {
		Inode file;
		int n = be_short(dir + i * 16);
		ReadInode(n, &file);
		char name[15];
		name[14] = '\0';
		strncpy(name, (char *)dir + i * 16 + 2, 14);
		if (strcmp(name, ".") == 0 || strcmp(name, "..") == 0)
			continue;

		if (I_IS_FILE(file.typePerm)) {
			SaveFile(n, name);
			savecnt++;
		}
		else if (I_IS_DIR(file.typePerm))
			savecnt += SaveTree(n, name, depth - 1);
	}

	if (*dirname) {
		if (chdir("..") != 0) {
			fprintf(stderr, "xfile: cannot go up a directory!  Aborting.\n");
			exit(2);
		}
	}

	free(dir);

	return savecnt;
}

int SaveDir(int inum, const char *name, int depth)
{
	ClearVisited();
	return SaveTree(inum, name, depth);
}

static const char *monthName[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

void StatFile(int inum)
{
	Inode file;
	ReadInode(inum, &file);
	int typeperm = file.typePerm;
	char type = '?';

	// Need examples of these
    // p  =  entry is a named pipe
    // s  =  entry is a semaphore
    // m  =  entry is a shared data (memory) file

	if (I_IS_FILE(typeperm))
		type = '-';
	else if (I_IS_DIR(typeperm))
		type = 'd';
	else if (I_IS_CHAR(typeperm))
		type = 'c';
	else if (I_IS_FIFO(typeperm))
		type = 'p';
	else if (I_IS_BLOCK(typeperm))
		type = 'b';
	else if (I_IS_NAME(typeperm))
		type = 'n'; // should be 'm' or 's', I guess.

	printf("%c", type);

	printf("%c%c%c",
		(I_OWNER(typeperm) & 4) ? 'r' : '-',
		(I_OWNER(typeperm) & 2) ? 'w' : '-',
		I_SETUID(typeperm) ? 's' : (I_OWNER(typeperm) & 1) ? 'x' : '-');

	printf("%c%c%c",
		(I_GROUP(typeperm) & 4) ? 'r' : '-',
		(I_GROUP(typeperm) & 2) ? 'w' : '-',
		I_SETGID(typeperm) ? 's' : (I_GROUP(typeperm) & 1) ? 'x' : '-');

	printf("%c%c%c",
		(I_WORLD(typeperm) & 4) ? 'r' : '-',
		(I_WORLD(typeperm) & 2) ? 'w' : '-',
		I_STICKY(typeperm) ? 't' : (I_WORLD(typeperm) & 1) ? 'x' : '-');

	printf(" %3d ", file.numLink);

	int known = sizeof known_uid_gid / sizeof known_uid_gid[0];

	if (file.uid >= 0 && file.uid < known)
		printf("%-8s ", known_uid_gid[file.uid]);
	else
		printf("%8d ", file.uid);

	if (file.gid >= 0 && file.gid < known)
		printf("%-8s ", known_uid_gid[file.gid]);
	else
		printf("%8d ", file.gid);

	if (I_IS_DIR(typeperm) || I_IS_FILE(typeperm))
		printf("%7d ", file.size);
	else
		printf("%3d,%3d ", file.major, file.minor);

	// Not sure what the algorithm for date printing is in ls.
	// For now I'll just print the time if the timestamp is in
	// our current month.

	time_t tnow = time(0);
	struct tm *now = localtime(&tnow);
	int current_month = now->tm_mon;
	int current_year = now->tm_year;

	struct tm *date = localtime(&file.mtime);

	printf("%s %2d ", monthName[date->tm_mon], date->tm_mday);

	if (date->tm_year == current_year && date->tm_mon == current_month)
		printf("%2d:%02d ", date->tm_hour, date->tm_min);
	else
		printf(" %4d ", 1900 + date->tm_year);
}

int be_short(const unsigned char *data)
{
	return (data[0] << 8) | data[1];
}

int be_long(const unsigned char *data)
{
	return (be_short(data) << 16) | be_short(data + 2);
}

// For reference.

#if 0
// Exceprt from sys/param.h with values needed for superblock

#define NICINOD 100	/* number of superblock inodes */
#define NICFREE 50	/* number of superblock free blocks */
#define NSBFILL 51	/* aligns s_magic & s_type at end of SB blk */

// daddr_t is long - unsigned 4 byte int
// ushort is unsigned 2 byte int
// ino_t is unsigned 2 byte int
// time_t is unsigned 4 byte int
// off_t is unsigned 4 byte int

// Excerpt from sys/filesys.h describing the superblock

struct filsys {
	ushort	s_isize;			/* size in blocks of i-list */
	daddr_t	s_fsize;			/* size in blocks of entire volume */
	short	s_nfree;			/* number of addresses in s_free */
	daddr_t	s_free[NICFREE];	/* free block list */
	short	s_ninode;			/* number of i-nodes in s_inode */
	ino_t	s_inode[NICINOD];	/* free i-node list */
	char	s_flock;			/* lock during free list manipulation */
	char	s_ilock;			/* lock during i-list manipulation */
	char  	s_fmod; 			/* super block modified flag */
	char	s_ronly;			/* mounted read-only flag */
	time_t	s_time; 			/* last super block update */
	daddr_t	s_tfree;			/* total free blocks*/
	ino_t	s_tinode;			/* total free inodes */
	short   s_dinfo[4];     	/* device information */
	char	s_fname[6];			/* file system name */
	char	s_fpack[6];			/* file system pack name */
	/* remainder is maintained for xenix */
	char   	s_clean;   			/* S_CLEAN if structure is properly closed */
	char    s_fill[NSBFILL];	/* space to make sizeof filsys be BSIZE */
	long    s_magic;        	/* indicates version of filsys */
	long	s_type;				/* type of new file system */
};
#define	S_CLEAN	0106        	/* arbitrary magic value  */
/* s_magic, magic value for file system version */
#define	S_S3MAGIC	0x2b5544	/* system 3 arbitrary magic value */
/* s_type, block size of file system */
#define	S_B512		1	/* 512 byte block */
#define	S_B1024		2	/* 1024 byte block */
/* codes for file system version (for utilities) */
#define	S_V2		1		/* version 7 */
#define	S_V3		2		/* system 3 */

// Funnily enough NSBFILL doesn't pad out struct filsys to BSIZE (512) bytes.
// It only pads so that s_magic is at 0x1F0.

// Excerpt from sys/ino.h detailing inode storage on disk.
/* Inode structure as it appears on a disk block. */
struct dinode {
	ushort	di_mode;		/* mode and type of file */
	short	di_nlink;		/* number of links to file */
	ushort	di_uid;			/* owner's user id */
	ushort	di_gid;			/* owner's group id */
	off_t	di_size;		/* number of bytes in file */
	char	di_addr[40];	/* disk block addresses */
	time_t	di_atime;		/* time last accessed */
	time_t	di_mtime;		/* time last modified */
	time_t	di_ctime;		/* time created */
};
/*
 * the 40 address bytes:
 *	39 used; 13 addresses
 *	of 3 bytes each.
 */

// Exceprt from sys/stat.h enumerating the type and permission bits.

#define	S_IFMT	0170000		/* type of file */
#define		S_IFDIR	0040000	/* directory */
#define		S_IFCHR	0020000	/* character special */
#define		S_IFBLK	0060000	/* block special */
#define		S_IFREG	0100000	/* regular */
#define		S_IFIFO	0010000	/* fifo */
#define		S_IFNAM	0050000	/* name space entry */
#define	S_INSEM			01	/* semaphore */
#define	S_INSHD			02	/* shared memory */
#define	S_ISUID		04000	/* set user id on execution */
#define	S_ISGID		02000	/* set group id on execution */
#define	S_ISVTX		01000	/* save swapped text even after use */
#define	S_IREAD		00400	/* read permission, owner */
#define	S_IWRITE	00200	/* write permission, owner */
#define	S_IEXEC		00100	/* execute/search permission, owner */

/* From a message sent to Valerie See.

May 13 07:16 1985  file.find Page 1

From uucp Fri May 10 18:44:47 1985
>From tif Fri May 10 11:38:34 1985 remote from techsup
To: qat!valsee
From: techsup!tif  (Paul Chamberlain x3978)
Sent: Fri May 10 1985, 11:38:27 CDT
Subject: finding files in XENIX
Return-receipt-to: techsup!diag1!tif
Cc: qat!root

Here is the information we talked about -- finding files in a XENIX file system.

To find a file in a ENIX file system you must follow these or similar steps:

note: all numbers in decimal unless otherwise noted

1.  Determine the device the file is located on (i.e. hd0, hd1, etc.).
2.  Determine the inode number of the file (try "l -i").
3.  Read (via zap or similar) the block containing the inode on the device,
    where the block number is given by this formula:  2 + (inode - 1) / 8
4.  The file's inode will be located within that block at the offset which is
    given by this formula:  64 * ( ( inode - 1 ) mode 8 )
5.  The inode will contain 64 bytes in the following format

        byte    length  description
        ----    ------  ------------------------------------------------------
          0       2     "mode" of file
          2       2     number of links to file
          4       2     owner user id number
          6       2     owner group id number
          8       4     length of file (It is my belief that the length will
                        include any "holes" caused by seeking past a block)
         12      30     pointers to blocks used by file,  room for ten 3-byte
                        block numbers
                        .
                        .
                        .
         42       3     block number (said to be "singly indirect") of a block
                        containing 128 4-byte block pointers to other blocks
                        in the file
         45       3     block number (said to be "doubly indirect")
         48       3     block number (said to be "triply indirect")
         51       1     -- unused --
         52       4     time last accessed (in seconds since 1/1/70 00:00:00)
         56       4     time last modified
         60       4     time created

6.  The file will reside in all blocks referred to by the above block pointers.
    If all pointers previous to the last pointer are not used (are not 0) then
    the file is said to have a "hole" in it.  This means the file does notuse
    disk space for the entire file.  This "hole" where the block pointer is 0
    will read as nulls (hex 00).

This information summarized from filsys(5) by Paul Chamberlain.

                                        Paul Chamberlain

May 13 07:16 1985  file.find Page 2


                                        Technical Support, Diagnostics Group
                                        400 Atrium, One Tandy Center
                                        Fort Worth, Texas  76102
                                        (ihnp4,trsvax}!sys1|techsup!diag1!tif

*/
#endif
