#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <utime.h>
#include <errno.h>
#include <dirent.h>
#include <time.h>
#include <sys/stat.h>

#include "fd_image.h"
#include "hd_image.h"

#ifdef _MSC_VER
#include <io.h> // _chsize
#define stricmp _stricmp
#else
#define stricmp strcasecmp
#endif

void process_sector(char* block, int length);

#define TRACK_COUNT (77)

#define TBLOCK	512
#define NAMSIZ	100

// for mkdir()
#ifdef _WIN32
	#define DIRMODE
#else
	#define DIRMODE , 0777
#endif

struct FileInfo {
	char name[NAMSIZ];
	char perm[8];		// mode
	char uid[8];	// uid
	char gid[8];	// gid
	char size[12];		// size
	char mtime[12];	// mtime
	char checksum[8];	// checksum
	char linkflag[1];
	char linkname[NAMSIZ];
	// Xenix extensions
	char part[4];
	char numpart[4];
	char parttotal[12];	// total size of all parts
	char rest[255-20];		// zeros
};

struct FileInfo2 {
	char name[NAMSIZ + 1];
	int typePerm;
	int size;
	int uid;
	int gid;
	time_t mtime;
	int checksum;
	int linkflag;
	char linkname[NAMSIZ + 1];
	// Xenix extensions
	int part;
	int numpart;
	int parttotal;	// total size of all parts
};

void CopyInfo(const FileInfo* src, FileInfo2* dst)
{
	int i;

	size_t len = sizeof(src->name);
	strncpy(dst->name, src->name, len);
	dst->name[len] = 0;

	sscanf(src->perm, "%o", &i);
	dst->typePerm = i;
	sscanf(src->uid, "%o", &i);
	dst->uid = i;
	sscanf(src->gid, "%o", &i);
	dst->gid = i;
	sscanf(src->size, "%o", &i);
	dst->size = i;
	sscanf(src->mtime, "%o", &i);
	dst->mtime = i;
	sscanf(src->checksum, "%o", &i);
	dst->checksum = i;

	dst->linkflag = 0;
	dst->linkname[0] = 0;
	if (src->linkflag[0] == '1') {
		dst->linkflag = 1;
		size_t len = sizeof(src->linkname);
		strncpy(dst->linkname, src->linkname, len);
		dst->linkname[len] = 0;
	}
	dst->part = dst->numpart = dst->parttotal = 0;
	if (src->part[0] != 0) {
		sscanf(src->part, "%o", &i);
		dst->part = i;
		sscanf(src->numpart, "%o", &i);
		dst->numpart = i;
		sscanf(src->parttotal, "%o", &dst->parttotal);
	}
}

void WriteInfo(FileInfo *dst, const char* name, int perm, int uid, int gid, int size, int mtime, int part = 0, int numpart = 0, int totalsize = 0)
{
	memset(dst, 0, sizeof(FileInfo));

	strcpy(dst->name, name);
	sprintf(dst->perm, "%6o ", perm);
	sprintf(dst->uid, "%6o ", uid);
	sprintf(dst->gid, "%6o ", gid);
	sprintf(dst->size, "%o ", size);
	sprintf(dst->mtime, "%o ", mtime);

	if (part != 0) {
		sprintf(dst->part, "%o", part);
		sprintf(dst->numpart, "%o", numpart);
		sprintf(dst->parttotal, "%o", totalsize);
	}
}

const char *known_uid_gid[] = {
	"root", "cron", "sys", "bin", "uucp", NULL, "asg"
};

#define I_TYPE(typeperm) (((typeperm) >> 12) & 0xf)
#define I_IS_FILE(typeperm) (I_TYPE(typeperm) == 8)
#define I_IS_DIR(typeperm) (I_TYPE(typeperm) == 4)
#define I_IS_CHAR(typeperm) (I_TYPE(typeperm) == 2)
#define I_IS_FIFO(typeperm) (I_TYPE(typeperm) == 1)
#define I_IS_BLOCK(typeperm) (I_TYPE(typeperm) == 6)
#define I_IS_NAME(typeperm) (I_TYPE(typeperm) == 5)
#define I_SETUID(typeperm) (((typeperm) >> 11) & 1)
#define I_SETGID(typeperm) (((typeperm) >> 10) & 1)
#define I_STICKY(typeperm) (((typeperm) >> 9) & 1)
#define I_OWNER(typeperm) (((typeperm) >> 6) & 7)
#define I_GROUP(typeperm) (((typeperm) >> 3) & 7)
#define I_WORLD(typeperm) ((typeperm) & 7)

static const char *monthName[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

void StatFile(FileInfo2 file)
{
	int typeperm = file.typePerm;
	char type = '?';

	// Tar only has permissions
	// Need examples of these
    // p  =  entry is a named pipe
    // s  =  entry is a semaphore
    // m  =  entry is a shared data (memory) file

//	if (I_IS_FILE(typeperm))
		type = '-';
//	else if (I_IS_DIR(typeperm))
//		type = 'd';
//	else if (I_IS_CHAR(typeperm))
//		type = 'c';
//	else if (I_IS_FIFO(typeperm))
//		type = 'p';
//	else if (I_IS_BLOCK(typeperm))
//		type = 'b';
//	else if (I_IS_NAME(typeperm))
//		type = 'n'; // should be 'm' or 's', I guess.

	printf("%c", type);

	printf("%c%c%c",
		(I_OWNER(typeperm) & 4) ? 'r' : '-',
		(I_OWNER(typeperm) & 2) ? 'w' : '-',
		I_SETUID(typeperm) ? 's' : (I_OWNER(typeperm) & 1) ? 'x' : '-');

	printf("%c%c%c",
		(I_GROUP(typeperm) & 4) ? 'r' : '-',
		(I_GROUP(typeperm) & 2) ? 'w' : '-',
		I_SETGID(typeperm) ? 's' : (I_GROUP(typeperm) & 1) ? 'x' : '-');

	printf("%c%c%c",
		(I_WORLD(typeperm) & 4) ? 'r' : '-',
		(I_WORLD(typeperm) & 2) ? 'w' : '-',
		I_STICKY(typeperm) ? 't' : (I_WORLD(typeperm) & 1) ? 'x' : '-');

	printf(" "); //printf(" %3d ", file.numLink);

	int known = sizeof known_uid_gid / sizeof known_uid_gid[0];

	if (file.uid >= 0 && file.uid < known)
		printf("%-8s ", known_uid_gid[file.uid]);
	else
		printf("%8d ", file.uid);

	if (file.gid >= 0 && file.gid < known)
		printf("%-8s ", known_uid_gid[file.gid]);
	else
		printf("%8d ", file.gid);

//	if (I_IS_DIR(typeperm) || I_IS_FILE(typeperm))
		printf("%7d ", file.size);
//	else
//		printf("%3d,%3d ", file.major, file.minor);

	// Not sure what the algorithm for date printing is in ls.
	// For now I'll just print the time if the timestamp is in
	// our current month.

	time_t tnow = time(0);
	struct tm *now = localtime(&tnow);
	int current_month = now->tm_mon;
	int current_year = now->tm_year;

	struct tm *date = localtime(&file.mtime);

	printf("%s %2d ", monthName[date->tm_mon], date->tm_mday);

	if (date->tm_year == current_year && date->tm_mon == current_month)
		printf("%2d:%02d ", date->tm_hour, date->tm_min);
	else
		printf(" %4d ", 1900 + date->tm_year);
}

int checksum( char* block)
{
	char copy[TBLOCK];

	memcpy(copy, block, TBLOCK);

	FileInfo *f = (FileInfo*)copy;
	memset(f->checksum, ' ', sizeof(f->checksum));

	int res = 0;
	for (int i = 0; i < TBLOCK; i++) {
		res += copy[i];
	}

	return res;
}

// Semi-working, very hacky, notable issues
// - pick a 'reasonable' UID, GID
// - breaking into pieces is a prolem
//   - mathematically not hard
//   - need to create new floppies.
// - command line parsing & usage

struct TimeDate {
	int hour, minute, second;
	int day, month, year;
};

void GetNow(TimeDate *td)
{
	time_t tnow = time(0);
	struct tm *now = localtime(&tnow);

	td->second = now->tm_sec;
	td->minute = now->tm_min;
	td->hour = now->tm_hour;
	td->day = now->tm_mday;
	td->month = now->tm_mon;
	td->year = 1900 + now->tm_year;
}

static void GetNow(DiskImageMetadata &meta)
{
	TimeDate now;
	GetNow(&now);

	meta.yyyy = now.year;
	meta.mm = now.month;
	meta.dd = now.day;
	meta.hour = now.hour;
	meta.min = now.minute;
	meta.sec = now.second;
}

char *concat(const char *s1, char sep, const char *s2)
{
	char *str = new char[strlen(s1) + 1 + strlen(s2) + 1];

	sprintf(str, "%s%c%s", s1, sep, s2);

	return str;
}

// 76*16 = 1216 blocks or 622,592 bytes max.
const char* imd_name = NULL;
char out_fname[256];
int out_format = 0; // 0=floppy, 1=tar (raw blocks)
IMDDiskDrive* disk_out;
FILE* tar_out = NULL;
int floppy_count = 1;
int block_track = 1;
int block_sector_index = 0;
int block_total = 0;

IMDDiskDrive* BlankFloppy()
{
	DriveConfiguration config = { 360,	500*1024, DISK_MFM, 1, TRACK_COUNT, 26 };
	DiskImageMetadata meta;

	GetNow(meta);
	meta.trs80_model_family = 2;

	IMDDiskDrive *imd = new IMDDiskDrive(config, &meta);

	imd->MakeTrack(0, 26, 128, 1, 0/*FM*/);
	for (int i = 1; i < TRACK_COUNT; i++) {
		imd->MakeTrack(i, 16, 512, 1, 1/*MFM*/);
	}

	return imd;
}

FILE* BlankTar()
{
	sprintf(out_fname, "%s%02d.tar", imd_name, floppy_count++);

	return fopen(out_fname, "wb");
}

void write_block(const void *data)
{
	if (block_track >= TRACK_COUNT) {
		fprintf(stderr, "out of space on floppy\n");
		exit(0);
	}

	if (tar_out)
		fwrite(data, 1, 512, tar_out);
	else
		disk_out->WriteSector(0, block_track, block_sector_index, 512, (const unsigned char*)data);

	block_sector_index++;
	if (block_sector_index >= 16) {
		block_sector_index = 0;
		block_track++;
	}

	block_total++;
}

int remain_block()
{
// Archives needs two blocks of zero trailers
#define PAD_BLOCKS (2)
	return 16 * 76 - block_total - PAD_BLOCKS;
}

void FinishFloppy()
{
	if (disk_out != NULL) {
		sprintf(out_fname, "%s%02d.imd", imd_name, floppy_count++);

		FILE* fp = fopen(out_fname, "wb");
		disk_out->Save(fp);
		delete disk_out;

		disk_out = NULL;
	}
	if (tar_out != NULL) {
		fclose(tar_out);
		tar_out = NULL;
	}
	printf("%s - %d remaining blocks\n", out_fname, remain_block());
}

void NextFloppy()
{
	FinishFloppy();

	block_track = 1;
	block_sector_index = 0;
	block_total = 0;

	disk_out = NULL;
	if (out_format == 0)
		disk_out = BlankFloppy();
	else
		tar_out = BlankTar();
}

// Multi-part logic can go here.
// Mostly:
// if size_in_blocks + 1 <= block space remaining -- no need to split
// otherwise, split.  block_space_remain - 1 determines
// how big the first piece, simply division determines the others
// Once that's known, you call a modified appendfile() which takes
// offset, part, part number size, etc.
// WriteInfo() can take the extra parameters.
// Note that "size" in a part is the size of the part
//
// 

void appendfile(const char* path, int len, const char* data, int mtime, int part = 0, int numpart = 0, int totalsize = 0)
{
	if (part == 0 && len > remain_block() * 512) {
		// Split into multiple parts

		// Don't want an empty header part, if only two blocks, go to next floppy and try again
		if (remain_block() < 2) {
			NextFloppy();
			appendfile(path, len, data, mtime);

			return;
		}

		// Compute number of parts.
		numpart = 1;
		int p1_len = (remain_block() - 1) * 512;
		int remainder = len - p1_len;

		int max = (16 * 76 - PAD_BLOCKS) * 512;
		numpart += (remainder + max - 1) / max;

		// Do part 1
		appendfile(path, p1_len, data, mtime, 1, numpart, len);
		// Do remaining parts
		for (int part = 2; part <= numpart; part++) {
			NextFloppy();
			appendfile(path, remainder >= max ? max : remainder, &data[p1_len + (part-2)*max], mtime, part, numpart, len);
			remainder -= max;
		}
		return;
	}

	FileInfo head;

	WriteInfo(&head, path, 0777, 3, 3, len, mtime, part, numpart, totalsize);

	memset(head.checksum, ' ', sizeof(head.checksum));
	sprintf(head.checksum, "%6o", checksum((char*)&head));

	write_block(&head);

	int i;
	for (i = 0; i + 512 < len; i += 512)
	{
		write_block(&data[i]);
	}

	int remain = len % 512;
	if (remain) {
		char buf[512];
		memset(buf, 0, 512);
		memcpy(buf, &data[i], remain);
		write_block(buf);
	}
	else {
		// exact multiple of 512
		if (len != 0) {	// no data if len is zero?
			write_block(&data[i]);
		}
	}
}

void write_tree(const char *path)
{
	// No error on open as directory since we may have a file.
	DIR *dp = opendir(path);
	if (dp) {
		for (;;) {
			struct dirent *dent = readdir(dp);
			if (!dent)
				break;

			if (strcmp(dent->d_name, ".") == 0 || strcmp(dent->d_name, "..") == 0)
				continue;

			const char *file = concat(path, '/', dent->d_name);
			write_tree(file);
			delete [] file;
		}
		closedir(dp);
		return;
	}

	//printf("%s\n", path);
	FILE *fp = fopen(path, "rb");
	if (!fp)
		fprintf(stderr, "xtar: error opening %s\n", path);
	else if (fseek(fp, 0, SEEK_END) == EOF) {
		fprintf(stderr, "xtar: seek error on %s\n", path);
		fclose(fp);
	}
	else {
		struct stat finfo;

		(void) stat(path, &finfo);

		size_t len = ftell(fp);
		if (len == 0)
			appendfile(path, 0, "", finfo.st_mtime);
		else {
			char *data = new char[len];
			if (fseek(fp, 0, SEEK_SET) == EOF)
				fprintf(stderr, "xtar: seek error on %s\n", path);
			else if (fread(data, len, 1, fp) != 1)
				fprintf(stderr, "xtar: read error on %s\n", path);
			else
				appendfile(path, len, data, finfo.st_mtime);
			delete [] data;
		}
		fclose(fp);
	}
}

int verbose;
int	extract;

#define SCAN_ENTRY (0)
#define SCAN_EXTENSION (1)
#define SCAN_BLOCK (2)
int scan_state = 0;

FILE* entry_fp;
int entry_remain;

FileInfo2 curr_entry;
FileInfo2 prev_entry;

void blk_entry(char *block);
void blk_append_entry(char* block);
void blk_open();
void blk_close();

char *get_path(FileInfo2& entry)
{
	int index = 0;

	while (entry.name[index] != 0 && entry.name[index] == '/')
		index++;

	return entry.name + index;
}

void process_sector(char* block, int length)
{
	for (int i = 0; i < length; i += TBLOCK)
	{
		int sum = checksum(block);
		FileInfo *fif = (FileInfo *)(block + i);

		int csum = 0;
		sscanf(fif->checksum, "%o", &csum);
		switch (scan_state)
		{
			case SCAN_ENTRY:
			case SCAN_EXTENSION:
				// block[0] test is to ignore "bogus" file names
				if (csum == sum && block[0] != 0) {
					blk_entry(block);
				}
				break;
			case SCAN_BLOCK:
				blk_append_entry(block);
				break;
		}
	}
}

void blk_entry(char *block)
{
	FileInfo *fif = (FileInfo *)(block);

	CopyInfo(fif, &curr_entry);
	StatFile(curr_entry);
	printf("%s", curr_entry.name);
	if (curr_entry.linkflag) {
		printf(" HARDLINK TO %s", curr_entry.linkname);
	}
	if (curr_entry.part != 0) {
		printf(" PART %d/%d size=%d", curr_entry.part, curr_entry.numpart, curr_entry.parttotal);
	}
	printf("\n");

	if (curr_entry.linkflag) {
		printf("  [ignored hard link]\n");
		return;
	}

	blk_open();
	scan_state = SCAN_BLOCK;
	if (curr_entry.size == 0) {
		blk_close();
	}
}

void blk_append_entry(char* block)
{
	if (entry_fp) {
		fwrite(block, 1, entry_remain < TBLOCK ? entry_remain : TBLOCK, entry_fp);
	}
	entry_remain -= TBLOCK;
	if (entry_remain <= 0) {
		blk_close();
	}
}

int mk_path(const char* path)
{
	char dirpath[NAMSIZ + 1];

	strcpy(dirpath, path);
	for (char* sep = strchr(dirpath, '/'); sep != NULL; ) {
		*sep = 0;

		if (mkdir(dirpath DIRMODE) != 0 && errno != EEXIST) {
			fprintf(stderr, "xtar: cannot create directory '%s'\n", dirpath);
			return 0;
		}

		*sep = '/';
		sep = strchr(sep + 1, '/');
	}

	return 1;
}

int extension_check()
{
	if (strcmp(curr_entry.name, prev_entry.name) != 0)
		return 0;

	if (curr_entry.parttotal != prev_entry.parttotal)
		return 0;

	if (curr_entry.part != prev_entry.part + 1)
		return 0;

	if (curr_entry.numpart != prev_entry.numpart)
		return 0;

	return 1;
}

void blk_open()
{
	entry_fp = NULL;
	entry_remain = curr_entry.size;

	int is_extension = scan_state == SCAN_EXTENSION;

	if (is_extension) {
		if (!extension_check()) {
			fprintf(stderr, "Incorrect archive, possible missing disk or bad disk order?\n");
			fprintf(stderr, "%s part %d/%d (size %d) not current extension for:\n%s part %d/%d (size %d)\n",
					curr_entry.name, curr_entry.part, curr_entry.numpart, curr_entry.parttotal,
					prev_entry.name, prev_entry.part, prev_entry.numpart, prev_entry.parttotal);
			exit(-1);
		}
	}

	if (extract) {
		char* path = get_path(curr_entry);
		if (!is_extension) {
			// New entry or first part of entry
			if (mk_path(get_path(curr_entry))) {
				entry_fp = fopen(path, "w+b");
				if (entry_fp == NULL) {
					fprintf(stderr, "xtar: cannot create %s, file skipped\n", path);
				}
			}
		}
		else {
			// Extension: open for append
			entry_fp = fopen(path, "ab");
			if (entry_fp == NULL) {
				fprintf(stderr, "xtar: cannot open %s for append, file skipped\n", path);
			}
		}
	}
}

void blk_close()
{
	if (entry_fp) {
		fclose(entry_fp);
		entry_fp = NULL;

		struct utimbuf accmod = { curr_entry.mtime, curr_entry.mtime };
		if (utime(get_path(curr_entry), &accmod)) {
			fprintf(stderr, "xtar: error setting times of file '%s'\n", curr_entry.name);
		}
	}
	if (curr_entry.part == 0 || curr_entry.part >= curr_entry.numpart) {
		scan_state = SCAN_ENTRY;	// look for new entry
	} else {
		scan_state = SCAN_EXTENSION;	// look for continuation
		prev_entry = curr_entry;
	}
}

int ends_with(const char* name, const char* ext)
{
	int ext_len = strlen(ext);
	int len = strlen(name);
	if (len < ext_len)
		return 0;

	return stricmp(name + len - ext_len, ext) == 0;
}

void trunc_ext(char *name, const char* ext)
{
	int ext_len = strlen(ext);
	int len = strlen(name);

	name[len - ext_len] = 0;
}

void usage(const char *msg)
{
	printf("xtar: %s\n", msg);
	printf("\n");
	printf("xtar version 1.0     http://48k.ca/xtar.html\n");
	printf("Lists and extracts files from Model 16 Xenix floppy tar archives.\n");
	printf("\n");
	printf("usage: xtar [-e] [-v] [-dd file] diskimages\n");
	printf("\t-e\tExtract each every file or directory contents.\n");
	printf("\t-dd\tExtract raw block image to file.\n");
	printf("\t-v\tPrint some floppy information.\n");
	printf("\t\tOtherwise files only listed.\n");
	printf("or:    xtar -c floppy paths\n");
	printf("\t-c\twrite multi-part floppy archive to floppy01.imd etc.\n");
	printf("\n");
	printf("Use .tar extension to read or create raw bock files, e.g.\n");
	printf("\txtar file.tar -- list contents\n");
	printf("\txtar -c file.tar paths -- create file01.tar file02.tar etc.\n");
	exit(1);
}

int main(int argc, char *argv[])
{
	DiskImageMetadata meta;
	DriveConfiguration drive_configuration[1] = { 360,	500*1024, DISK_MFM, 1, TRACK_COUNT, 26 };
	Notifier report;

#ifdef TEST_CREATE
	// xtar -c floppy pathname -- writes path multi-floppy archive floppy01.imd, floppy02.imd etc.
	imd_name = "out";
	disk_out = BlankFloppy();

	write_tree("usr");// games/utils");

	// No need to write two zero blocks since block_remain() always reserves two blocks
	// and floppies are initialized to zero
	FinishFloppy();

	return 0;
#endif

	verbose = 0;
	extract = 0;
	const char* dump_file = NULL;

	for (; argc >= 2 && argv[1][0] == '-'; argc--, argv++) {
		if (strcmp(argv[1], "-v") == 0)
			verbose++;
		else if (strcmp(argv[1], "-e") == 0 || strcmp(argv[1], "-x") == 0)
			extract = 1;
		else if (strcmp(argv[1], "-dd") == 0) {
			if (argc < 3)
				usage("no file argument to -dd");
			dump_file = argv[2];
			argv++;
			argc--;
		}
		else if (strcmp(argv[1], "-c") == 0) {
			if (argc < 3)
				usage("no file argument to -c");
			imd_name = argv[2];
			out_format = 0;
			if (ends_with(imd_name, ".tar")) {
				trunc_ext((char*)imd_name, ".tar");
				out_format = 1;
			}
			else if (ends_with(imd_name, ".imd")) {
				trunc_ext((char*)imd_name, ".imd");
			}
			argv++;
			argc--;
		}
		else
			usage("unknown argument");
	}

	if (imd_name) {
		if (argc < 2) {
			usage("no path given to put in floppy");
		}
		else {
			tar_out = NULL;
			disk_out = NULL;
			if (out_format == 0)
				disk_out = BlankFloppy();
			else
				tar_out = BlankTar();

			for (int arg = 1; arg < argc; arg++) {
				write_tree(argv[arg]);
			}
			FinishFloppy();
		}

		return 0;
	}

	if (argc < 2) {
		usage("no floppy image files given");
	}

	FILE* tar_out = NULL;
	if (dump_file) {
		if (verbose)
			printf("Writing raw image to %s...\n", dump_file);

		tar_out = fopen(dump_file, "wb");
		if (tar_out == NULL) {
			fprintf(stderr, "Can't write to %s\n", dump_file);
			return -1;
		}
	}

	for (int arg = 1; arg < argc; arg++) {
		const char* error = NULL;

		const char* path = argv[arg];
		if (ends_with(path, ".tar")) {
			FILE* f = fopen(path, "rb");
			int len = 512;
			char buf[512];
			while (fread(buf, 1, len, f) == len) {
				if (tar_out)
					fwrite(buf, 1, len, tar_out);

				process_sector(buf, len);
			}
			fclose(f);

			continue;
		}
		DiskDrive* dsk = LoadDisk(path, NULL, 0, error, drive_configuration[0]);
		if (error != NULL) { 
			fprintf(stderr, "xtar: error loading %s: %s\n", path, error);
			return -1;
		}
	
		const DriveConfiguration *dc = dsk->GetConfig();
		int num_sides = dsk->GetNumSides();
		int num_tracks = dc->tracks;
		if (verbose)
			printf("%s: sides %d tracks %d sectors %d\n", path, num_sides, num_tracks, dc->sectors_per_track);
	
		// iterate, skipping track 0 of side 0
		int z = 0;
		for (int track = 0; track < num_tracks; track++) {
			for (int side = 0; side < num_sides; side++) {
				// skip track 0 of side 0
				if (track == 0 && side == 0)
					continue;

				TrackInfo trk;
				dsk->GetTrackInfo(trk, side, track);
				if (verbose) {
					if (num_sides < 2)
						printf("trk %02d %02d sectors\n", track, trk.sectors); 
					else
						printf("trk %02d side %d %02d sectors\n", track, side, trk.sectors); 
				}
				SectorInfo *sec = new SectorInfo[trk.sectors];
				int *order = new int[trk.sectors];
				for (int s = 0; s < trk.sectors; s++) {
					dsk->GetSectorInfo(sec[s], side, track, s);
					int snum = sec[s].si_sector;
					// assumed 1..sectors labels
					if (snum < 1 || snum > trk.sectors) {
						fprintf(stderr, "xtar: sector id %d is too large\n", snum);
						return -1;
					}
					order[snum - 1] = s;
				}
				// Go in order.
				for (int s = 0; s < trk.sectors; s++) {
					unsigned char sdata[16384];
					dsk->GetSectorData(sdata, side, track, order[s]);
					SectorInfo *sinfo = &sec[ order[s] ];

					if (tar_out)
						fwrite(sdata, 1, sinfo->si_length, tar_out);

					process_sector((char*)sdata, sinfo->si_length);
				}

				delete [] sec;
				delete [] order;
			}
		}
		delete dsk;
	}
	if (tar_out)
		fclose(tar_out);

	// Check for correct end state
	if (scan_state == SCAN_EXTENSION) {
		fprintf(stderr, "xtar: last file missing extension, previous entry was:\n");
		fprintf(stderr, "%s part %d/%d (size %d)\n", prev_entry.name, prev_entry.part, prev_entry.numpart, prev_entry.parttotal);

		exit(-1);
	}

	if (scan_state == SCAN_BLOCK && entry_remain > 0) {
		fprintf(stderr, "xtar: insufficient blocks for last file, %d bytes missing\n", entry_remain);

		exit(-1);
	}

	return 0;
}

